/* LaHaShem HaAretz U'Mloah */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "hlib-c.h"
#include "hmm.h"
#include "data.h"
#include "control.h"

static RVec **g_logL = NULL;
static RVec **g_gamma = NULL;
static RVec **g_gamma_r = NULL;
static int g_ll;
static int g_uu;
static RMat *g_lastData = NULL;

/* -- function definition -- */

/* -- utility functions -- */

Model *allocModel(int ii, int uu);
void defineContFeat(Model *m, int i, Real priorVar);
void defineRadialFeat(Model *m, int i);
void displayModel(FILE *f, Model *m);
void freeModel(Model *m);
void initGlobals(int ll, int uu, RMat *data);
void print_model(Model *m);
void randomiseModel(Model *m, int ut, int rab);

/* -- processing functions -- */

Real logLike(Model *m, RMat data);
int *best_state_sequence(Model *m, RMat data);
void force_init_globals(void);
Real update(Model *m, RMat *data, int ll);

/* -- accessor functions -- */
int model_ii(Model *m);
int model_feature_type(Model *m, int i);
Real model_parameter(Model *m, int u, int i, int n);
int model_uu(Model *m);
Real model_a(Model *m, int u, int v);
Real model_b(Model *m, int u);
void set_model_parameter(Model *m, int u, int i, int n, Real x);
void set_model_a(Model *m, int u, int v, Real x);
void set_model_b(Model *m, int u, Real x);

/* -- function declarations -- */

/* -- utility functions -- */

/* allocModel - Allocates memory for a model containing uu states and ii
 *  features.
 */
Model *allocModel(int ii, int uu)
{ Model *m;
  int u;
  if (ii<=0||uu<=0) panic("allocModel(): %s must be>0", ii<=0?"ii":"uu");
  m = (Model *)safe_malloc(sizeof(Model));
  m->ffm = (Ffm **)safe_malloc(uu*sizeof(Ffm *));
  m->uu = uu;
  m->hmm = allocHMM(uu);
  for (u = 0; u<uu; u++) m->ffm[u] = allocFFM(ii);
  return m;}

/* defineContFeat - Defines feature i to be a continuous feature whose
 *  value is in the range (-sigma, sigma] with p = 0.68.  Non-zero mean
 *  features should be translated appropiately.
 */
void defineContFeat(Model *m, int i, Real sigma)
{ struct ContFI *fi;
  void *old;
  int u, ii = m->ffm[0]->ii;
  if (i<0||i>=ii)
  { panic("defineContFeat(): i = %d is outside range [0,%d]", i, ii-1);}
  if (sigma<=0.0) panic("defineContFeat(): sigma = %g must be>0.0", sigma);
  for (u = 0; u<m->uu; u++)
  { fi = (struct ContFI *)safe_malloc(sizeof(struct ContFI));
    fi->initialSigma = sigma;
    old = setFeatType(m->ffm[u], i, FT_CONTINUOUS, (void *)fi);
    if (old!=NULL) free(old);}}

/* defineRadialFeat - Defines feature i to be a radial feature whose
 *  value is in the range [0, 2*pi).
 */
void defineRadialFeat(Model *m, int i)
{ void *old;
  int u, ii = m->ffm[0]->ii;
  if (i<0||i>=ii)
  { panic("defineRadialFeat(): i = %d is outside range [0,%d]", i, ii-1);}
  for (u = 0; u<m->uu; u++)
  { old = setFeatType(m->ffm[u], i, FT_RADIAL, NULL);
    if (old!=NULL) free(old);}}

/* initGlobals - Allocates or reallocates global variables for the modules.
 *  Could be improved to reallocate space only when it is necessary.
 */
void initGlobals(int ll, int uu, RMat *data)
{ int l, u;
  if (g_logL!=NULL)
  { for (l = 0; l<g_ll; l++)
    { for (u = 0; u<g_uu; u++)
      { freeRVec(g_logL[l][u]);
	freeRVec(g_gamma[l][u]);}
      free(g_logL[l]);
      free(g_gamma[l]);}
    for (u = 0; u<g_uu; u++) free(g_gamma_r[u]);
    free(g_logL);
    free(g_gamma);
    free(g_gamma_r);}
  g_logL = (RVec **)safe_malloc(sizeof(RVec *)*ll);
  g_gamma = (RVec **)safe_malloc(sizeof(RVec *)*ll);
  g_gamma_r = (RVec **)safe_malloc(sizeof(RVec *)*uu);
  for (l = 0; l<ll; l++)
  { g_logL[l] = (RVec *)safe_malloc(sizeof(RVec)*uu);
    g_gamma[l] = (RVec *)safe_malloc(sizeof(RVec)*uu);
    for (u = 0; u<uu; u++)
    { g_logL[l][u] = allocRVec(COLUMNS(data[l]));
      g_gamma[l][u] = allocRVec(COLUMNS(data[l]));}}
  for (u = 0; u<uu; u++)
  { g_gamma_r[u] = (RVec *)safe_malloc(sizeof(RVec)*ll);
    for (l = 0; l<ll; l++) g_gamma_r[u][l] = g_gamma[l][u];}
  g_uu = uu;
  g_ll = ll;}

void displayModel(FILE *f, Model *m)
{ int u;
  fprintf(f, "Model Display\n");
  displayHMM(f, m->hmm);
  for (u = 0; u<m->uu; u++)
  { fprintf(f, "-------\n");
    fprintf(f, "U = %3d\n", u);
    fprintf(f, "-------\n\n");
    displayFFM(f, m->ffm[u]);}}

void freeModel(Model *m)
{ int u;
  freeHMM(m->hmm);
  for (u = 0; u<m->uu; u++) freeFFM(m->ffm[u]);
  free(m);}

void print_model(Model *m) {displayModel(stdout, m);}

void randomiseModel(Model *m, int ut, int rab)
{ int u;
  if (rab) randomiseHMM(m->hmm, ut);
  else constantHMM(m->hmm, ut);
  for (u = 0; u<m->uu; u++) randomiseParams(m->ffm[u]);}

/* -- processing functions -- */

/* logLike -
 *  Calculates log likelihood of entire model given the feature data.
 */
Real logLike(Model *m, RMat data)
{ int u, uu = m->uu;
  RVec *logL;
  Real rv;
  logL = (RVec *)safe_malloc(uu*sizeof(RVec));
  for (u = 0; u<uu; u++)
  { assert(ROWS(data)==m->ffm[u]->ii);
    logL[u] = allocRVec(COLUMNS(data));
    FFM_logL(m->ffm[u], logL[u], data);}
  rv = HMM_logL(m->hmm, logL);
  for (u = 0; u<uu; u++) freeRVec(logL[u]);
  free(logL);
  return rv/model_ii(m);}

/* best_state_sequence -
 *  Calculates the best state sequence of entire model given the feature data.
 */
int *best_state_sequence(Model *m, RMat data)
{ int u, uu = m->uu, *rv;
  RVec *logL;
  logL = (RVec *)safe_malloc(uu*sizeof(RVec));
  for (u = 0; u<uu; u++)
  { assert(ROWS(data)==m->ffm[u]->ii);
    logL[u] = allocRVec(COLUMNS(data));
    FFM_logL(m->ffm[u], logL[u], data);}
  rv = HMM_best_state_sequence(m->hmm, logL);
  for (u = 0; u<uu; u++) freeRVec(logL[u]);
  free(logL);
  return rv;}

void force_init_globals(void) {g_lastData = NULL;}

/* update - Performs a single update on the HMM model for the given data. */
Real update(Model *m, RMat *data, int ll)
{ static Hmm *new = NULL;
  int l, u, uu = m->uu;
  Real logL = 0.0;
  if (g_lastData!=data||ll>g_ll||uu>g_uu)
  { g_lastData = data;
    initGlobals(ll, uu, data);}
  if (new==NULL||m->uu!=new->uu)
  { if (new!=NULL) freeHMM(new);
    new = allocHMM(uu);}
  zeroHMM(new);
  for (l = 0; l<ll; l++)
  { for (u = 0; u<uu; u++) FFM_logL(m->ffm[u], g_logL[l][u], data[l]);
    logL += HMM_updateModel(m->hmm, new, g_logL[l], g_gamma[l]);}
  normaliseHMM(new);
  copyHMM(m->hmm, new);
  for (u = 0; u<uu; u++) FFM_maximise(m->ffm[u], data, g_gamma_r[u], ll);
  return logL;}

/* -- accessor functions -- */

int model_ii(Model *m)
{ assert(m!=NULL&&m->uu>0);
  return m->ffm[0]->ii;}

int model_feature_type(Model *m, int i)
{ assert(m!=NULL&&m->uu>0);
  if (i<0||i>=model_ii(m))
  { panic("model_feature_type(): i = %d out of range [0, %d]",
	  i, model_ii(m));}
  return m->ffm[0]->ft[i];}

Real model_parameter(Model *m, int u, int i, int n)
{ assert(m!=NULL&&m->uu>0);
  if (i<0||i>=model_ii(m))
  { panic("model_parameter(): i = %d out of range [0, %d]", i, model_ii(m));}
  if (u<0||u>=m->uu)
  { panic("model_parameter(): u = %d out of range [0, %d]", u, m->uu);}
  switch (m->ffm[u]->ft[i])
  { case FT_CONTINUOUS:
    switch (n)
    { case 0: return ((Param *)m->ffm[u]->p[i])->mu;
      case 1: return ((Param *)m->ffm[u]->p[i])->sigma;
      default: panic("model_parameter(): n = %d out of range [0, 1]", n);}
    case FT_RADIAL:
    switch (n)
    { case 0: return ((VMParam *)m->ffm[u]->p[i])->mean;
      case 1: return ((VMParam *)m->ffm[u]->p[i])->kappa;
      default: panic("model_parameter(): n = %d out of range [0, 1]", n);}
    default:
    panic("model_parameter(): Unrecognised feature type: %d",
	  m->ffm[u]->ft[i]);}
  panic("model_parameter(): Control shouldn't reach this point");
  return 0.0;}

int model_uu(Model *m)
{ assert(m!=NULL);
  return m->uu;}

Real model_a(Model *m, int u, int v)
{ assert(m!=NULL&&m->hmm!=NULL);
  if (u<0||u>=m->uu||v<0||v>=m->uu)
  { panic("model_a(): u = %d or v = %d out of range [0, %d]", u, v, m->uu);}
  return my_exp(MATRIX(m->hmm->logA)[u][v]);}

Real model_b(Model *m, int u)
{ assert(m!=NULL&&m->hmm!=NULL);
  if (u<0||u>=m->uu) panic("model_b(): u = %d out of range [0, %d]", u, m->uu);
  return my_exp(VECTOR(m->hmm->logB)[u]);}

void set_model_parameter(Model *m, int u, int i, int n, Real x)
{ assert(m!=NULL&&m->uu>0);
  if (i<0||i>=model_ii(m))
  { panic("set_model_parameter(): i = %d out of range [0, %d]",
	  i, model_ii(m));}
  if (u<0||u>=m->uu)
  { panic("set_movel_parameter(): u = %d out of range [0, %d]", u, m->uu);}
  if (m->ffm[u]->p[i]==NULL)
  { panic("set_model_parameter(): feature %d is not initialised", i);}
  switch (m->ffm[u]->ft[i])
  { case FT_CONTINUOUS:
    switch (n)
    { case 0:
      ((Param *)m->ffm[u]->p[i])->mu = x;
      break;
      case 1:
      if (x<=0.0)
      { panic("set_model_parameter(): x = %f is out of range (0, inf)", x);}
      ((Param *)m->ffm[u]->p[i])->sigma = x;
      break;
      default:
      panic("set_model_parameter(): n = %d out of range [0, 1]", n);}
    break;
    case FT_RADIAL:
    switch (n)
    { case 0:
      ((VMParam *)m->ffm[u]->p[i])->mean = x;
      break;
      case 1:
      if (x<=0.0)
      { panic("set_model_parameter(): x = %f is out of range (0, inf)", x);}
      ((VMParam *)m->ffm[u]->p[i])->kappa = x;
      break;
      default:
      panic("set_model_parameter(): n = %d out of range [0, 1]", n);}
    break;
    default:
    panic("set_model_paramter(): Unrecognised feature type: %d, %d, %d",
	  m->ffm[u]->ft[i], u, i);}}

void set_model_a(Model *m, int u, int v, Real x)
{ assert(m!=NULL&&m->hmm!=NULL);
  if (u<0||u>=m->uu||v<0||v>=m->uu)
  { panic("set_model_a(): u = %d or v = %d out of range [0, %d]",
	  u, v, m->uu);}
  if (x<0.0||x>1.0) panic("set_model_a(): x = %f is out of range [0, 1]", x);
  MATRIX(m->hmm->logA)[u][v] = my_log(x);}

void set_model_b(Model *m, int u, Real x)
{ assert(m!=NULL&&m->hmm!=NULL);
  if (u<0||u>=m->uu)
  { panic("set_model_b(): u = %d out of range [0, %d]", u, m->uu);}
  if (x<0.0||x>1.0) panic("set_model_b(): x = %f is out of range [0, 1]", x);
  VECTOR(m->hmm->logB)[u] = my_log(x);}

/* Tam V'Nishlam Shevah L'El Borei Olam */
