/* LaHaShem HaAretz U'Mloah */

/* Data model for Factorial Features -
 *  assumes that all feature values are independent of each other, and
 *  distributions are phase-invariant wrt a mean
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "hlib-c.h"
#include "data.h"

#define MIN_STD_DEV    0.01
#define MAX_KAPPA      100.0
#define SQRT_2PI       2.506628	/* calculated as sqrt(2.0*M_PI) */
#define LOG_2PI        1.837877	/* calculated as log(2.0*PI) */

/* typedefs for LogLikeF and MaxF go here */

/* -- global variables -- */

static Real g_initialSigma = 10.0; /* default initial sigma for cont feat */
static Real g_rx = 0.0;		/* used in VonMisesOpt */
static Real g_vonMisesTol = 1e-6; /* tolerance for kappa root finder */

/* -- extern Numerical Recipes functions -- */

extern float lbessi0(float x);
extern float lbessi1(float x);
extern float zriddr(float (*func)(float), float x1, float x2, float xacc);

/* -- function definitions -- */

/* utility functions */

Ffm *allocFFM(int ii);
void displayFFM(FILE *f, Ffm *m);
void freeFFM(Ffm *m);

/* standard evaluation and maximisation functions */

Real logGauss(Real x, void *param);
Real logVonMises(Real x, void *param);
void maxGauss(void *param, Real **x, RVec *w, int ll);
void maxVonMises(void *param, Real **x, RVec *w, int ll);
float vonMisesOpt(float x);

/* accessor functions */

FeatType getFeatType(Ffm *m, int i);
void *getFeatInfo(Ffm *m, int i);
void *getParam(Ffm *m, int i);
int getFeatNum(Ffm *m);
Real getInitialSigma(void);
void setInitialSigma(Real x);

/* setup functions */

void randomiseParams(Ffm *m);
void *setFeatType(Ffm *m, int i, FeatType ft, void *fi);
void setMaxFunc(Ffm *m, int i, MaxF f);
void setLogLikeFunc(Ffm *m, int i, LogLikeF f);
void *setParam(Ffm *m, int i, void *p);

/* model evaluation and maximisation functions */

void FFM_logL(Ffm *m, RVec logL, RMat data);
void FFM_maximise(Ffm *m, RMat *data, RVec *r, int ll);

/* -- function declarations -- */

/* utility functions */

Ffm *allocFFM(int ii)
{ Ffm *m = (Ffm *)safe_malloc(sizeof(Ffm));
  m->p = (void **)safe_malloc(ii*sizeof(void *));
  m->lf = (LogLikeF *)safe_malloc(ii*sizeof(LogLikeF));
  m->mf = (MaxF *)safe_malloc(ii*sizeof(MaxF));
  m->ft = (FeatType *)safe_malloc(ii*sizeof(FeatType));
  m->fi = (void **)safe_malloc(ii*sizeof(void *));
  m->ii = ii;
  for (ii--; ii>=0; ii--)
  { m->ft[ii] = FT_OTHER;
    m->fi[ii] = NULL;
    m->p[ii] = NULL;}
  return m;}

void displayFFM(FILE *f, Ffm *m)
{ int i;
  fprintf(f, "FFM Features: Type       Mean      Variance\n");
  for (i = 0; i<m->ii; i++)
  { switch (m->ft[i])
    { case FT_CONTINUOUS:
      fprintf(f, "%13d %4s %10.2f %13.2f\n", i, "CONT",
	      ((Param *)m->p[i])->mu, ((Param *)m->p[i])->sigma);
      break;
      case FT_RADIAL:
      fprintf(f, "%13d %4s %10.2f %13.2f\n", i, "RAD",
	      ((VMParam *)m->p[i])->mean, ((VMParam *)m->p[i])->kappa);
      break;
      default:
      fprintf(f, "%13d %4s\n", i, "UNK");
      break;}}}

void freeFFM(Ffm *m)
{ int ii = m->ii;
  for (ii--; ii>=0; ii--)
  { if (m->p[ii]!=NULL) free(m->p[ii]);
    if (m->fi[ii]!=NULL) free(m->fi[ii]);}
  free(m->fi);
  free(m->ft);
  free(m->lf);
  free(m->mf);
  free(m->p);
  free(m);}

/* standard evaluation and maximisation functions */

/* kappaEstimate: (Mardia 1972, pp 122-123)
 *  gives an estimate of ML kappa within ~ 0.01 relative error of A^{-1}(R)
 *      |  1/6*r*(12+6*r^2+5*r^4), r<0.45
 *  k = |  31.1282*r^3-47.2724*r^2+26.9889*r-4.3941, 0.45<=r<=0.8
 *      |  1/(2*(1-r)-(1-r)^2-(1-r)^3), r>0.8
 */
Real kappaEstimate(Real r)
{ assert(r>=0.0&&r<=1.0);
  return (r<0.45)?r*(2.0+r*r*(1.0+5.0/6.0*r*r)):
  (r<=0.8)?((31.1282*r-47.2724)*r+26.9889)*r-4.3941:
  /*r<=1.0*/ 1.0/((1.0-r)*r*(3.0-r));}

Real logGauss(Real x, void *param)
{ Param *p = (Param *)param;
  assert(p->sigma>=MIN_STD_DEV);
  x -= p->mu;
  return (-0.5*x*x/(p->sigma*p->sigma))-my_log(p->sigma*SQRT_2PI);}

/* logVonMises -
 *  as per Mardia, K. V. "Statistics of Directional Data", 1972.  p 57
 */
Real logVonMises(Real x, void *param)
{ VMParam *p = (VMParam *)param;
  assert(p->kappa>0.0);
  assert(p->mean>=-PI&&p->mean<PI);
  assert(x>=-PI&&x<PI+1e-6);	/* wear goggles */
  x -= p->mean;
  return p->kappa*cos(x)-LOG_2PI-lbessi0(p->kappa);}

void maxGauss(void *param, Real **x, RVec *weight, int ll)
{ int l, t, tt;
  Real mom1 = 0.0, mom2 = 0.0, scale = 0.0, *w;
  Param *p = (Param *)param;
  for (l = 0; l<ll; l++)
  { w = VECTOR(weight[l]);
    tt = VLENGTH(weight[l]);
    for (t = 0; t<tt; t++)
    { mom1 += x[l][t]*w[t];
      mom2 += x[l][t]*x[l][t]*w[t];
      scale += w[t];}}
  if (scale>0.0)
  { mom1 /= scale;
    mom2 /= scale;
    mom2 = mom2-mom1*mom1;
    if (mom2<SQR(MIN_STD_DEV))
    { /* reportWarn("maxGauss(): Replacing variance estimate %g with %g",
		 mom2, SQR(MIN_STD_DEV)); */
      mom2 = SQR(MIN_STD_DEV);}
    p->mu = mom1;
    p->sigma = sqrt(mom2);}
  /*
  else reportWarn("maxGauss(): Sum of weights %s 0", (scale<0.0)?"<":"==");
  */}

/* maxVonMises -
 *  as per Mardia, K. V. "Statistics of Directional Data", 1972.  p26, p122
 */
void maxVonMises(void *param, Real **x, RVec *weight, int ll)
{ int i, l, t, tt;
  Real sinx = 0.0, cosx = 0.0, rx, scale = 0.0, k0, k, a0, a, factor, *w;
  VMParam *p = (VMParam *)param;
  for (l = 0; l<ll; l++)
  { w = VECTOR(weight[l]);
    tt = VLENGTH(weight[l]);
    for (t = 0; t<tt; t++)
    { sinx += sin(x[l][t])*w[t];
      cosx += cos(x[l][t])*w[t];
      scale += w[t];}}
  if (scale>1e-6)		/* wear goggles */
  { sinx /= scale;
    cosx /= scale;
    rx = sqrt(sinx*sinx+cosx*cosx);
    p->mean = my_atan2(sinx, cosx);
    if (p->mean==PI) p->mean = -PI; /* wear goggles */
    assert(p->mean>=-PI&&p->mean<PI);
    /* It should never be the case that rx>1.0. But for some reason this
       happens sometimes. And it happens by a large margin, i.e. rx=1.068,
       that cannot be attributed simply to roundoff. */
    if (fabs(rx-1.0)<1e-6||rx>1.0) p->kappa = INFINITY;
    else
    /* estimating kappa */
    { g_rx = rx;
      /* initial estimate of kappa */
      k = (k0 = kappaEstimate(rx));
      a = (a0 = vonMisesOpt(k));
      factor = (a>0.0)?-0.02:0.02;
      /* bracketing root of A(k)-R */
      i = 0;
      while (a*a0>0.0)
      { assert(i<10);
	k0 = k;
	a0 = a;
	k = k0*(1.0+factor*pow(2.0, MIN(i, 5)));
	i++;
	a = vonMisesOpt(k);}
      p->kappa = zriddr(vonMisesOpt, k0, k, g_vonMisesTol);
      assert(p->kappa>0.0);}
    if (p->kappa>MAX_KAPPA)
    { /* reportWarn("maxVonMises(): Replacing kappa estimate %g with %g",
		 p->kappa, MAX_KAPPA); */
      p->kappa = MAX_KAPPA;}}
  /*
  else reportWarn("maxVonMises(): Sum of weights %s 0", (scale<0.0)?"<":"==");
  */}

/* vonMisesOpt - called by zriddr to evaluate f(k) = A(k)-r */
float vonMisesOpt(float k) {return my_exp(lbessi1(k)-lbessi0(k))-g_rx;}

/* accessor functions */

FeatType getFeatType(Ffm *m, int i)
{ assert(i>=0&&i<m->ii);
  return m->ft[i];}

void *getFeatInfo(Ffm *m, int i)
{ assert(i>=0&&i<m->ii);
  return m->fi[i];}

void *getParam(Ffm *m, int i)
{ assert(i>=0&&i<m->ii);
  return m->p[i];}

int getFeatNum(Ffm *m) {return m->ii;}

Real getInitialSigma(void) {return g_initialSigma;}

void setInitialSigma(Real x)
{ assert(x>0.0);
  g_initialSigma = x;}

/* setup functions */

void randomiseParams(Ffm *m)
{ int i;
  Real s;
  for (i = 0; i<m->ii; i++)
  { switch(m->ft[i])
    { case FT_CONTINUOUS:
      s = ((struct ContFI *)m->fi[i])->initialSigma;
      ((Param *)m->p[i])->mu = sqrt(s)*rand_gaussian();
      ((Param *)m->p[i])->sigma = s;
      break;
      case FT_RADIAL:
      ((VMParam *)m->p[i])->mean = 2.0*PI*RANDOM-PI;
      ((VMParam *)m->p[i])->kappa = 3.0*PI;
      break;
      case FT_OTHER:
      break;
      default:
      panic("randomizeParams(): Unrecognised feature type: %d", m->ft[i]);}}}

void *setFeatType(Ffm *m, int i, FeatType ft, void *fi)
{ void *old;
  assert(i>=0&&i<m->ii);
  m->ft[i] = ft;
  old = m->fi[i];
  m->fi[i] = fi;
  switch(ft)
  { case FT_CONTINUOUS:
    m->mf[i] = maxGauss;
    m->lf[i] = logGauss;
    if (m->fi[i]==NULL)
    { m->fi[i] = (void *)safe_malloc(sizeof(struct ContFI));
      ((struct ContFI *)m->fi[i])->initialSigma = g_initialSigma;}
    if (m->p[i]==NULL) m->p[i] = safe_malloc(sizeof(Param));
    break;
    case FT_RADIAL:
    m->mf[i] = maxVonMises;
    m->lf[i] = logVonMises;
    if (m->p[i]==NULL) m->p[i] = safe_malloc(sizeof(VMParam));
    break;
    case FT_OTHER:
    break;
    default:
    panic("setFeatType(): Unrecognised feature type: %d", ft);}
  return old;}

void setMaxFunc(Ffm *m, int i, MaxF f)
{ assert(i>=0&&i<m->ii);
  m->mf[i] = f;}

void setLogLikeFunc(Ffm *m, int i, LogLikeF f)
{ assert(i>=0&&i<m->ii);
  m->lf[i] = f;}

/* setParam -
 *  Sets the param value to the new value and returns a pointer to the
 *  old value so that the user can free the associated memory.
 */
void *setParam(Ffm *m, int i, void *p)
{ void *old;
  assert(i>=0&&i<m->ii);
  old = m->p[i];
  m->p[i] = p;
  return old;}

/* model evaluation and maximisation functions */

void FFM_logL(Ffm *m, RVec logL, RMat data)
{ int i, t, tt = COLUMNS(data), ii = m->ii;
  Real **x = MATRIX(data), *l = VECTOR(logL);
  void **p = m->p;
  assert(tt==VLENGTH(logL));
  assert(ROWS(data)==ii);
  /* logL[t] = sum(i = 1->ii, logLike(x[t][i], modelParams[i]) */
  for (t = 0; t<tt; t++)
  { l[t] = 0.0;
    for (i = 0; i<ii; i++)
    { l[t] += (m->lf[i])(x[i][t], p[i]);}}}

/* FFM_maximise -
 *  Given a FFM, feature vectors for a number of movies, and the
 *  responsibility of the model for each feature vector, this function
 *  maximises the parameter values of the FFM.
 */
void FFM_maximise(Ffm *m, RMat *data, RVec *w, int ll)
{ int i, l, ii = m->ii;
  static Real **x = NULL;
  static int last_ll = 0;
  if (ll>last_ll)
  { if (x!=NULL) free(x);
    x = (Real **)safe_malloc(ll*sizeof(Real *));
    last_ll = ll;}
  for (i = 0; i<ii; i++)
  { for (l = 0; l<ll; l++) x[l] = MATRIX(data[l])[i];
    (m->mf[i])(m->p[i], x, w, ll);}}

/* need something which determines which data models match which states */

/* Tam V'Nishlam Shevah L'El Borei Olam */
