#ifndef _HLIBC_H
#define _HLIBC_H

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "hlib-c-safe.h"

/* macros which pretend that they are functions */
#define matrix2d(x, y) __matrix2d(x, y, __LINE__, __FILE__)
#define matrix3d(x, y, z) __matrix3d(x, y, z, __LINE__, __FILE__)
#define matrix4d(x, y, z, w) __matrix4d(x, y, z, w, __LINE__, __FILE__)
/* disabled for now
#define malloc(n) my_malloc(n, __LINE__, __FILE__)
*/
#define vector(x, y) __vector(x, y, __LINE__, __FILE__)
#define matrix(x, y, z, w) __matrix(x, y, z, w, __LINE__, __FILE__)

/* If the below #define is uncommented, messages are displayed that track
   the amound of memory being allocated */
/* #define HLIBC_TRACE_SAFE_MALLOC */

/* global variable declarations */

extern int gtrace[];

/* function declarations */

/* vector allocation and deallocation procedures */

extern RMat *allocate_rmat_vector(size_t n);
extern RVec allocRVec(size_t x);
extern RMat allocRMat(size_t y, size_t x);
extern RMat3d allocRMat3d(size_t z, size_t y, size_t x);
extern void free_rmat_vector(RMat *v);
extern void freeRVec(RVec v);
extern void freeRMat(RMat m);
extern void freeRMat3d(RMat3d m);

/* vector operations */

extern Real *addRVec(Real *u, const Real *v, int size);
extern Real *copyRVec(Real *u, const Real *v, int size);
extern Real dotProdRVec(const Real *u, const Real *v, int size);
extern Real normaliseRVec(Real *v, int size);
extern Real rmat_get(RMat rmat, int i, int j);
extern void rmat_set(RMat rmat, int i, int j, Real x);
extern void rmat_vector_set(RMat *v, int i, RMat r);
extern RMat rmat_vector(RMat *v, int i);
extern Real *scaleRVec(Real *v, Real k, int size);
extern Real sumOfLogs(Real *v, int size);
extern Real sumRVec(Real *v, int size);

/* functions of general use */

extern Real add_exp(Real e1, Real e2);
extern void clip_symmatrix(Real *xx, Real *xy, Real *yy, Real clip);
extern int distinct_reals(Real a, Real b);
extern Real dot_product(Real *u, Real *v, int size);
extern Real eigenvalue1(Real xx, Real xy, Real yx, Real yy);
extern Real eigenvalue2(Real xx, Real xy, Real yx, Real yy);
extern Real eigenvector_angle1(Real xx, Real xy, Real yx, Real yy);
extern Real eigenvector_angle2(Real xx, Real xy, Real yx, Real yy);
extern void eigenvectors(Real xx, Real xy, Real yx, Real yy, Real *exx,
			 Real *exy, Real *eyx, Real *eyy);
extern void free_matrix(Real **m, int nrl, int nrh, int ncl, int nch);
extern void free_matrix2d(Real **m, int n);
extern void free_matrix3d(Real ***a, int x, int y);
extern void free_matrix4d(Real ****a, int x, int y, int z);
extern void free_vector(Real *v, int nl, int nh);
extern int is_nan(Real x);
extern Real *map_real(Real (*f)(Real), Real *v, Real *rv, int size);
#ifdef SUN
extern int matherr(struct exception *exc);
#endif
extern Real **__matrix(int nrl, int nrh, int ncl, int nch, int line,
		       const char *file);
extern Real **__matrix2d(int x, int y, int line, const char *file);
extern Real ***__matrix3d(int x, int y, int z, int line, const char *file);
extern Real ****__matrix4d(int x, int y, int z, int w, int line,
			   const char *file);
extern int max_int(int *v, int size);
extern Real my_add_exp(Real e1, Real e2, int line, const char *file);
extern Real my_exp(Real x);
extern void *my_malloc(int n, int line, const char *file);
extern Real my_sqrt(Real x);
extern Real __my_sqrt(Real x, int line, const char *file);
extern Real my_log(Real x);
extern Real my_atan2(Real x, Real y);
extern Real normalise_rotation(Real rotation);
extern void panic(const char *error_text, ...);
extern void print_version(const char *version, const char *led,
			  const char *file);
extern Real quadratic1(Real a, Real b, Real c);
extern Real quadratic2(Real a, Real b, Real c);
extern Real radial_distance(Real t, Real u);
extern void reportWarn(const char *error_text, ...);
extern void *_safe_malloc(size_t size, int line, char *file);
extern void set_version_printing(int boolean);
extern Real sqr_real(Real x);
extern Real sum_real(Real *v, int size);
extern void trace(int n, const char *format, ...);
extern Real *__vector(int nl, int nh, int line, const char *file);

/* PDFs */

extern Real log_normal(int n , Real *x, Real *mu,
		       Real sqrt_big_sigma_determinant,
		       Real **big_sigma_inverse);
extern Real log_normal_1d(Real x, Real mu, Real sigma);
extern Real normal(int n, Real *x, Real *mu, Real sqrt_big_sigma_determinant,
	    Real **big_sigma_inverse);
extern Real normal_1d(Real x, Real mu, Real sigma);

/* debugging redefines to catch math errors */
#ifdef DEBUG
#define add_exp(x, y) my_add_exp(x, y, __LINE__, __FILE__)
#define my_sqrt(x) __my_sqrt(x, __LINE__, __FILE__)
#endif
#endif
