(MODULE HLIB-INTERPRET)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-interpret.sch")
(include "c-externals.sc")

;;; These assume that angle is in [-pi, pi) where -pi and pi point to the left
;;; and increasing angle denotes counterclockwise movement.

;;; Variables

(define *language* #f)

(define *english-velocity-descriptions*
 '#("leftward"
    "leftward-and-downward"
    "downward"
    "rightward-and-downward"
    "rightward"
    "rightward-and-upward"
    "upward"
    "leftward-and-upward"))

(define *english-relative-position-descriptions*
 '#("to-the-right-of"
    "above-and-to-the-right-of"
    "above"
    "above-and-to-the-left-of"
    "to-the-left-of"
    "below-and-to-the-left-of"
    "below"
    "below-and-to-the-right-of"))

(define *japanese-velocity-descriptions*
 '#("hidari"				;leftward
    "hidarishita"			;leftward and downward
    "shita"				;downward
    "migishita"				;rightward and downward
    "migi"				;rightward
    "migiue"				;rightward and upward
    "ue"				;upward
    "hidariue"))			;leftward and upward

(define *japanese-relative-position-descriptions*
 '#("migi"				;to the right of
    "migiue"				;above and to the right of
    "ue"				;above
    "hidariue"				;above and to the left of
    "hidari"				;to the left of
    "hidarishita"			;below and to the left of
    "shita"				;below
    "migishita"))			;below and to the right of

(define *sigma-threshold* 10000.0)
(define *kappa-threshold* 0.0)
(define *slow-motion-threshold* 10.0)
(define *fast-motion-threshold* 20.0)
(define *acceleration-threshold* 1.0)
(define *slow-acceleration-threshold* 10.0)
(define *fast-acceleration-threshold* 20.0)
(define *rotation-threshold* 5.0)
(define *slow-rotation-threshold* 15.0)
(define *fast-rotation-threshold* 30.0)
(define *growth-threshold* 0.0)
(define *slow-growth-threshold* 50.0)
(define *fast-growth-threshold* 100.0)
(define *stretching-threshold* 0.0)
(define *slow-stretching-threshold* 50.0)
(define *fast-stretching-threshold* 100.0)
(define *distance-threshold* 50.0)
(define *relative-motion-threshold* 1.0)
(define *slow-relative-motion-threshold* 10.0)
(define *fast-relative-motion-threshold* 20.0)
(define *changing-relative-position-threshold* 5.0)
(define *slow-changing-relative-position-threshold* 15.0)
(define *fast-changing-relative-position-threshold* 30.0)
(define *roles* #f)

;;; Procedures

(define (describe-angle angle descriptions)
 ;; angle should be in the range -PI to PI, index should make that range
 ;; into 0 to 7, where [7/8 PI, PI) & [-PI, -7/8 PI) translate into index 0
 ;; [-7/8 PI, -5/8 PI) translates into index 1, etc.
 (let ((index
	(modulo (inexact->exact (truncate (+ (* 4.0 (/ angle pi)) 4.5))) 8)))
  (vector-ref descriptions index)))

(define (describe-angular-velocity angle)
 (case *language*
  ((english) (if (positive? angle) "counterclockwise" "clockwise"))
  ((japanese) (if (positive? angle) "hidarimawarini" "migimawarini"))
  (else (fuck-up))))

(define (describe-rate motion slow fast)
 (case *language*
  ((english)
   (cond ((< motion slow) " slowly") ((< motion fast) "") (else " quickly")))
  ((japanese)
   (cond ((< motion slow) " yukkuri") ((< motion fast) "") (else " subayaku")))
  (else (fuck-up))))

(define (role k) (if (vector? *roles*) (vector-ref *roles* k) k))

(define (describe-motion unknown? psi u k)
 (case *language*
  ((english)
   (if (<= (psi-sigma-magnitude-of-the-velocity-vector psi u k)
	   *sigma-threshold*)
       (let ((motion (psi-mu-magnitude-of-the-velocity-vector psi u k)))
	(if (>= motion *motion-threshold*)
	    (if (>= (psi-kappa-orientation-of-the-velocity-vector psi u k)
		    *kappa-threshold*)
		(format #f "~a is-moving~a ~a"
			(role k)
			(describe-rate motion
				       *slow-motion-threshold*
				       *fast-motion-threshold*)
			(describe-angle
			 (psi-mean-orientation-of-the-velocity-vector psi u k)
			 *english-velocity-descriptions*))
		(format #f "~a is-moving~a"
			(role k)
			(describe-rate motion
				       *slow-motion-threshold*
				       *fast-motion-threshold*)))
	    (format #f "~a is-not-moving" (role k))))
       (if unknown?
	   (format #f "it-is-not-known-whether ~a is-moving" (role k))
	   #f)))
  ((japanese)
   (if (<= (psi-sigma-magnitude-of-the-velocity-vector psi u k)
	   *sigma-threshold*)
       (let ((motion (psi-mu-magnitude-of-the-velocity-vector psi u k)))
	(if (>= motion *motion-threshold*)
	    (if (>= (psi-kappa-orientation-of-the-velocity-vector psi u k)
		    *kappa-threshold*)
		(format #f "~a ga ~a no-hou-he~a ugoiteimasu"
			(role k)
			(describe-angle
			 (psi-mean-orientation-of-the-velocity-vector psi u k)
			 *japanese-velocity-descriptions*)
			(describe-rate motion
				       *slow-motion-threshold*
				       *fast-motion-threshold*))
		(format #f "~a ga~a ugoiteimasu"
			(role k)
			(describe-rate motion
				       *slow-motion-threshold*
				       *fast-motion-threshold*)))
	    (format #f "~a ga-ugoiteimasen" (role k))))
       (if unknown?
	   (format #f "~a ga-ugoiteiru-kadouka-wakarimasen" (role k))
	   #f)))
  (else (fuck-up))))

(define (describe-acceleration unknown? psi u k)
 (case *language*
  ((english)
   (if (<= (psi-sigma-finite-difference-of-the-magnitude-of-the-velocity-vector
	    psi u k)
	   *sigma-threshold*)
       (let ((acceleration
	      (psi-mu-finite-difference-of-the-magnitude-of-the-velocity-vector
	       psi u k)))
	(if (>= acceleration *acceleration-threshold*)
	    (format #f "~a is-accelerating~a"
		    (role k)
		    (describe-rate acceleration
				   *slow-acceleration-threshold*
				   *fast-acceleration-threshold*))
	    (format #f "~a is-not-accelerating" (role k))))
       (if unknown?
	   (format #f "it-is-not-known-whether ~a is-accelerating" (role k))
	   #f)))
  ((japanese)
   (if (<= (psi-sigma-finite-difference-of-the-magnitude-of-the-velocity-vector
	    psi u k)
	   *sigma-threshold*)
       (let ((acceleration
	      (psi-mu-finite-difference-of-the-magnitude-of-the-velocity-vector
	       psi u k)))
	(if (>= acceleration *acceleration-threshold*)
	    (format #f "~a ga~a kasoku-shiteimasu"
		    (role k)
		    (describe-rate acceleration
				   *slow-acceleration-threshold*
				   *fast-acceleration-threshold*))
	    (format #f "~a ga-kasoku-shiteimasen" (role k))))
       (if unknown?
	   (format #f "~a ga-kasoku-shiteiru-kadouka-wakarimasen" (role k))
	   #f)))
  (else (fuck-up))))

(define (describe-rotation unknown? psi u k)
 (case *language*
  ((english)
   (if (>= (psi-kappa-angular-velocity psi u k) *kappa-threshold*)
       (let ((rotation (radians->degrees (psi-mean-angular-velocity psi u k))))
	(if (>= (abs rotation) *rotation-threshold*)
	    (format #f "~a is-rotating~a ~a"
		    (role k)
		    (describe-rate rotation
				   *slow-rotation-threshold*
				   *fast-rotation-threshold*)
		    (describe-angular-velocity rotation))
	    (format #f "~a is-not-rotating" (role k))))
       (if unknown?
	   (format #f "it-is-not-known-whether ~a is-rotating" (role k))
	   #f)))
  ((japanese)
   (if (>= (psi-kappa-angular-velocity psi u k) *kappa-threshold*)
       (let ((rotation (radians->degrees (psi-mean-angular-velocity psi u k))))
	(if (>= (abs rotation) *rotation-threshold*)
	    (format #f "~a ga ~a~a mawatteimasu"
		    (role k)
		    (describe-angular-velocity rotation)
		    (describe-rate rotation
				   *slow-rotation-threshold*
				   *fast-rotation-threshold*))
	    (format #f "~a ga-mawatteimasen" (role k))))
       (if unknown?
	   (format #f "~a ga-mawatteiru-kadouka-wakarimasen" (role k))
	   #f)))
  (else (fuck-up))))

(define (describe-growth unknown? psi u k)
 (case *language*
  ((english)
   (if (<= (psi-sigma-finite-difference-of-the-area psi u k) *sigma-threshold*)
       (let ((growth (psi-mu-finite-difference-of-the-area psi u k)))
	(if (>= (abs growth) *growth-threshold*)
	    (if (positive? growth)
		(format #f "~a is-growing~a"
			(role k)
			(describe-rate growth
				       *slow-growth-threshold*
				       *fast-growth-threshold*))
		(format #f "~a is-shrinking~a"
			(role k)
			(describe-rate (- growth)
				       *slow-growth-threshold*
				       *fast-growth-threshold*)))
	    (format #f "~a is-not-changing-size" (role k))))
       (if unknown?
	   (format #f "it-is-not-known-whether ~a is-changing-size" (role k))
	   #f)))
  ((japanese)
   (if (<= (psi-sigma-finite-difference-of-the-area psi u k) *sigma-threshold*)
       (let ((growth (psi-mu-finite-difference-of-the-area psi u k)))
	(if (>= (abs growth) *growth-threshold*)
	    (if (positive? growth)
		(format #f "~a ga~a ookikunatteikimasu"
			(role k)
			(describe-rate growth
				       *slow-growth-threshold*
				       *fast-growth-threshold*))
		(format #f "~a ga~a chiisakunatteikimasu"
			(role k)
			(describe-rate (- growth)
				       *slow-growth-threshold*
				       *fast-growth-threshold*)))
	    (format #f "~a no-ookisa-ga-kawatteimasen" (role k))))
       (if unknown?
	   (format #f "~a no-ookisa-ga-kawatteiru-kadouka-wakarimasen"
		   (role k))
	   #f)))
  (else (fuck-up))))

(define (describe-stretching unknown? psi u k)
 (case *language*
  ((english)
   (if (<= (psi-sigma-finite-difference-of-the-eccentricity psi u k)
	   *sigma-threshold*)
       (let ((stretching
	      (psi-mu-finite-difference-of-the-eccentricity psi u k)))
	(if (>= (abs stretching) *stretching-threshold*)
	    (if (positive? stretching)
		(format #f "~a is~a becoming-more-flat"
			(role k)
			(describe-rate stretching
				       *slow-stretching-threshold*
				       *fast-stretching-threshold*))
		(format #f "~a is~a becoming-more-round"
			(role k)
			(describe-rate (- stretching)
				       *slow-stretching-threshold*
				       *fast-stretching-threshold*)))
	    (format #f "~a is-not-changing-shape" (role k))))
       (if unknown?
	   (format #f "it-is-not-known-whether ~a is-changing-shape" (role k))
	   #f)))
  ((japanese)
   (if (<= (psi-sigma-finite-difference-of-the-eccentricity psi u k)
	   *sigma-threshold*)
       (let ((stretching
	      (psi-mu-finite-difference-of-the-eccentricity psi u k)))
	(if (>= (abs stretching) *stretching-threshold*)
	    (if (positive? stretching)
		(format #f "~a ga~a hirataku-natteikimasu"
			(role k)
			(describe-rate stretching
				       *slow-stretching-threshold*
				       *fast-stretching-threshold*))
		(format #f "~a ga~a maruku-natteikimasu"
			(role k)
			(describe-rate (- stretching)
				       *slow-stretching-threshold*
				       *fast-stretching-threshold*)))
	    (format #f "~a no-katachi-ga-kawatteimasen" (role k))))
       (if unknown?
	   (format #f "~a no-katachi-ga-kawatteiru-kadouka-wakarimasen"
		   (role k))
	   #f)))
  (else (fuck-up))))

(define (describe-relative-position unknown? psi u k1 k2)
 (case *language*
  ((english)
   (if (<= (psi-sigma-distance-between-ellipse-centres psi u k1 k2)
	   *sigma-threshold*)
       (let ((distance (psi-mu-distance-between-ellipse-centres psi u k1 k2)))
	(if (>= (psi-kappa-orientation-of-the-vector-between-ellipse-centres
		 psi u k1 k2)
		*kappa-threshold*)
	    (let ((relative-position
		   (psi-mean-orientation-of-the-vector-between-ellipse-centres
		    psi u k1 k2)))
	     (format
	      #f "~a is ~a and ~a ~a"
	      (role k1)
	      (if (>= distance *distance-threshold*) "far-from" "close-to")
	      (describe-angle relative-position
			      *english-relative-position-descriptions*)
	      (role k2)))
	    (format
	     #f "~a is ~a ~a"
	     (role k1)
	     (if (>= distance *distance-threshold*) "far-from" "close-to")
	     (role k2))))
       (if (>= (psi-kappa-orientation-of-the-vector-between-ellipse-centres
		psi u k1 k2)
	       *kappa-threshold*)
	   (let ((relative-position
		  (psi-mean-orientation-of-the-vector-between-ellipse-centres
		   psi u k1 k2)))
	    (format
	     #f "~a is ~a ~a"
	     (role k1)
	     (describe-angle relative-position
			     *english-relative-position-descriptions*)
	     (role k2)))
	   (if unknown?
	       (format #f "the-relative-position-of ~a and ~a is-not-known"
		       (role k1) (role k2))
	       #f))))
  ((japanese)
   (if (<= (psi-sigma-distance-between-ellipse-centres psi u k1 k2)
	   *sigma-threshold*)
       (let ((distance (psi-mu-distance-between-ellipse-centres psi u k1 k2)))
	(if (>= (psi-kappa-orientation-of-the-vector-between-ellipse-centres
		 psi u k1 k2)
		*kappa-threshold*)
	    (let ((relative-position
		   (psi-mean-orientation-of-the-vector-between-ellipse-centres
		    psi u k1 k2)))
	     (if (>= distance *distance-threshold*)
		 (format
		  #f "~a ga ~a no ~a no-hanareta-tokoro-ni-arimasu"
		  (role k1)
		  (role k2)
		  (describe-angle relative-position
				  *japanese-relative-position-descriptions*))
		 (format
		  #f "~a ga ~a no-sugu ~a ni-arimasu"
		  (role k1)
		  (role k2)
		  (describe-angle relative-position
				  *japanese-relative-position-descriptions*))))
	    (format #f "~a ga ~a ~a ni-arimasu"
		    (role k1)
		    (role k2)
		    (if (>= distance *distance-threshold*)
			"kara-hanareta-tokoro"
			"no-chikaku"))))
       (if (>= (psi-kappa-orientation-of-the-vector-between-ellipse-centres
		psi u k1 k2)
	       *kappa-threshold*)
	   (let ((relative-position
		  (psi-mean-orientation-of-the-vector-between-ellipse-centres
		   psi u k1 k2)))
	    (format
	     #f "~a ga ~a no ~a ni-arimasu"
	     (role k1)
	     (role k2)
	     (describe-angle relative-position
			     *japanese-relative-position-descriptions*)))
	   (if unknown?
	       (format #f "~a to ~a no-soutaiteki-ichi-ha-wakarimasen"
		       (role k1) (role k2))
	       #f))))
  (else (fuck-up))))

(define (describe-relative-motion unknown? psi u k1 k2)
 ;; needs work: If k2 is moving and k1 is not then switch the argument order.
 (case *language*
  ((english)
   (if (<= (psi-sigma-finite-difference-of-the-distance-between-ellipse-centres
	    psi u k1 k2)
	   *sigma-threshold*)
       (let ((relative-motion
	      (psi-mu-finite-difference-of-the-distance-between-ellipse-centres
	       psi u k1 k2)))
	(if (>= (abs relative-motion) *relative-motion-threshold*)
	    (format #f "~a is-moving~a ~a ~a"
		    (role k1)
		    (describe-rate relative-motion
				   *slow-relative-motion-threshold*
				   *fast-relative-motion-threshold*)
		    (if (positive? relative-motion) "away-from" "towards")
		    (role k2))
	    (format #f "~a is-not-moving-towards-or-away-from ~a"
		    (role k1) (role k2))))
       (if unknown?
	   (format
	    #f "it-is-not-known-whether ~a is-moving-towards-or-away-from ~a"
	    (role k1) (role k2))
	   #f)))
  ((japanese)
   (if (<= (psi-sigma-finite-difference-of-the-distance-between-ellipse-centres
	    psi u k1 k2)
	   *sigma-threshold*)
       (let ((relative-motion
	      (psi-mu-finite-difference-of-the-distance-between-ellipse-centres
	       psi u k1 k2)))
	(if (>= (abs relative-motion) *relative-motion-threshold*)
	    (format #f "~a ga ~a ~a~a ~a"
		    (role k1)
		    (role k2)
		    (if (positive? relative-motion)
			"kara"
			"no-hou-he")
		    (describe-rate relative-motion
				   *slow-relative-motion-threshold*
				   *fast-relative-motion-threshold*)
		    (if (positive? relative-motion)
			"hanarete-ikimasu"
			"ugoiteimasu"))
	    (format #f "~a ga ~a ni-chikazuki-mo-shinai-shi-hanare-mo-shimasen"
		    (role k1) (role k2))))
       (if unknown?
	   (format
	    #f "~a ga ~a kara-hanarete-iku-noka-te-he-chikazuite-iku-noka-wakarimasen"
	    (role k1) (role k2))
	   #f)))
  (else (fuck-up))))

(define (describe-changing-relative-position unknown? psi u k1 k2)
 (case *language*
  ((english)
   (if (>= (psi-kappa-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
	    psi u k1 k2)
	   *kappa-threshold*)
       (let ((changing-relative-position
	      (psi-mean-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
	       psi u k1 k2)))
	(if (>= (abs changing-relative-position)
		*changing-relative-position-threshold*)
	    (format #f "~a is-moving~a ~a around ~a"
		    (role k1)
		    (describe-rate changing-relative-position
				   *slow-changing-relative-position-threshold*
				   *fast-changing-relative-position-threshold*)
		    (describe-angular-velocity changing-relative-position)
		    (role k2))
	    (format #f "~a is-not-moving-around ~a" (role k1) (role k2))))
       (if unknown?
	   (format #f "it-is-not-known-whether ~a is-moving-around ~a"
		   (role k1) (role k2))
	   #f)))
  ((japanese)
   (if (>= (psi-kappa-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
	    psi u k1 k2)
	   *kappa-threshold*)
       (let ((changing-relative-position
	      (psi-mean-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
	       psi u k1 k2)))
	(if (>= (abs changing-relative-position)
		*changing-relative-position-threshold*)
	    (format #f "~a ga ~a no-mawari-wo ~a~a mawatteimasu"
		    (role k1)
		    (role k2)
		    (describe-angular-velocity changing-relative-position)
		    (describe-rate
		     changing-relative-position
		     *slow-changing-relative-position-threshold*
		     *fast-changing-relative-position-threshold*))
	    (format #f "~a ga ~a no-mawari-wo-mawatteimasen"
		    (role k1) (role k2))))
       (if unknown?
	   (format #f "~a ga ~a no-mawari-wo-mawatteiru-kadouka-wakarimasen"
		   (role k1) (role k2))
	   #f)))
  (else (fuck-up))))

(define (interpret unknown? psi)
 (let ((description '())
       (features (psi-features psi)))
  (define (describe string)
   (when string
    (set! description (cons (string-append "  " string) description))))
  (for-each-n
   (lambda (u)
    (set! description
	  (cons (format #f "~a ~a"
			(case *language*
			 ((english) "state")
			 ((japanese) "jotai")
			 (else (fuck-up)))
			u)
		description))
    (for-each-n
     (lambda (k1)
      (when (and (memq 'magnitude-of-the-velocity-vector features)
		 (memq 'orientation-of-the-velocity-vector features))
       (describe (describe-motion unknown? psi u k1)))
      (when (memq
	     'finite-difference-of-the-magnitude-of-the-velocity-vector
	     features)
       (describe (describe-acceleration unknown? psi u k1)))
      ;; finite-difference-of-the-orientation-of-the-velocity-vector
      (when (memq 'angular-velocity features)
       (describe (describe-rotation unknown? psi u k1)))
      ;; finite-difference-of-the-angular-velocity
      (when (memq 'finite-difference-of-the-area features)
       (describe (describe-growth unknown? psi u k1)))
      (when (memq 'finite-difference-of-the-eccentricity features)
       (describe (describe-stretching unknown? psi u k1)))
      ;; second-finite-difference-of-the-area
      ;; second-finite-difference-of-the-eccentricity
      )
     (psi-kk psi))
    (for-each-n
     (lambda (k1)
      (for-each-n
       (lambda (k2)
	(when (and (memq 'distance-between-ellipse-centres features)
		   (memq 'orientation-of-the-vector-between-ellipse-centres
			 features))
	 (when (> k1 k2)
	  (describe (describe-relative-position unknown? psi u k1 k2))))
	(when (memq 'finite-difference-of-the-distance-between-ellipse-centres
		    features)
	 (when (> k1 k2)
	  (describe (describe-relative-motion unknown? psi u k1 k2))))
	(when (memq 'finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
		    features)
	 (when (> k1 k2)
	  (describe-changing-relative-position unknown? psi u k1 k2)))
	;; difference-between-the-orientations-of-the-major-axes
	;; mouthful
	;; finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
	;; finite-difference-of-mouthful
	)
       (psi-kk psi)))
     (psi-kk psi)))
   (vector-length (psi-b psi)))
  (reverse description)))

(define (english-event-name->language event-name)
 (case *language*
  ((english) event-name)
  ((japanese)
   (cond ((string=? event-name "pick-up") "hirou")
	 ((string=? event-name "put-down") "oku")
	 ((string=? event-name "push") "osu")
	 ((string=? event-name "pull") "hiku")
	 ((string=? event-name "tip") "katamukeru")
	 ((string=? event-name "tap") "tataku")
	 (else (fuck-up))))
  (else (fuck-up))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
