(MODULE HLIB-KINEMATIC)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-kinematic.sch")
(include "c-externals.sc")

;;; needs work: There is no notion of end-plugs on prismatic joints.
;;; needs work: coincident line segments, corner-to-corner, closed-loop

;;; Structures

(define-structure kinematic-model
 same-layer? joint-i? joint-j? joint-w? ground?)

(define-structure contact p l1 l2 i j)

;;; Parameters

(define *epsilon1* 7.0)
(define *epsilon2* 3.0)

;;; Variables

(define *kinematic-models-cache* '())
(define *colours* #f)
(define *cache?* #f)
(define *fast?* #f)
(define *cheap?* #f)
(define *twice?* #f)
(define *new-temporal-circumscription-cost?* #f)
(define *jair2001-bug?* #f)

;;; Procedures

(define (vector-remove v i) (list->vector (list-remove (vector->list v) i)))

(define (inexactify-line-segment l)
 (make-line-segment
  (map-vector exact->inexact (p l)) (map-vector exact->inexact (q l))))

(define (orthogonal v) (vector (- (y v)) (x v)))

(define (midpoint l)
 (vector (* 0.5 (+ (x (p l)) (x (q l)))) (* 0.5 (+ (y (p l)) (y (q l))))))

(define (swap-endpoints l) (make-line-segment (q l) (p l)))

(define (maybe-swap-endpoints l1 l2)
 ;; Swapping of endpoints fixed a bug in the version of Leonard used to
 ;; produce the JAIR2001 results.
 (if (and (not *jair2001-bug?*)
	  (< (distance-from-point-to-line (p l1) l2)
	     (distance-from-point-to-line (q l1) l2)))
     (swap-endpoints l1)
     l1))

(define (sin-atan y x) (/ y (sqrt (+ (sqr x) (sqr y)))))

(define (cos-atan y x) (/ x (sqrt (+ (sqr x) (sqr y)))))

(define (line-segment-orientation l) (orientation (p l) (q l)))

(define (sin-line-segment-orientation l)
 (sin-atan (if *jair2001-bug?*
	       ;; This is to replicate the behaviour of JAIR2001 which had a
	       ;; bug where it didn't correctly handle the fact that screen
	       ;; coordinates have Y increasing downward.
	       (- (y (q l)) (y (p l)))
	       ;; This intentionally is negated because screen coordinates have
	       ;; Y increasing downward.
	       (- (y (p l)) (y (q l))))
	   (- (x (q l)) (x (p l)))))

(define (cos-line-segment-orientation l)
 (cos-atan (if *jair2001-bug?*
	       ;; This is to replicate the behaviour of JAIR2001 which had a
	       ;; bug where it didn't correctly handle the fact that screen
	       ;; coordinates have Y increasing downward.
	       (- (y (q l)) (y (p l)))
	       ;; This intentionally is negated because screen coordinates have
	       ;; Y increasing downward.
	       (- (y (p l)) (y (q l))))
	   (- (x (q l)) (x (p l)))))

(define (rho r l)
 (if (zero? (x (v- (q l) (p l))))
     (/ (- (y r) (y (p l))) (- (y (q l)) (y (p l))))
     (/ (- (x r) (x (p l))) (- (x (q l)) (x (p l))))))

(define (closest-point-on-line-to-point r l)
 (point-of-intersection-of-lines
  l (make-line-segment r (v+ r (orthogonal (v- (q l) (p l)))))))

(define (distance-from-point-to-line r l)
 (distance r (closest-point-on-line-to-point r l)))

(define (point-touches-point? p q epsilon) (<= (distance p q) epsilon))

(define (point-touches-line? r l epsilon)
 (<= (distance-from-point-to-line r l) epsilon))

(define (point-touches-line-segment? r l epsilon)
 (and (point-touches-line? r l epsilon)
      (<= (- (min (x (p l)) (x (q l))) epsilon)
	  (x r)
	  (+ (max (x (p l)) (x (q l))) epsilon))
      (<= (- (min (y (p l)) (y (q l))) epsilon)
	  (y r)
	  (+ (max (y (p l)) (y (q l))) epsilon))))

(define (point-of-intersection-of-lines l1 l2)
 ;; This will fail if the lines do not intersect or if they are collinear.
 (m*v (invert-matrix
       (vector (vector (- (y (p l1)) (y (q l1))) (- (x (q l1)) (x (p l1))))
	       (vector (- (y (p l2)) (y (q l2))) (- (x (q l2)) (x (p l2))))))
      (vector (+ (* (y (p l1)) (- (x (q l1)) (x (p l1))))
		 (* (x (p l1)) (- (y (p l1)) (y (q l1)))))
	      (+ (* (y (p l2)) (- (x (q l2)) (x (p l2))))
		 (* (x (p l2)) (- (y (p l2)) (y (q l2))))))))

(define (point-of-intersection-of-line-segments l1 l2 epsilon)
 ;; This will fail if the line segments do not intersect or if they overlap.
 (cond ((point-touches-line-segment? (p l1) l2 epsilon) (p l1))
       ((point-touches-line-segment? (q l1) l2 epsilon) (q l1))
       ((point-touches-line-segment? (p l2) l1 epsilon) (p l2))
       ((point-touches-line-segment? (q l2) l1 epsilon) (q l2))
       (else (point-of-intersection-of-lines l1 l2))))

(define (lines-intersect? l1 l2)
 ;; This will fail if the lines are collinear.
 (not (zero? (dot (v- (q l1) (p l1)) (orthogonal (v- (q l2) (p l2)))))))

(define (line-segments-intersect? l1 l2 epsilon)
 (or (point-touches-line-segment? (p l1) l2 epsilon)
     (point-touches-line-segment? (q l1) l2 epsilon)
     (point-touches-line-segment? (p l2) l1 epsilon)
     (point-touches-line-segment? (q l2) l1 epsilon)
     (and (lines-intersect? l1 l2)
	  (let ((r (point-of-intersection-of-lines l1 l2)))
	   (and (point-touches-line-segment? r l1 epsilon)
		(point-touches-line-segment? r l2 epsilon))))))

(define (line-segments-cross? l1 l2 epsilon)
 (and (lines-intersect? l1 l2)
      (let ((r (point-of-intersection-of-lines l1 l2)))
       (and (point-touches-line-segment? r l1 *epsilon2*)
	    (point-touches-line-segment? r l2 *epsilon2*)))
      (not (point-touches-line-segment? (p l1) l2 epsilon))
      (not (point-touches-line-segment? (q l1) l2 epsilon))
      (not (point-touches-line-segment? (p l2) l1 epsilon))
      (not (point-touches-line-segment? (q l2) l1 epsilon))))

(define (line-segments-touch? l1 l2 epsilon)
 (or (point-touches-line-segment? (p l1) l2 epsilon)
     (point-touches-line-segment? (q l1) l2 epsilon)
     (point-touches-line-segment? (p l2) l1 epsilon)
     (point-touches-line-segment? (q l2) l1 epsilon)))

(define (endpoint-touches-at-non-corner? r1 r2 l1 l2 epsilon)
 (let ((epsilon1 0.0))			;tolerance
  (and (point-touches-line-segment? (r1 l1) l2 epsilon)
       (implies (point-touches-point? (r1 l1) (p l2) epsilon)
		(> (dot (v- (r2 l1) (r1 l1)) (v- (q l2) (p l2))) epsilon1))
       (implies (point-touches-point? (r1 l1) (q l2) epsilon)
		(> (dot (v- (r2 l1) (r1 l1)) (v- (p l2) (q l2))) epsilon1)))))

(define (polygon-endpoint-touches-at-non-corner? p1 p1a p1b l2 epsilon)
 (let ((epsilon1 0.0))			;tolerance
  (and (point-touches-line-segment? p1 l2 epsilon)
       (implies (point-touches-point? p1 (p l2) epsilon)
		(or (> (dot (v- p1a p1) (v- (q l2) (p l2))) epsilon1)
		    (> (dot (v- p1b p1) (v- (q l2) (p l2))) epsilon1)))
       (implies (point-touches-point? p1 (q l2) epsilon)
		(or (> (dot (v- p1a p1) (v- (p l2) (q l2))) epsilon1)
		    (> (dot (v- p1b p1) (v- (p l2) (q l2))) epsilon1))))))

(define (line-segments-touch-at-non-corner? l1 l2 epsilon)
 (or (endpoint-touches-at-non-corner? p q l1 l2 epsilon)
     (endpoint-touches-at-non-corner? q p l1 l2 epsilon)
     (endpoint-touches-at-non-corner? p q l2 l1 epsilon)
     (endpoint-touches-at-non-corner? q p l2 l1 epsilon)))

(define (square-matrix? m) (= (matrix-rows m) (matrix-columns m)))

(define (upper-triangular-transitive? m)
 (unless (square-matrix? m) (panic "Matrix must be square"))
 (every-n (lambda (k)
	   (every-n (lambda (j)
		     (every-n (lambda (i)
			       (and (or (not (matrix-ref m i j))
					(not (matrix-ref m j k))
					(matrix-ref m i k))
				    (or (not (matrix-ref m i j))
					(not (matrix-ref m i k))
					(matrix-ref m j k))
				    (or (not (matrix-ref m i k))
					(not (matrix-ref m j k))
					(matrix-ref m i j))))
			      j))
		    k))
	  (matrix-rows m)))

(define (transitive-closure m)
 (unless (square-matrix? m) (panic "Matrix must be square"))
 (let ((m (map-vector (lambda (row) (map-vector identity row)) m))
       (n (matrix-rows m)))
  (let loop ()
   (let ((again? #f))
    (for-each-n
     (lambda (i)
      (for-each-n
       (lambda (j)
	(when (matrix-ref m i j)
	 (for-each-n (lambda (k)
		      (when (and (matrix-ref m j k) (not (matrix-ref m i k)))
		       (matrix-set! m i k #t)
		       (set! again? #t)))
		     n)))
       n))
     n)
    (when again? (loop))))
  m))

(define (reflexive-transitive-closure m)
 (unless (square-matrix? m) (panic "Matrix must be square"))
 (let ((m (map-vector (lambda (row) (map-vector identity row)) m))
       (n (matrix-rows m)))
  (let loop ()
   (let ((again? #f))
    (for-each-n
     (lambda (i)
      (for-each-n
       (lambda (j)
	(when (matrix-ref m i j)
	 (for-each-n (lambda (k)
		      (when (and (matrix-ref m j k) (not (matrix-ref m i k)))
		       (matrix-set! m i k #t)
		       (set! again? #t)))
		     n)))
       n))
     n)
    (when again? (loop))))
  (for-each-n (lambda (i) (matrix-set! m i i #t)) n)
  m))

(define (vector-append . &rest)
 (list->vector (apply append (map vector->list &rest))))

(define (matrix-append . &rest) (apply map-vector vector-append &rest))

(define (d l)
 ;; This is the derivative of {x(p(l)), y(p(l)), x(q(l)), y(q(l))} with
 ;; respect to {x(c(l)), y(c(l)), theta(l)}.
 (vector (vector 1.0 0.0 (* 0.5 (- (y (q l)) (y (p l)))))
	 (vector 0.0 1.0 (* 0.5 (- (x (p l)) (x (q l)))))
	 (vector 1.0 0.0 (* 0.5 (- (y (p l)) (y (q l)))))
	 (vector 0.0 1.0 (* 0.5 (- (x (q l)) (x (p l)))))))

(define (d-polygon l polygon)
 ;; This is the derivative of {x(p(l)), y(p(l)), x(q(l)), y(q(l))} with
 ;; respect to {x(c(p)), y(c(p)), theta(p)}.
 (let ((c (centroid polygon)))
  (vector (vector 1.0 0.0 (- (y c) (y (p l))))
	  (vector 0.0 1.0 (- (x (p l)) (x c)))
	  (vector 1.0 0.0 (- (y c) (y (q l))))
	  (vector 0.0 1.0 (- (x (q l)) (x c))))))

(define (d-polygon2 l polygon)
 ;; This is the derivative of {x(c(l)), y(c(l)), theta(l)} with respect to
 ;; {x(c(p)), y(c(p)), theta(p)}.
 (let ((c (centroid polygon)))
  (vector (vector 1.0 0.0 (- (y c) (* 0.5 (+ (y (p l)) (y (q l))))))
	  (vector 0.0 1.0 (- (* 0.5 (+ (x (p l)) (x (q l)))) (x c)))
	  (vector 0.0 0.0 1.0))))

(define (dlrhodalpha rho l)
 ;; This is the derivative of {x(l(rho)), y(l(rho))} with respect to
 ;; {x(c(l)), y(c(l)), theta(l)}.
 (m*
  ;; This is the derivative of {x(l(rho)), y(l(rho))} with respect to
  ;; {x(p(l)), y(p(l)), x(q(l)), y(q(l))}.
  (vector (vector (- 1.0 rho) 0.0 rho 0.0) (vector 0.0 (- 1.0 rho) 0.0 rho))
  ;; This is the derivative of {x(p(l)), y(p(l)), x(q(l)), y(q(l))} with
  ;; respect to {x(c(l)), y(c(l)), theta(l)}.
  (d l)))

(define (dlrhodalpha-polygon rho l p)
 ;; This is the derivative of {x(l(rho)), y(l(rho))} with respect to
 ;; {x(c(p)), y(c(p)), theta(p)}.
 (m*
  ;; This is the derivative of {x(l(rho)), y(l(rho))} with respect to
  ;; {x(p(l)), y(p(l)), x(q(l)), y(q(l))}.
  (vector (vector (- 1.0 rho) 0.0 rho 0.0) (vector 0.0 (- 1.0 rho) 0.0 rho))
  ;; This is the derivative of {x(p(l)), y(p(l)), x(q(l)), y(q(l))} with
  ;; respect to {x(c(p)), y(c(p)), theta(p)}.
  (d-polygon l p)))

(define (drhodalpha l1 l2)
 ;; This is the derivative of rho(I(l1,l2),l1) with respect to
 ;; {x(c(l1)), y(c(l1)), theta(l1), x(c(l2)), y(c(l2)), theta(l2)}
 (let* ((x1 (x (midpoint l1)))
	(y1 (y (midpoint l1)))
	(x2 (x (midpoint l2)))
	(y2 (y (midpoint l2)))
	(t1 (cos-line-segment-orientation l2))
	(t4 (sin-line-segment-orientation l1))
	(t6 (cos-line-segment-orientation l1))
	(t7 (sin-line-segment-orientation l2))
	(l1 (line-segment-length l1))
	(l2 (line-segment-length l2))
	(t2 (/ l1))
	(t3 (* t1 t2))
	(t9 (- (* t6 t7) (* t4 t1)))
	(t10 (/ t9))
	(t12 (* 0.5 l1 t4))
	(t14 (* (- y1 t12) l1))
	(t17 (* 0.5 l1 t6))
	(t19 (* (- x1 t17) l1))
	(t21 (- (* t14 t6) (* t19 t4)))
	(t22 (* t10 t21))
	(t23 (* t3 t22))
	(t24 (/ l2))
	(t25 (* t6 t24))
	(t29 (* (- y2 (* 0.5 l2 t7)) l2))
	(t34 (* (- x2 (* 0.5 l2 t1)) l2))
	(t36 (- (* t29 t1) (* t34 t7)))
	(t37 (* t10 t36))
	(t38 (* t25 t37))
	(t39 (+ t17 (- t23 (- t38 x1))))
	(t44 (* t7 t2))
	(t45 (* t44 t22))
	(t46 (* t4 t24))
	(t47 (* t46 t37))
	(t48 (+ t12 (- t45 (- t47 y1))))
	(t50 (* t10 t4))
	(t51 (* t48 t7 t50))
	(t54 (* t10 t6))
	(t55 (* t39 t1 t54))
	(t61 (* t9 t9))
	(t62 (/ t61))
	(t63 (* t62 t21))
	(t66 (- (+ (* t6 t1) (* t4 t7))))
	(t67 (* t63 t66))
	(t69 (* l1 l1))
	(t70 (* t6 t6))
	(t74 (* t4 t4))
	(t79 (- (* t10
		   (+ (* 0.5 t69 t70)
		      (* t14 t4)
		      (* 0.5 t69 t74)
		      (* t19 t6)))))
	(t81 (* t62 t36))
	(t82 (* t81 t66))
	(t98 (- (* t63 t66)))
	(t100 (- (* t81 t66)))
	(t102 (* l2 l2))
	(t103 (* t1 t1))
	(t107 (* t7 t7))
	(t112 (- (* t10
		    (+ (* 0.5 t102 t103)
		       (* t29 t7)
		       (* 0.5 t102 t107)
		       (* t34 t1))))))
  (vector
   (* -2.0 (+ (* t39 (+ (* t1 t10 t4) 1.0)) t51))
   (* 2.0 (+ t55 (* t48 (- (* t7 t10 t6) 1.0))))
   (* 2.0
      (+ (* t39 (- (+ (* t3 t79) t47 (* t25 t82)) (+ (* t3 t67) t12)))
	 (* t48 (- (+ (* t44 t79) (* t46 t82) t17) (+ (* t44 t67) t38)))))
   (* 2.0 (+ (* t39 t7 t54) t51))
   (* -2.0 (+ t55 (* t48 t1 t50)))
   (* 2.0
      (+ (* t39 (- (* t25 t100) (+ t45 (* t3 t98) (* t25 t112))))
	 (* t48 (- (+ t23 (* t46 t100)) (+ (* t44 t98) (* t46 t112)))))))))

(define (drhodalpha-polygon l1 l2 p1 p2)
 ;; This is the derivative of rho(I(l1,l2),l1) with respect to
 ;; {x(c(p1)), y(c(p1)), theta(p1), x(c(p2)), y(c(p2)), theta(p2)}.
 (vector-ref
  (m*
   ;; This is the derivative of rho(I(l1,l2),l1) with respect to
   ;; {x(c(l1)), y(c(l1)), theta(l1), x(c(l2)), y(c(l2)), theta(l2)}
   (vector (drhodalpha l1 l2))
   ;; This is the derivative of
   ;; {x(c(l1)), y(c(l1)), theta(l1), x(c(l2)), y(c(l2)), theta(l2)} with
   ;; respect to
   ;; {x(c(p1)), y(c(p1)), theta(p1), x(c(p2)), y(c(p2)), theta(p2)}.
   (vector-append (matrix-append (d-polygon2 l1 p1) (make-matrix 3 3 0.0))
		  (matrix-append (make-matrix 3 3 0.0) (d-polygon2 l2 p2))))
  0))

(define (make-equation-for-dpe line-segments)
 (let ((v (make-vector (* 3 (vector-length line-segments)) 0.0)))
  (for-each-n
   (lambda (i)
    (let* ((l (inexactify-line-segment (vector-ref line-segments i)))
	   (line-segment-magnitude (* 0.5 (magnitude (v- (q l) (p l))))))
     (vector-set! v (+ (* 3 i) 1) line-segment-magnitude)))
   (vector-length line-segments))
  v))

(define (make-inequalities-for-line-segments-on-same-layer
	 line-segments same-layer?)
 (unless (square-matrix? same-layer?) (panic "Matrix must be square"))
 (unless (upper-triangular-transitive? same-layer?)
  (panic "Matrix must be transitive"))
 (let ((a '())
       (theta 20.0))			;tolerance
  (for-each-n
   (lambda (j)
    (for-each-n
     (lambda (i)
      (when (matrix-ref same-layer? i j)
       (when (line-segments-cross? (vector-ref line-segments i)
				   (vector-ref line-segments j)
				   *epsilon1*)
	(panic "Some line segments on same layer cross"))
       (let ((l1 (inexactify-line-segment (vector-ref line-segments i)))
	     (l2 (inexactify-line-segment (vector-ref line-segments j))))
	(when (endpoint-touches-at-non-corner? p q l1 l2 *epsilon1*)
	 (let ((v1 (make-vector (* 3 (vector-length line-segments)) 0.0))
	       (v2 (make-vector (* 3 (vector-length line-segments)) 0.0)))
	  (define (tilt! theta v)
	   (let* ((perp (unit (orthogonal (v- (q l2) (p l2)))))
		  ;; Sigma is a vector away from p(l1). The aaai2000 paper has
		  ;; a mistake. In the text, it defines it as towards but the
		  ;; equation has a sign error that defines it as away from.
		  ;; The code originally matched the text. It has been changed
		  ;; to match the equation. The text should read `away from'.
		  ;; The equation in the paper is OK except that the minus
		  ;; sign can be eliminated and q(l1)-p(l1) inverted to read
		  ;; p(l1)-q(l1). The paper has another mistake. The text says
		  ;; `a penetration will occur [...] is less than', which would
		  ;; indicate an inconsistency, but the inequality to enforce
		  ;; consistency says <=. When sigma is a vector away from
		  ;; p(l1), the text should read `is greater than'. The
		  ;; inequality is OK. As an aside, the paper uses the
		  ;; equation environment when it should use displaymath and
		  ;; the label eqSix should be eliminated.
		  (sigma (m*v (rotation-matrix-2d theta)
			      (k*v (dot perp (v- (p l1) (q l1))) perp)))
		  ;; The fact that the code for sigma originally matched the
		  ;; text was compensated for in the original code by the fact
		  ;; that a<=b was mistakingly coded as b-a<=0. It has been
		  ;; changed to a-b<=0.
		  (b (m*
		      (vector sigma)
		      (matrix-append
		       ;; This is the derivative of {x(p(l1)), y(p(l1))}
		       ;; with respect to {x(c(l1)), y(c(l1)), theta(l1)}.
		       (vector
			(vector 1.0 0.0 (* 0.5 (- (y (q l1)) (y (p l1)))))
			(vector 0.0 1.0 (* 0.5 (- (x (p l1)) (x (q l1))))))
		       (k*m -1.0 (dlrhodalpha (rho (p l1) l2) l2))))))
	    (vector-set! v (+ (* 3 i) 0) (matrix-ref b 0 0))
	    (vector-set! v (+ (* 3 i) 1) (matrix-ref b 0 1))
	    (vector-set! v (+ (* 3 i) 2) (matrix-ref b 0 2))
	    (vector-set! v (+ (* 3 j) 0) (matrix-ref b 0 3))
	    (vector-set! v (+ (* 3 j) 1) (matrix-ref b 0 4))
	    (vector-set! v (+ (* 3 j) 2) (matrix-ref b 0 5))))
	  (tilt! (degrees->radians (- theta)) v1)
	  (tilt! (degrees->radians theta) v2)
	  (set! a (cons v1 a))
	  (set! a (cons v2 a))))
	(when (endpoint-touches-at-non-corner? q p l1 l2 *epsilon1*)
	 (let ((v1 (make-vector (* 3 (vector-length line-segments)) 0.0))
	       (v2 (make-vector (* 3 (vector-length line-segments)) 0.0)))
	  (define (tilt! theta v)
	   (let* ((perp (unit (orthogonal (v- (q l2) (p l2)))))
		  ;; See the note above.
		  (sigma (m*v (rotation-matrix-2d theta)
			      (k*v (dot perp (v- (q l1) (p l1))) perp)))
		  (b (m*
		      (vector sigma)
		      (matrix-append
		       ;; This is the derivative of {x(q(l1)), y(q(l1))}
		       ;; with respect to {x(c(l1)), y(c(l1)), theta(l1)}.
		       (vector
			(vector 1.0 0.0 (* 0.5 (- (y (p l1)) (y (q l1)))))
			(vector 0.0 1.0 (* 0.5 (- (x (q l1)) (x (p l1))))))
		       (k*m -1.0 (dlrhodalpha (rho (q l1) l2) l2))))))
	    (vector-set! v (+ (* 3 i) 0) (matrix-ref b 0 0))
	    (vector-set! v (+ (* 3 i) 1) (matrix-ref b 0 1))
	    (vector-set! v (+ (* 3 i) 2) (matrix-ref b 0 2))
	    (vector-set! v (+ (* 3 j) 0) (matrix-ref b 0 3))
	    (vector-set! v (+ (* 3 j) 1) (matrix-ref b 0 4))
	    (vector-set! v (+ (* 3 j) 2) (matrix-ref b 0 5))))
	  (tilt! (degrees->radians (- theta)) v1)
	  (tilt! (degrees->radians theta) v2)
	  (set! a (cons v1 a))
	  (set! a (cons v2 a))))
	(when (endpoint-touches-at-non-corner? p q l2 l1 *epsilon1*)
	 (let ((v1 (make-vector (* 3 (vector-length line-segments)) 0.0))
	       (v2 (make-vector (* 3 (vector-length line-segments)) 0.0)))
	  (define (tilt! theta v)
	   (let* ((perp (unit (orthogonal (v- (q l1) (p l1)))))
		  ;; See the note above.
		  (sigma (m*v (rotation-matrix-2d theta)
			      (k*v (dot perp (v- (p l2) (q l2))) perp)))
		  (b (m*
		      (vector sigma)
		      (matrix-append
		       (k*m -1.0 (dlrhodalpha (rho (p l2) l1) l1))
		       ;; This is the derivative of {x(p(l2)), y(p(l2))}
		       ;; with respect to {x(c(l2)), y(c(l2)), theta(l2)}.
		       (vector
			(vector 1.0 0.0 (* 0.5 (- (y (q l2)) (y (p l2)))))
			(vector 0.0 1.0 (* 0.5 (- (x (p l2)) (x (q l2))))))))))
	    (vector-set! v (+ (* 3 i) 0) (matrix-ref b 0 0))
	    (vector-set! v (+ (* 3 i) 1) (matrix-ref b 0 1))
	    (vector-set! v (+ (* 3 i) 2) (matrix-ref b 0 2))
	    (vector-set! v (+ (* 3 j) 0) (matrix-ref b 0 3))
	    (vector-set! v (+ (* 3 j) 1) (matrix-ref b 0 4))
	    (vector-set! v (+ (* 3 j) 2) (matrix-ref b 0 5))))
	  (tilt! (degrees->radians (- theta)) v1)
	  (tilt! (degrees->radians theta) v2)
	  (set! a (cons v1 a))
	  (set! a (cons v2 a))))
	(when (endpoint-touches-at-non-corner? q p l2 l1 *epsilon1*)
	 (let ((v1 (make-vector (* 3 (vector-length line-segments)) 0.0))
	       (v2 (make-vector (* 3 (vector-length line-segments)) 0.0)))
	  (define (tilt! theta v)
	   (let* ((perp (unit (orthogonal (v- (q l1) (p l1)))))
		  ;; See the note above.
		  (sigma (m*v (rotation-matrix-2d theta)
			      (k*v (dot perp (v- (q l2) (p l2))) perp)))
		  (b (m*
		      (vector sigma)
		      (matrix-append
		       (k*m -1.0 (dlrhodalpha (rho (q l2) l1) l1))
		       ;; This is the derivative of {x(q(l2)), y(q(l2))} with
		       ;; respect to {x(c(l2)), y(c(l2)), theta(l2)}.
		       (vector
			(vector 1.0 0.0 (* 0.5 (- (y (p l2)) (y (q l2)))))
			(vector 0.0 1.0 (* 0.5 (- (x (q l2)) (x (p l2))))))))))
	    (vector-set! v (+ (* 3 i) 0) (matrix-ref b 0 0))
	    (vector-set! v (+ (* 3 i) 1) (matrix-ref b 0 1))
	    (vector-set! v (+ (* 3 i) 2) (matrix-ref b 0 2))
	    (vector-set! v (+ (* 3 j) 0) (matrix-ref b 0 3))
	    (vector-set! v (+ (* 3 j) 1) (matrix-ref b 0 4))
	    (vector-set! v (+ (* 3 j) 2) (matrix-ref b 0 5))))
	  (tilt! (degrees->radians (- theta)) v1)
	  (tilt! (degrees->radians theta) v2)
	  (set! a (cons v1 a))
	  (set! a (cons v2 a)))))))
     j))
   (matrix-rows same-layer?))
  a))

(define (make-equations-for-joints-with-i-rigid line-segments joint-i?)
 (unless (square-matrix? joint-i?) (panic "Matrix must be square"))
 (let ((a '()))
  (for-each-n
   (lambda (j)
    (for-each-n
     (lambda (i)
      (when (matrix-ref joint-i? i j)
       (unless (line-segments-intersect? (vector-ref line-segments i)
					 (vector-ref line-segments j)
					 *epsilon2*)
	(panic "Some joined line segments don't intersect"))
       (unless (lines-intersect? (vector-ref line-segments i)
				 (vector-ref line-segments j))
	(panic "Cannot handle joined collinear line segments"))
       (let* ((v (make-vector (* 3 (vector-length line-segments)) 0.0))
	      (l1 (inexactify-line-segment (vector-ref line-segments i)))
	      (l2 (inexactify-line-segment (vector-ref line-segments j)))
	      (b (drhodalpha (maybe-swap-endpoints l1 l2)
			     (maybe-swap-endpoints l2 l1))))
	(vector-set! v (+ (* 3 i) 0) (vector-ref b 0))
	(vector-set! v (+ (* 3 i) 1) (vector-ref b 1))
	(vector-set! v (+ (* 3 i) 2) (vector-ref b 2))
	(vector-set! v (+ (* 3 j) 0) (vector-ref b 3))
	(vector-set! v (+ (* 3 j) 1) (vector-ref b 4))
	(vector-set! v (+ (* 3 j) 2) (vector-ref b 5))
	(set! a (cons v a)))))
     j))
   (matrix-rows joint-i?))
  a))

(define (make-equations-for-joints-with-j-rigid line-segments joint-j?)
 (unless (square-matrix? joint-j?) (panic "Matrix must be square"))
 (let ((a '()))
  (for-each-n
   (lambda (j)
    (for-each-n
     (lambda (i)
      (when (matrix-ref joint-j? i j)
       (unless (line-segments-intersect? (vector-ref line-segments i)
					 (vector-ref line-segments j)
					 *epsilon2*)
	(panic "Some joined line segments don't intersect"))
       (unless (lines-intersect? (vector-ref line-segments i)
				 (vector-ref line-segments j))
	(panic "Cannot handle joined collinear line segments"))
       (let* ((v (make-vector (* 3 (vector-length line-segments)) 0.0))
	      (l1 (inexactify-line-segment (vector-ref line-segments i)))
	      (l2 (inexactify-line-segment (vector-ref line-segments j)))
	      (b (drhodalpha (maybe-swap-endpoints l2 l1)
			     (maybe-swap-endpoints l1 l2))))
	(vector-set! v (+ (* 3 j) 0) (vector-ref b 0))
	(vector-set! v (+ (* 3 j) 1) (vector-ref b 1))
	(vector-set! v (+ (* 3 j) 2) (vector-ref b 2))
	(vector-set! v (+ (* 3 i) 0) (vector-ref b 3))
	(vector-set! v (+ (* 3 i) 1) (vector-ref b 4))
	(vector-set! v (+ (* 3 i) 2) (vector-ref b 5))
	(set! a (cons v a)))))
     j))
   (matrix-rows joint-j?))
  a))

(define (make-equations-for-joints-with-w-rigid line-segments joint-w?)
 (unless (square-matrix? joint-w?) (panic "Matrix must be square"))
 (let ((a '()))
  (for-each-n
   (lambda (j)
    (for-each-n
     (lambda (i)
      (when (matrix-ref joint-w? i j)
       (unless (line-segments-intersect? (vector-ref line-segments i)
					 (vector-ref line-segments j)
					 *epsilon2*)
	(panic "Some joined line segments don't intersect"))
       (let ((v (make-vector (* 3 (vector-length line-segments)) 0.0)))
	(vector-set! v (+ (* 3 i) 2) 1.0)
	(vector-set! v (+ (* 3 j) 2) -1.0)
	(set! a (cons v a)))))
     j))
   (matrix-rows joint-w?))
  a))

(define (make-equations-for-ground line-segments ground?)
 (let ((a '()))
  (for-each-n
   (lambda (i)
    (when (vector-ref ground? i)
     (let ((v (make-vector (* 3 (vector-length line-segments)) 0.0)))
      (vector-set! v (+ (* 3 i) 0) 1.0)
      (set! a (cons v a)))
     (let ((v (make-vector (* 3 (vector-length line-segments)) 0.0)))
      (vector-set! v (+ (* 3 i) 1) 1.0)
      (set! a (cons v a)))
     (let ((v (make-vector (* 3 (vector-length line-segments)) 0.0)))
      (vector-set! v (+ (* 3 i) 2) 1.0)
      (set! a (cons v a)))))
   (vector-length ground?))
  a))

(define (patched-simplex a m1 m2 m3)
 (unless (and (>= m1 0)
	      (>= m2 0)
	      (>= m3 0)
	      (= (matrix-rows a) (+ m1 m2 m3 2)))
  (fuck-up))
 (let* ((m12 (+ m1 m2 1))
	(m (- (matrix-rows a) 2))
	(n (- (matrix-columns a) 1))
	(l1 (make-vector (+ n 1)))
	(l2 (make-vector m))
	(l3 (make-vector m2))
	(nl1 n)
	(iposv (make-vector m))
	(izrov (make-vector n))
	(ip 0)
	(kp #f)
	(bmax #f))
  (define (simp1 mm abs?)
   (set! kp (vector-ref l1 0))
   (set! bmax (matrix-ref a mm kp))
   (do ((k 1 (+ k 1))) ((>= k nl1))
    (when (positive?
	   (if abs?
	       (- (abs (matrix-ref a mm (vector-ref l1 k))) (abs bmax))
	       (- (matrix-ref a mm (vector-ref l1 k)) bmax)))
     (set! kp (vector-ref l1 k))
     (set! bmax (matrix-ref a mm (vector-ref l1 k))))))
  (define (simp2)
   (set! ip 0)
   (let ((q1 0.0)
	 (flag? #f))
    (for-each-n
     (lambda (i)
      (if flag?
	  (when (< (matrix-ref a (vector-ref l2 i) kp) (- *epsilon*))
	   (let ((q (/ (- (matrix-ref a (vector-ref l2 i) 0))
		       (matrix-ref a (vector-ref l2 i) kp))))
	    (cond
	     ((< q q1) (set! ip (vector-ref l2 i)) (set! q1 q))
	     ((= q q1)
	      (let ((qp 0.0)
		    (q0 0.0))
	       (let loop ((k 1))
		(when (<= k n)
		 (set! qp (/ (- (matrix-ref a ip k)) (matrix-ref a ip kp)))
		 (set! q0 (/ (- (matrix-ref a (vector-ref l2 i) k))
			     (matrix-ref a (vector-ref l2 i) kp)))
		 (when (= q0 qp) (loop (+ k 1)))))
	       (when (< q0 qp) (set! ip (vector-ref l2 i))))))))
	  (when (< (matrix-ref a (vector-ref l2 i) kp) (- *epsilon*))
	   (set! q1 (/ (- (matrix-ref a (vector-ref l2 i) 0))
		       (matrix-ref a (vector-ref l2 i) kp)))
	   (set! ip (vector-ref l2 i))
	   (set! flag? #t))))
     m)))
  (define (simp3 one?)
   (let ((piv (/ (matrix-ref a ip kp)))
	 (m (- (matrix-rows a) 2))
	 (n (- (matrix-columns a) 1)))
    (for-each-n
     (lambda (ii)
      (unless (= ii ip)
       (matrix-set! a ii kp (* piv (matrix-ref a ii kp)))
       (for-each-n (lambda (kk)
		    (unless (= kk kp)
		     (matrix-set! a ii kk (- (matrix-ref a ii kk)
					     (* (matrix-ref a ip kk)
						(matrix-ref a ii kp))))))
		   (+ n 1))))
     (+ m (if one? 2 1)))
    (for-each-n (lambda (kk)
		 (unless (= kk kp)
		  (matrix-set! a ip kk (* (- piv) (matrix-ref a ip kk)))))
		(+ n 1))
    (matrix-set! a ip kp piv)))
  (define (pass2)
   (simp1 0 #f)
   (cond ((> bmax *epsilon*)		;change
	  (simp2)
	  (cond ((zero? ip) #t)
		(else (simp3 #f)
		      (let ((t (vector-ref izrov (- kp 1))))
		       (vector-set! izrov (- kp 1) (vector-ref iposv (- ip 1)))
		       (vector-set! iposv (- ip 1) t))
		      (pass2))))
	 (else (list iposv izrov))))
  (for-each-n (lambda (k)
	       (vector-set! l1 k (+ k 1))
	       (vector-set! izrov k k))
	      n)
  (for-each-n (lambda (i)
	       (when (negative? (matrix-ref a (+ i 1) 0))
		(panic "Constants in SIMPLEX tableaux must be nonnegative"))
	       (vector-set! l2 i (+ i 1))
	       (vector-set! iposv i (+ n i)))
	      m)
  (for-each-n (lambda (i) (vector-set! l3 i #t)) m2)
  (cond
   ((positive? (+ m2 m3))
    (for-each-n
     (lambda (k)
      (do ((i (+ m1 1) (+ i 1)) (sum 0.0 (+ sum (matrix-ref a i k))))
	((> i m) (matrix-set! a (+ m 1) k (- sum)))))
     (+ n 1))
    (let loop ()
     (define (one)
      (simp3 #t)
      (cond
       ((>= (vector-ref iposv (- ip 1)) (+ n m12 -1))
	(let loop ((k 0))
	 (cond
	  ((and (< k nl1) (not (= kp (vector-ref l1 k)))) (loop (+ k 1)))
	  (else (set! nl1 (- nl1 1))
		(do ((is k (+ is 1))) ((>= is nl1))
		 (vector-set! l1 is (vector-ref l1 (+ is 1))))
		(matrix-set! a (+ m 1) kp (+ (matrix-ref a (+ m 1) kp) 1))
		(for-each-n
		 (lambda (i) (matrix-set! a i kp (- (matrix-ref a i kp))))
		 (+ m 2))))))
       ((and (>= (vector-ref iposv (- ip 1)) (+ n m1))
	     (vector-ref l3 (- (vector-ref iposv (- ip 1)) m1 n)))
	(vector-set! l3 (- (vector-ref iposv (- ip 1)) m1 n) #f)
	(matrix-set! a (+ m 1) kp (+ (matrix-ref a (+ m 1) kp) 1))
	(for-each-n (lambda (i) (matrix-set! a i kp (- (matrix-ref a i kp))))
		    (+ m 2))))
      (let ((t (vector-ref izrov (- kp 1))))
       (vector-set! izrov (- kp 1) (vector-ref iposv (- ip 1)))
       (vector-set! iposv (- ip 1) t))
      (loop))
     (simp1 (+ m 1) #f)
     (cond
      ((<= bmax *epsilon*)
       (cond ((< (matrix-ref a (+ m 1) 0) (- *epsilon*)) #f)
	     ((<= (matrix-ref a (+ m 1) 0) *epsilon*)
	      (let loop ((ip1 m12))
	       (cond
		((<= ip1 m)
		 (cond ((= (vector-ref iposv (- ip1 1)) (+ ip n -1))
			(simp1 ip1 #t)
			(cond ((> bmax *epsilon*) (set! ip ip1) (one)) ;change
			      (else (loop (+ ip1 1)))))
		       (else (loop (+ ip1 1)))))
		(else
		 (do ((i (+ m1 1) (+ i 1))) ((>= i m12))
		  (when (vector-ref l3 (- i m1 1))
		   (for-each-n
		    (lambda (k) (matrix-set! a i k (- (matrix-ref a i k))))
		    (+ n 1))))
		 (pass2)))))
	     (else (simp2) (if (zero? ip) #f (one)))))
      (else (simp2) (if (zero? ip) #f (one))))))
   (else (pass2)))))

(define (specialized-simplex a)
 (let* ((m (- (matrix-rows a) 1))
	(n (- (matrix-columns a) 1))
	(m1 (+ m 1))
	(n1 (+ n 1))
	(iposv (make-vector m))
	(izrov (make-vector n))
	(ip 0)
	(kp #f)
	(bmax #f))
  (for-each-n (lambda (k) (vector-set! izrov k k)) n)
  (for-each-n (lambda (i)
	       (when (negative? (matrix-ref a (+ i 1) 0))
		(panic "Constants in SIMPLEX tableaux must be nonnegative"))
	       (vector-set! iposv i (+ n i)))
	      m)
  (let loop ()
   (set! kp 1)
   (set! bmax (matrix-ref a 0 kp))
   (do ((k 2 (+ k 1))) ((>= k n1))
    (when (> (matrix-ref a 0 k) bmax)
     (set! kp k)
     (set! bmax (matrix-ref a 0 kp))))
   (cond
    ((> bmax *epsilon*)			;change
     (set! ip 0)
     (let ((q1 0.0)
	   (flag? #f))
      (for-each-n
       (lambda (i)
	(if flag?
	    (when (< (matrix-ref a (+ i 1) kp) (- *epsilon*))
	     (let ((q (/ (- (matrix-ref a (+ i 1) 0))
			 (matrix-ref a (+ i 1) kp))))
	      (cond
	       ((< q q1) (set! ip (+ i 1)) (set! q1 q))
	       ((= q q1)
		(let ((qp 0.0)
		      (q0 0.0))
		 (let loop ((k 1))
		  (when (<= k n)
		   (set! qp (/ (- (matrix-ref a ip k)) (matrix-ref a ip kp)))
		   (set! q0 (/ (- (matrix-ref a (+ i 1) k))
			       (matrix-ref a (+ i 1) kp)))
		   (when (= q0 qp) (loop (+ k 1)))))
		 (when (< q0 qp) (set! ip (+ i 1))))))))
	    (when (< (matrix-ref a (+ i 1) kp) (- *epsilon*))
	     (set! q1 (/ (- (matrix-ref a (+ i 1) 0))
			 (matrix-ref a (+ i 1) kp)))
	     (set! ip (+ i 1))
	     (set! flag? #t))))
       m))
     (cond ((zero? ip) #t)
	   (else (let ((piv (/ (matrix-ref a ip kp))))
		  (for-each-n
		   (lambda (ii)
		    (unless (= ii ip)
		     (matrix-set! a ii kp (* piv (matrix-ref a ii kp)))
		     (for-each-n
		      (lambda (kk)
		       (unless (= kk kp)
			(matrix-set! a ii kk (- (matrix-ref a ii kk)
						(* (matrix-ref a ip kk)
						   (matrix-ref a ii kp))))))
		      n1)))
		   m1)
		  (for-each-n
		   (lambda (kk)
		    (unless (= kk kp)
		     (matrix-set! a ip kk (* (- piv) (matrix-ref a ip kk)))))
		   n1)
		  (matrix-set! a ip kp piv))
		 (let ((t (vector-ref izrov (- kp 1))))
		  (vector-set! izrov (- kp 1) (vector-ref iposv (- ip 1)))
		  (vector-set! iposv (- ip 1) t))
		 (loop))))
    (else (list iposv izrov))))))

(define (presolve a b c)
 ;; Presolves the system:
 ;;   cx=1
 ;;   Ax<=0
 ;;   Bx=0
 (unless (and (or (zero? (matrix-rows a))
		  (= (matrix-columns a) (vector-length c)))
	      (or (zero? (matrix-rows b))
		  (= (matrix-columns b) (vector-length c))))
  (panic "The number of columns in A, the number of columns of B, and the length of C must all be equal"))
 (let ((a (map-vector (lambda (row) (map-vector identity row)) a))
       (b (map-vector (lambda (row) (map-vector identity row)) b))
       (c (map-vector identity c))
       (m1 (matrix-rows a))
       (m2 (matrix-rows b))
       (n (vector-length c))
       (r 0))
  (call-with-current-continuation
   (lambda (outer)
    (do ((i 0 (+ i 1))) ((or (>= i m2) (>= i n)))
     (call-with-current-continuation
      (lambda (inner)
       (do ((i1 i (+ i1 1))) ((>= i1 m2))
	(do ((j1 i (+ j1 1))) ((>= j1 n))
	 (unless (< (abs (matrix-ref b i1 j1)) 1e-6) ;tolerance
	  (unless (= i1 i)
	   (do ((j 0 (+ j 1))) ((>= j n))
	    (let ((t (matrix-ref b i j)))
	     (matrix-set! b i j (matrix-ref b i1 j))
	     (matrix-set! b i1 j t))))
	  (unless (= j1 i)
	   (let ((j i))
	    (do ((i 0 (+ i 1))) ((>= i m1))
	     (let ((t (matrix-ref a i j)))
	      (matrix-set! a i j (matrix-ref a i j1))
	      (matrix-set! a i j1 t)))
	    (do ((i 0 (+ i 1))) ((>= i m2))
	     (let ((t (matrix-ref b i j)))
	      (matrix-set! b i j (matrix-ref b i j1))
	      (matrix-set! b i j1 t)))
	    (let ((t (vector-ref c j)))
	     (vector-set! c j (vector-ref c j1))
	     (vector-set! c j1 t))))
	  (inner #f))))
       (outer #f)))
     (set! r (+ i 1))
     (let ((p (matrix-ref b i i)))
      (do ((j 0 (+ j 1))) ((>= j n))
       (matrix-set! b i j (/ (matrix-ref b i j) p))))
     (do ((i1 0 (+ i1 1))) ((>= i1 m1))
      (let ((p (matrix-ref a i1 i)))
       (do ((j 0 (+ j 1))) ((>= j n))
	(matrix-set!
	 a i1 j (- (matrix-ref a i1 j) (* p (matrix-ref b i j)))))))
     (do ((i1 0 (+ i1 1))) ((>= i1 m2))
      (unless (= i1 i)
       (let ((p (matrix-ref b i1 i)))
	(do ((j 0 (+ j 1))) ((>= j n))
	 (matrix-set!
	  b i1 j (- (matrix-ref b i1 j) (* p (matrix-ref b i j))))))))
     (let ((p (vector-ref c i)))
      (do ((j 0 (+ j 1))) ((>= j n))
       (vector-set! c j (- (vector-ref c j) (* p (matrix-ref b i j)))))))))
  (do ((i 0 (+ i 1))) ((>= i m1))
   (do ((j 0 (+ j 1))) ((>= j r))
    (unless (zero? (matrix-ref a i j))
     (panic "Some A entry is nonzero after presolve"))))
  (do ((i 0 (+ i 1))) ((>= i m2))
   (do ((j 0 (+ j 1))) ((>= j r))
    (unless (= (matrix-ref b i j) (if (= i j) 1.0 0.0))
     (panic "Some B entry is non-zero/one after presolve"))))
  (do ((j 0 (+ j 1))) ((>= j r))
   (unless (zero? (vector-ref c j))
    (panic "Some C entry is nonzero after presolve")))
  (list
   (map-vector (lambda (row) (subvector row r n)) a) '#() (subvector c r n))))

(define (numrec-consistent? a b c)
 ;; Determines whether:
 ;;   cx=1
 ;;   Ax<=0
 ;;   Bx=0
 ;; has a solution.
 (let* ((presolution (presolve a b c))
	(a (first presolution))
	(b (second presolution))
	(c (third presolution))
	(t (list->vector
	    (cons
	     (list->vector
	      (cons 0.0 (append (vector->list c) (map - (vector->list c)))))
	     (append
	      (map (lambda (row)
		    (list->vector (cons 0.0
					(append (map - (vector->list row))
						(vector->list row)))))
		   (vector->list a))
	      (map (lambda (row)
		    (list->vector (cons 0.0
					(append (map - (vector->list row))
						(vector->list row)))))
		   (vector->list b))
	      (list (make-vector (+ (* 2 (vector-length c)) 1) 0.0))))))
	(result (patched-simplex t (matrix-rows a) 0 (matrix-rows b))))
  (unless result (fuck-up))
  (unless (or (eq? result #t) (zero? (matrix-ref t 0 0))) (fuck-up))
  (eq? result #t)))

(define (maple-consistent? a b c)
 ;; Determines whether:
 ;;   cx=1
 ;;   Ax<=0
 ;;   Bx=0
 ;; has a solution.
 (let* ((presolution (presolve a b c))
	(a (first presolution))
	(b (second presolution))
	(c (third presolution)))
  (call-with-output-file "/tmp/QobiScheme.maple"
   (lambda (maple-port)
    (format maple-port "simplex[feasible]({")
    (for-each-vector
     (lambda (row)
      (for-each-n (lambda (j)
		   (let ((aij (vector-ref row j)))
		    (cond ((zero? aij) #f)
			  ((negative? aij) (format maple-port "~s*X~s" aij j))
			  (else (format maple-port "+~s*X~s" aij j)))))
		  (vector-length row))
      (format maple-port "<=0,~%"))
     a)
    (for-each-vector
     (lambda (row)
      (for-each-n (lambda (j)
		   (let ((bij (vector-ref row j)))
		    (cond ((zero? bij) #f)
			  ((negative? bij) (format maple-port "~s*X~s" bij j))
			  (else (format maple-port "+~s*X~s" bij j)))))
		  (vector-length row))
      (format maple-port "=0,~%"))
     b)
    (for-each-n (lambda (j)
		 (let ((cj (vector-ref c j)))
		  (cond ((zero? cj) #f)
			((negative? cj) (format maple-port "~s*X~s" cj j))
			(else (format maple-port "+~s*X~s" cj j)))))
		(vector-length c))
    (format maple-port "=1},UNRESTRICTED);~%"))))
 (system "maple /tmp/QobiScheme.maple >/tmp/QobiScheme.output")
 (let ((file (read-file "/tmp/QobiScheme.output")))
  (rm "/tmp/QobiScheme.maple")
  (rm "/tmp/QobiScheme.output")
  (unless (or (some (lambda (line)
		     (some (lambda (field) (string=? field "true"))
			   (fields line)))
		    file)
	      (some (lambda (line)
		     (some (lambda (field) (string=? field "false"))
			   (fields line)))
		    file))
   (fuck-up))
  (some (lambda (line)
	 (some (lambda (field) (string=? field "true")) (fields line)))
	file)))

(define (write-matlab-matrix a)
 (call-with-output-file "/tmp/foo.m"
  (lambda (matlab-port)
   (format matlab-port "function a=foo()~%")
   (format matlab-port "a=zeros(~a,~a);~%" (matrix-rows a) (matrix-columns a))
   (for-each-n (lambda (i)
		(for-each-n (lambda (j)
			     (unless (zero? (matrix-ref a i j))
			      (format matlab-port "a(~a,~a)=~a;~%"
				      (+ i 1) (+ j 1) (matrix-ref a i j))))
			    (matrix-columns a)))
	       (matrix-rows a)))))

(define (matlab-consistent? a b c)
 ;; Determines whether:
 ;;   cx=1
 ;;   Ax<=0
 ;;   Bx=0
 ;; has a solution.
 (let* ((presolution (presolve a b c))
	(a (first presolution))
	(b (second presolution))
	(c (third presolution))
	(m1 (matrix-rows a))
	(m2 (matrix-rows b))
	(n (vector-length c)))
  (cond
   ((zero? (+ m1 m2)) (not (every-vector zero? c)))
   (else
    (call-with-output-file "/tmp/QobiScheme.matlab"
     (lambda (matlab-port)
      (format matlab-port "lp([")
      (for-each-n
       (lambda (j) (format matlab-port "~s, ...~%" (- (vector-ref c j))))
       (- n 1))
      (format matlab-port "~s], ...~%[" (- (vector-ref c (- n 1))))
      (unless (zero? m2)
       (for-each-n
	(lambda (i)
	 (for-each-n
	  (lambda (j) (format matlab-port "~s, ...~%" (matrix-ref b i j)))
	  (- n 1))
	 (format matlab-port "~s; ...~%" (matrix-ref b i (- n 1))))
	(- m2 1))
       (for-each-n
	(lambda (j)
	 (format matlab-port "~s, ...~%" (matrix-ref b (- m2 1) j)))
	(- n 1))
       (if (zero? m1)
	   (format matlab-port "~s" (matrix-ref b (- m2 1) (- n 1)))
	   (format matlab-port "~s; ...~%"
		   (matrix-ref b (- m2 1) (- n 1)))))
      (cond
       ((zero? m1) (format matlab-port "], ...~%["))
       (else
	(for-each-n
	 (lambda (i)
	  (for-each-n
	   (lambda (j) (format matlab-port "~s, ...~%" (matrix-ref a i j)))
	   (- n 1))
	  (format matlab-port "~s; ...~%" (matrix-ref a i (- n 1))))
	 (- m1 1))
	(for-each-n
	 (lambda (j)
	  (format matlab-port "~s, ...~%" (matrix-ref a (- m1 1) j)))
	 (- n 1))
	(format matlab-port "~s], ...~%[" (matrix-ref a (- m1 1) (- n 1)))))
      (for-each-n
       (lambda (i) (format matlab-port "0; ...~%")) (- (+ m1 m2) 1))
      (format matlab-port "0],[],[],[],~s)~%" m2)))
    (system
     "unset DISPLAY;matlab </tmp/QobiScheme.matlab >/tmp/QobiScheme.output")
    (let ((file (read-file "/tmp/QobiScheme.output")))
     (rm "/tmp/QobiScheme.matlab")
     (rm "/tmp/QobiScheme.output")
     (some
      (lambda (line)
       (some (lambda (field) (string=? field "unbounded")) (fields line)))
      file))))))

(define (lp-solve-consistent? a b c)
 ;; Determines whether:
 ;;   cx=1
 ;;   Ax<=0
 ;;   Bx=0
 ;; has a solution.
 (let* ((presolution (presolve a b c))
	(a (first presolution))
	(b (second presolution))
	(c (third presolution))
	(m1 (matrix-rows a))
	(m2 (matrix-rows b))
	(n (vector-length c)))
  (if (zero? (+ m1 m2))
      (not (every-vector zero? c))
      (let ((lp (make-lp (+ m1 m2) (* 2 n))))
       (do ((i 0 (+ i 1))) ((>= i m1))
	(set-constr-type! lp (+ i 1) 0)
	(set-rh! lp (+ i 1) 0.0)
	(do ((j 0 (+ j 1))) ((>= j n))
	 (set-mat! lp (+ i 1) (+ j 1) (matrix-ref a i j))
	 (set-mat! lp (+ i 1) (+ n j 1) (- (matrix-ref a i j)))))
       (do ((i 0 (+ i 1))) ((>= i m2))
	(set-constr-type! lp (+ m1 i 1) 1)
	(set-rh! lp (+ m1 i 1) 0.0)
	(do ((j 0 (+ j 1))) ((>= j n))
	 (set-mat! lp (+ m1 i 1) (+ j 1) (matrix-ref b i j))
	 (set-mat! lp (+ m1 i 1) (+ n j 1) (- (matrix-ref b i j)))))
       (do ((j 0 (+ j 1))) ((>= j n))
	(set-mat! lp 0 (+ j 1) (vector-ref c j))
	(set-mat! lp 0 (+ n j 1) (- (vector-ref c j))))
       (set-maxim! lp)
       (auto-scale lp)
       (let ((result (solve lp)))
	(when (= result 6)
	 (format #t "Inverting failed~%")
	 (write (list a b c))
	 (newline))
	(unless (or (= result 0) (= result 3) (= result 6)) (fuck-up))
	(delete-lp lp)
	(= result 3))))))

(define (check-consistent1? a b c)
 (let ((numrec? (numrec-consistent? a b c))
       (maple? (maple-consistent? a b c))
       (matlab? (matlab-consistent? a b c)))
  (cond ((and (eq? numrec? maple?) (eq? numrec? matlab?)) numrec?)
	((eq? numrec? maple?)
	 (format #t "Matlab disagrees ~s ~s ~s~%"
		 matlab?
		 (matrix-rows a)
		 (matrix-rows b))
	 numrec?)
	((eq? numrec? matlab?)
	 (format #t "Maple disagrees ~s ~s ~s~%"
		 maple?
		 (matrix-rows a)
		 (matrix-rows b))
	 numrec?)
	(else (format #t "Numerical Recipes disagrees ~s ~s ~s~%"
		      numrec?
		      (matrix-rows a)
		      (matrix-rows b))
	      matlab?))))

(define (check-consistent2? a b c)
 (let ((lp-solve? (lp-solve-consistent? a b c))
       (maple? (maple-consistent? a b c))
       (matlab? (matlab-consistent? a b c)))
  (cond ((and (eq? lp-solve? maple?) (eq? lp-solve? matlab?)) lp-solve?)
	((eq? lp-solve? maple?)
	 (format #t "Matlab disagrees ~s ~s ~s~%"
		 matlab?
		 (matrix-rows a)
		 (matrix-rows b))
	 lp-solve?)
	((eq? lp-solve? matlab?)
	 (format #t "Maple disagrees ~s ~s ~s~%"
		 maple?
		 (matrix-rows a)
		 (matrix-rows b))
	 lp-solve?)
	(else (format #t "lp_solve disagrees ~s ~s ~s~%"
		      lp-solve?
		      (matrix-rows a)
		      (matrix-rows b))
	      matlab?))))

(define (boolean-upper-triangular? a)
 (every-n (lambda (i)
	   (every-n (lambda (j) (or (> j i) (not (matrix-ref a i j))))
		    (matrix-columns a)))
	  (matrix-rows a)))

(define (stable? line-segments kinematic-model)
 ;; Only the upper triangle of SAME-LAYER?, JOINT-I?, JOINT-J?, and JOINT-W?
 ;; is used.
 (unless (and (boolean-upper-triangular?
	       (kinematic-model-same-layer? kinematic-model))
	      (boolean-upper-triangular?
	       (kinematic-model-joint-i? kinematic-model))
	      (boolean-upper-triangular?
	       (kinematic-model-joint-j? kinematic-model))
	      (boolean-upper-triangular?
	       (kinematic-model-joint-w? kinematic-model)))
  (panic "Some matrix is not boolean upper triangular"))
 (let ((line-segments (list->vector line-segments)))
  (not (lp-solve-consistent?
	(list->vector
	 (make-inequalities-for-line-segments-on-same-layer
	  line-segments (kinematic-model-same-layer? kinematic-model)))
	(list->vector
	 (append (make-equations-for-joints-with-i-rigid
		  line-segments (kinematic-model-joint-i? kinematic-model))
		 (make-equations-for-joints-with-j-rigid
		  line-segments (kinematic-model-joint-j? kinematic-model))
		 (make-equations-for-joints-with-w-rigid
		  line-segments (kinematic-model-joint-w? kinematic-model))
		 (make-equations-for-ground
		  line-segments (kinematic-model-ground? kinematic-model))))
	(make-equation-for-dpe line-segments)))))

(define (vector->matrix v)
 (let ((n (inexact->exact (round (sqrt (vector-length v))))))
  (unless (= (sqr n) (vector-length v))
   (panic "The length of the vector is not a perfect square"))
  (map-n-vector (lambda (i) (subvector v (* i n) (* (+ i 1) n))) n)))

(define (vector->upper-triangular-matrix v)
 (let ((n (quadratic1 0.5 -0.5 (- (vector-length v)))))
  (unless (integer? n)
   (panic "The length of the vector is inappropriate for a square upper-triangular matrix"))
  (let* ((n (inexact->exact (round n)))
	 (a (make-matrix n n #f))
	 (k 0))
   (do ((i 0 (+ i 1))) ((>= i n))
    (do ((j (+ i 1) (+ j 1))) ((>= j n))
     (matrix-set! a i j (vector-ref v k))
     (set! k (+ k 1))))
   a)))

(define (upper-triangular-matrix->vector m)
 (unless (square-matrix? m) (panic "Matrix must be square"))
 (let ((v
	(make-vector (quotient (* (matrix-rows m) (- (matrix-rows m) 1)) 2)))
       (i 0))
  (do ((j 0 (+ j 1))) ((>= j (matrix-rows m)))
   (do ((k (+ j 1) (+ k 1))) ((>= k (matrix-rows m)))
    (vector-set! v i (matrix-ref m j k))
    (set! i (+ i 1))))
  v))

(define (last-position-vectorq x v)
 (let loop ((i (- (vector-length v) 1)))
  (if (or (negative? i) (eq? x (vector-ref v i))) i (loop (- i 1)))))

(define (bit-clear v i)
 (map-n-vector (lambda (j) (and (not (= j i)) (vector-ref v j)))
	       (vector-length v)))

(define (bit-set v i)
 (map-n-vector (lambda (j) (or (= j i) (vector-ref v j))) (vector-length v)))

(define (k->i k n) (if (< k (- n 1)) 0 (+ (k->i (- k (- n 1)) (- n 1)) 1)))

(define (k->j k n)
 (if (< k (- n 1)) (+ k 1) (+ (k->j (- k (- n 1)) (- n 1)) 1)))

(define (a-daughter i p v?)
 (let loop ((i i) (v? v?))
  (let ((k (an-integer-between (+ i 1) (- (vector-length v?) 1))))
   (unless (vector-ref v? k) (fail))
   (let ((v? (bit-clear v? k))) (if (p v?) v? (loop k v?))))))

(define (a-mother i p v?)
 (let loop ((i i) (v? v?))
  (let ((k (an-integer-between (+ i 1) (- (vector-length v?) 1))))
   (when (vector-ref v? k) (fail))
   (let ((v? (bit-set v? k))) (if (p v?) v? (loop k v?))))))

(define (a-lexicographic-ground?-daughter ground?)
 (bit-clear ground?
	    (an-integer-between (+ (last-position-vectorq #f ground?) 1)
				(- (vector-length ground?) 1))))

(define (a-lexicographic-ground?-mother ground?)
 (bit-set ground?
	  (an-integer-between (+ (last-position-vectorq #t ground?) 1)
			      (- (vector-length ground?) 1))))

(define (a-ground?-daughter ground?)
 (let ((k (an-integer-between 0 (- (vector-length ground?) 1))))
  (unless (vector-ref ground? k) (fail))
  (bit-clear ground? k)))

(define (a-ground?-mother ground?)
 (let ((k (an-integer-between 0 (- (vector-length ground?) 1))))
  (when (vector-ref ground? k) (fail))
  (bit-set ground? k)))

(define (ground?-eve line-segments)
 (make-vector (length line-segments) #t))

(define (ground?-star-daughter line-segments)
 (make-vector (length line-segments) #f))

(define (a-lexicographic-joint?-daughter joint?)
 (bit-clear joint?
	    (an-integer-between (+ (last-position-vectorq #f joint?) 1)
				(- (vector-length joint?) 1))))

(define (a-lexicographic-joint?-mother line-segments joint?)
 (let* ((n (length line-segments))
	(k (an-integer-between (+ (last-position-vectorq #t joint?) 1)
			       (- (vector-length joint?) 1))))
  (unless (line-segments-intersect? (list-ref line-segments (k->i k n))
				    (list-ref line-segments (k->j k n))
				    *epsilon2*)
   (fail))
  (bit-set joint? k)))

(define (a-joint?-daughter joint?)
 (let ((k (an-integer-between 0 (- (vector-length joint?) 1))))
  (unless (vector-ref joint? k) (fail))
  (bit-clear joint? k)))

(define (a-joint?-mother line-segments joint?)
 (let* ((n (length line-segments))
	(k (an-integer-between 0 (- (vector-length joint?) 1))))
  (when (or (vector-ref joint? k)
	    (not (line-segments-intersect? (list-ref line-segments (k->i k n))
					   (list-ref line-segments (k->j k n))
					   *epsilon2*)))
   (fail))
  (bit-set joint? k)))

(define (joint?-eve line-segments)
 (let ((joint?
	(make-vector
	 (quotient (* (length line-segments) (- (length line-segments) 1)) 2)))
       (n (length line-segments))
       (k 0))
  (for-each (lambda (i li)
	     (for-each (lambda (j lj)
			(when (< i j)
			 (vector-set!
			  joint? k (line-segments-intersect? li lj *epsilon2*))
			 (set! k (+ k 1))))
		       (enumerate n)
		       line-segments))
	    (enumerate n)
	    line-segments)
  joint?))

(define (joint?-star-daughter line-segments)
 (make-vector
  (quotient (* (length line-segments) (- (length line-segments) 1)) 2) #f))

(define (same-layer?-consistent? line-segments same-layer?)
 (let ((same-layer? (vector->upper-triangular-matrix same-layer?)))
  (and (upper-triangular-transitive? same-layer?)
       (not (some-n (lambda (j)
		     (some-n (lambda (i)
			      (and (matrix-ref same-layer? i j)
				   (line-segments-cross?
				    (list-ref line-segments i)
				    (list-ref line-segments j)
				    *epsilon1*)))
			     j))
		    (length line-segments))))))

(define (a-lexicographic-same-layer?-daughter line-segments same-layer?)
 (a-daughter
  (last-position-vectorq #f same-layer?)
  (lambda (same-layer?) (same-layer?-consistent? line-segments same-layer?))
  same-layer?))

(define (a-lexicographic-same-layer?-mother line-segments same-layer?)
 (a-mother
  (last-position-vectorq #t same-layer?)
  (lambda (same-layer?) (same-layer?-consistent? line-segments same-layer?))
  same-layer?))

(define (a-same-layer?-daughter line-segments same-layer?)
 (a-daughter
  -1
  (lambda (same-layer?) (same-layer?-consistent? line-segments same-layer?))
  same-layer?))

(define (a-same-layer?-mother line-segments same-layer?)
 (a-mother
  -1
  (lambda (same-layer?) (same-layer?-consistent? line-segments same-layer?))
  same-layer?))

(define (a-same-layer?-eve line-segments)
 ;; note: In the case of polygons, we only consider same-layer relations
 ;;       between polygons that touch. This is an optimization since same-layer
 ;;       relations between polygons that don't touch are redundant. We haven't
 ;;       (yet) added this optimization to the line-segment case.
 (let ((same-layer?
	(make-vector
	 (quotient (* (length line-segments) (- (length line-segments) 1)) 2)
	 #t)))
  (if (same-layer?-consistent? line-segments same-layer?)
      same-layer?
      (a-same-layer?-daughter line-segments same-layer?))))

(define (same-layer?-star-daughter line-segments)
 (make-vector
  (quotient (* (length line-segments) (- (length line-segments) 1)) 2) #f))

(define (a-top-down-ground? line-segments)
 (let loop ((ground? (ground?-eve line-segments)))
  (let ((ground?s
	 (remove-if-not
	  (lambda (ground?)
	   (possibly? (stable? line-segments
			       (make-kinematic-model
				(vector->upper-triangular-matrix
				 (a-same-layer?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				ground?))))
	  (all-values (a-lexicographic-ground?-daughter ground?)))))
   (if (null? ground?s)
       (if (possibly? (stable? line-segments
			       (make-kinematic-model
				(vector->upper-triangular-matrix
				 (a-same-layer?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(a-ground?-daughter ground?))))
	   (fail)
	   ground?)
       (loop (a-member-of ground?s))))))

(define (a-bottom-up-ground? line-segments)
 (let loop ((ground? (ground?-star-daughter line-segments)))
  (if (possibly?
       (stable?
	line-segments
	(make-kinematic-model
	 (vector->upper-triangular-matrix (a-same-layer?-eve line-segments))
	 (vector->upper-triangular-matrix (joint?-eve line-segments))
	 (vector->upper-triangular-matrix (joint?-eve line-segments))
	 (vector->upper-triangular-matrix (joint?-eve line-segments))
	 ground?)))
      (if (possibly?
	   (stable?
	    line-segments
	    (make-kinematic-model
	     (vector->upper-triangular-matrix
	      (a-same-layer?-eve line-segments))
	     (vector->upper-triangular-matrix (joint?-eve line-segments))
	     (vector->upper-triangular-matrix (joint?-eve line-segments))
	     (vector->upper-triangular-matrix (joint?-eve line-segments))
	     (a-ground?-daughter ground?))))
	  (fail)
	  ground?)
      (loop (a-lexicographic-ground?-mother ground?)))))

(define (a-top-down-joint-w? line-segments ground?)
 (let loop ((joint-w? (joint?-eve line-segments)))
  (let ((joint-w?s
	 (remove-if-not
	  (lambda (joint-w?)
	   (possibly? (stable? line-segments
			       (make-kinematic-model
				(vector->upper-triangular-matrix
				 (a-same-layer?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix joint-w?)
				ground?))))
	  (all-values (a-lexicographic-joint?-daughter joint-w?)))))
   (if (null? joint-w?s)
       (if (possibly? (stable? line-segments
			       (make-kinematic-model
				(vector->upper-triangular-matrix
				 (a-same-layer?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix
				 (a-joint?-daughter joint-w?))
				ground?)))
	   (fail)
	   (vector->upper-triangular-matrix joint-w?))
       (loop (a-member-of joint-w?s))))))

(define (a-bottom-up-joint-w? line-segments ground?)
 (let loop ((joint-w? (joint?-star-daughter line-segments)))
  (if (possibly?
       (stable?
	line-segments
	(make-kinematic-model
	 (vector->upper-triangular-matrix (a-same-layer?-eve line-segments))
	 (vector->upper-triangular-matrix (joint?-eve line-segments))
	 (vector->upper-triangular-matrix (joint?-eve line-segments))
	 (vector->upper-triangular-matrix joint-w?)
	 ground?)))
      (if (possibly?
	   (stable?
	    line-segments
	    (make-kinematic-model
	     (vector->upper-triangular-matrix
	      (a-same-layer?-eve line-segments))
	     (vector->upper-triangular-matrix (joint?-eve line-segments))
	     (vector->upper-triangular-matrix (joint?-eve line-segments))
	     (vector->upper-triangular-matrix (a-joint?-daughter joint-w?))
	     ground?)))
	  (fail)
	  (vector->upper-triangular-matrix joint-w?))
      (loop (a-lexicographic-joint?-mother line-segments joint-w?)))))

(define (a-top-down-joint-j? line-segments joint-w? ground?)
 (let loop ((joint-j? (joint?-eve line-segments)))
  (let ((joint-j?s
	 (remove-if-not
	  (lambda (joint-j?)
	   (possibly? (stable? line-segments
			       (make-kinematic-model
				(vector->upper-triangular-matrix
				 (a-same-layer?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix joint-j?)
				joint-w?
				ground?))))
	  (all-values (a-lexicographic-joint?-daughter joint-j?)))))
   (if (null? joint-j?s)
       (if (possibly? (stable? line-segments
			       (make-kinematic-model
				(vector->upper-triangular-matrix
				 (a-same-layer?-eve line-segments))
				(vector->upper-triangular-matrix
				 (joint?-eve line-segments))
				(vector->upper-triangular-matrix
				 (a-joint?-daughter joint-j?))
				joint-w?
				ground?)))
	   (fail)
	   (vector->upper-triangular-matrix joint-j?))
       (loop (a-member-of joint-j?s))))))

(define (a-bottom-up-joint-j? line-segments joint-w? ground?)
 (let loop ((joint-j? (joint?-star-daughter line-segments)))
  (if (possibly?
       (stable?
	line-segments
	(make-kinematic-model
	 (vector->upper-triangular-matrix (a-same-layer?-eve line-segments))
	 (vector->upper-triangular-matrix (joint?-eve line-segments))
	 (vector->upper-triangular-matrix joint-j?)
	 joint-w?
	 ground?)))
      (if (possibly?
	   (stable?
	    line-segments
	    (make-kinematic-model
	     (vector->upper-triangular-matrix
	      (a-same-layer?-eve line-segments))
	     (vector->upper-triangular-matrix (joint?-eve line-segments))
	     (vector->upper-triangular-matrix (a-joint?-daughter joint-j?))
	     joint-w?
	     ground?)))
	  (fail)
	  (vector->upper-triangular-matrix joint-j?))
      (loop (a-lexicographic-joint?-mother line-segments joint-j?)))))

(define (a-top-down-joint-i? line-segments joint-j? joint-w? ground?)
 (let loop ((joint-i? (joint?-eve line-segments)))
  (let ((joint-i?s
	 (remove-if-not
	  (lambda (joint-i?)
	   (stable? line-segments
		    (make-kinematic-model
		     (vector->upper-triangular-matrix
		      (a-same-layer?-eve line-segments))
		     (vector->upper-triangular-matrix joint-i?)
		     joint-j?
		     joint-w?
		     ground?)))
	  (all-values (a-lexicographic-joint?-daughter joint-i?)))))
   (if (null? joint-i?s)
       (if (possibly? (stable? line-segments
			       (make-kinematic-model
				(vector->upper-triangular-matrix
				 (a-same-layer?-eve line-segments))
				(vector->upper-triangular-matrix
				 (a-joint?-daughter joint-i?))
				joint-j?
				joint-w?
				ground?)))
	   (fail)
	   (vector->upper-triangular-matrix joint-i?))
       (loop (a-member-of joint-i?s))))))

(define (a-bottom-up-joint-i? line-segments joint-j? joint-w? ground?)
 (let loop ((joint-i? (joint?-star-daughter line-segments)))
  (if (possibly?
       (stable?
	line-segments
	(make-kinematic-model
	 (vector->upper-triangular-matrix (a-same-layer?-eve line-segments))
	 (vector->upper-triangular-matrix joint-i?)
	 joint-j?
	 joint-w?
	 ground?)))
      (if (possibly?
	   (stable?
	    line-segments
	    (make-kinematic-model
	     (vector->upper-triangular-matrix
	      (a-same-layer?-eve line-segments))
	     (vector->upper-triangular-matrix (a-joint?-daughter joint-i?))
	     joint-j?
	     joint-w?
	     ground?)))
	  (fail)
	  (vector->upper-triangular-matrix joint-i?))
      (loop (a-lexicographic-joint?-mother line-segments joint-i?)))))

(define (a-top-down-same-layer?
	 line-segments joint-i? joint-j? joint-w? ground?)
 (let loop ((same-layer? (a-same-layer?-eve line-segments)))
  (let ((same-layer?s
	 (remove-if-not
	  (lambda (same-layer?)
	   (stable? line-segments
		    (make-kinematic-model
		     (vector->upper-triangular-matrix same-layer?)
		     joint-i?
		     joint-j?
		     joint-w?
		     ground?)))
	  (all-values
	   (a-lexicographic-same-layer?-daughter line-segments same-layer?)))))
   (if (null? same-layer?s)
       (if (possibly? (stable? line-segments
			       (make-kinematic-model
				(vector->upper-triangular-matrix
				 (a-same-layer?-daughter
				  line-segments same-layer?))
				joint-i?
				joint-j?
				joint-w?
				ground?)))
	   (fail)
	   (vector->upper-triangular-matrix same-layer?))
       (loop (a-member-of same-layer?s))))))

(define (a-bottom-up-same-layer?
	 line-segments joint-i? joint-j? joint-w? ground?)
 (let loop ((same-layer? (same-layer?-star-daughter line-segments)))
  (if (stable?
       line-segments
       (make-kinematic-model (vector->upper-triangular-matrix same-layer?)
			     joint-i?
			     joint-j?
			     joint-w?
			     ground?))
      (if (possibly?
	   (stable? line-segments
		    (make-kinematic-model
		     (vector->upper-triangular-matrix
		      (a-same-layer?-daughter line-segments same-layer?))
		     joint-i?
		     joint-j?
		     joint-w?
		     ground?)))
	  (fail)
	  (vector->upper-triangular-matrix same-layer?))
      (loop (a-lexicographic-same-layer?-mother line-segments same-layer?)))))

(define (top-down-minimal-kinematic-models line-segments)
 (all-values
  (let* ((ground? (a-top-down-ground? line-segments))
	 (joint-w? (a-top-down-joint-w? line-segments ground?))
	 (joint-j? (a-top-down-joint-j? line-segments joint-w? ground?))
	 (joint-i?
	  (a-top-down-joint-i? line-segments joint-j? joint-w? ground?))
	 (same-layer? (a-top-down-same-layer?
		       line-segments joint-i? joint-j? joint-w? ground?)))
   (make-kinematic-model same-layer? joint-i? joint-j? joint-w? ground?))))

(define (bottom-up-minimal-kinematic-models line-segments)
 (all-values
  (let* ((ground? (a-bottom-up-ground? line-segments))
	 (joint-w? (a-bottom-up-joint-w? line-segments ground?))
	 (joint-j? (a-bottom-up-joint-j? line-segments joint-w? ground?))
	 (joint-i?
	  (a-bottom-up-joint-i? line-segments joint-j? joint-w? ground?))
	 (same-layer? (a-bottom-up-same-layer?
		       line-segments joint-i? joint-j? joint-w? ground?)))
   (make-kinematic-model same-layer? joint-i? joint-j? joint-w? ground?))))

(define (minimal-kinematic-models line-segments)
 (let ((top-down (top-down-minimal-kinematic-models line-segments))
       (bottom-up (bottom-up-minimal-kinematic-models line-segments)))
  (unless (set-equal? top-down bottom-up)
   (panic "Top-down and bottom-up differ"))
  bottom-up))

;;; Polygons

(define (polygon->line-segments polygon)
 (let ((ps (vector->list polygon)))
  (map make-line-segment ps (append (rest ps) (list (first ps))))))

(define (polygons->line-segments polygons)
 (reduce append (map polygon->line-segments polygons) '()))

(define (point-touches-polygon? point polygon epsilon)
 (some (lambda (l) (point-touches-line-segment? point l epsilon))
       (polygon->line-segments polygon)))

(define (contact-orthogonality c)
 (dot (v- (q (contact-l1 c)) (p (contact-l1 c)))
      (orthogonal (v- (q (contact-l2 c)) (p (contact-l2 c))))))

(define (contacts p1 p2 i j epsilon)
 (let ((contacts '()))
  (for-each-n
   (lambda (i1)
    (let ((l1 (make-line-segment
	       (vector-ref p1 i1)
	       (vector-ref p1 (modulo (+ i1 1) (vector-length p1))))))
     (for-each-n
      (lambda (j1)
       (let ((l2 (make-line-segment
		  (vector-ref p2 j1)
		  (vector-ref p2 (modulo (+ j1 1) (vector-length p2))))))
	(when (and (line-segments-intersect? l1 l2 epsilon)
		   ;; This is here because we cannot handle joined collinear
		   ;; line segments.
		   (lines-intersect? l1 l2))
	 (let ((r (point-of-intersection-of-line-segments l1 l2 epsilon)))
	  (set! contacts (cons (make-contact r l1 l2 i j) contacts))))))
      (vector-length p2))))
   (vector-length p1))
  ;; There can be two coincident contacts with vertex-edge contacts and four
  ;; coincident contacts with vertex-vertex contacts. These are redundant.
  ;; We used to pick one arbitrarily. That cause problems because sometimes
  ;; one of the contacts had l1 and l2 parallel or close to parallel. Now we
  ;; form the set of all coincident contacts and for each set choose the one
  ;; whose l1 and l2 are most orthogonal.
  (let ((contacts
	 (map
	  (lambda (contact-set)
	   (minp (lambda (c1 c2)
		  (> (contact-orthogonality c1) (contact-orthogonality c2)))
		 contact-set))
	  (transitive-equivalence-classesp
	   (lambda (c1 c2)
	    (and (= (x (contact-p c1)) (x (contact-p c2)))
		 (= (y (contact-p c1)) (y (contact-p c2)))
		 (= (contact-i c1) (contact-i c2))
		 (= (contact-j c1) (contact-j c2))))
	   contacts))))
   (if *fast?*
       (if (null? contacts) '() (list (first contacts)))
       contacts))))

(define (all-contacts polygons epsilon)
 (let ((the-contacts '()))
  (do ((p1s polygons (rest p1s)) (i 0 (+ i 1))) ((null? p1s))
   (let ((p1 (first p1s)))
    (for-each
     (lambda (p2 j)
      (set! the-contacts
	    (append (contacts p1 p2 i (+ i 1 j) epsilon) the-contacts)))
     (rest p1s)
     (enumerate (length (rest p1s))))))
  the-contacts))

(define (point-inside-polygon? point polygon epsilon)
 ;; note: POLYGON must be convex.
 (and (some (lambda (triangle)
	     (point-inside-triangle?
	      point (first triangle) (second triangle) (third triangle)))
	    (triangulate (vector->list polygon)))
      (not (point-touches-polygon? point polygon epsilon))))

(define (polygons-cross? polygon1 polygon2 epsilon)
 (let ((line-segments1 (polygon->line-segments polygon1))
       (line-segments2 (polygon->line-segments polygon2)))
  (or (some (lambda (line-segment1)
	     (some (lambda (line-segment2)
		    (line-segments-cross? line-segment1 line-segment2 epsilon))
		   line-segments2))
	    line-segments1)
      (some (lambda (line-segment1)
	     (point-inside-polygon? (midpoint line-segment1) polygon2 epsilon))
	    line-segments1)
      (some (lambda (line-segment2)
	     (point-inside-polygon? (midpoint line-segment2) polygon1 epsilon))
	    line-segments2)
      (and (every-vector
	    (lambda (point1) (point-touches-polygon? point1 polygon2 epsilon))
	    polygon1)
	   (every-vector
	    (lambda (point2) (point-touches-polygon? point2 polygon1 epsilon))
	    polygon2)))))

(define (polygons-touch? polygon1 polygon2 epsilon)
 (let ((line-segments1 (polygon->line-segments polygon1))
       (line-segments2 (polygon->line-segments polygon2)))
  (and
   (not (polygons-cross? polygon1 polygon2 epsilon))
   (some (lambda (line-segment1)
	  (some (lambda (line-segment2)
		 (line-segments-touch? line-segment1 line-segment2 epsilon))
		line-segments2))
	 line-segments1))))

(define (polygons-touch-at-non-corner? polygon1 polygon2 epsilon)
 (let ((line-segments1 (polygon->line-segments polygon1))
       (line-segments2 (polygon->line-segments polygon2)))
  (and
   (not (polygons-cross? polygon1 polygon2 epsilon))
   (or (some (lambda (l1)
	      (some-n (lambda (j)
		       (polygon-endpoint-touches-at-non-corner?
			(vector-ref polygon2 j)
			(vector-ref polygon2
				    (modulo (- j 1) (vector-length polygon2)))
			(vector-ref polygon2
				    (modulo (+ j 1) (vector-length polygon2)))
			l1
			epsilon))
		      (vector-length polygon2)))
	     line-segments1)
       (some (lambda (l2)
	      (some-n (lambda (i)
		       (polygon-endpoint-touches-at-non-corner?
			(vector-ref polygon1 i)
			(vector-ref polygon1
				    (modulo (- i 1) (vector-length polygon1)))
			(vector-ref polygon1
				    (modulo (+ i 1) (vector-length polygon1)))
			l2
			epsilon))
		      (vector-length polygon1)))
	     line-segments2)))))

(define (polygon-make-equation-for-dpe polygons)
 (let ((v (make-vector (* 3 (vector-length polygons)) 0.0)))
  (for-each-n
   (lambda (i)
    (vector-set!
     v
     (+ (* 3 i) 1)
     ;; note: POLYGON must be convex.
     (* 0.5 (area-of-polygon (vector->list (vector-ref polygons i))))))
   (vector-length polygons))
  v))

(define (make-inequalities-for-polygons-on-same-layer polygons same-layer?)
 (unless (square-matrix? same-layer?) (panic "Matrix must be square"))
 (unless (upper-triangular-transitive? same-layer?)
  (panic "Matrix must be transitive"))
 (let ((a '())
       (theta 20.0))			;tolerance
  (for-each-n
   (lambda (j)
    (let ((polygon2 (vector-ref polygons j)))
     (for-each-n
      (lambda (i)
       (let ((polygon1 (vector-ref polygons i)))
	(when (matrix-ref same-layer? i j)
	 (when (polygons-cross? polygon1 polygon2 *epsilon1*)
	  (panic "Some polygons on same layer cross"))
	 (for-each
	  (lambda (l1)
	   (for-each-n
	    (lambda (j1)
	     (let ((p2 (vector-ref polygon2 j1))
		   (l1 (inexactify-line-segment l1)))
	      (when (polygon-endpoint-touches-at-non-corner?
		     p2
		     (vector-ref polygon2
				 (modulo (- j1 1) (vector-length polygon2)))
		     (vector-ref polygon2
				 (modulo (+ j1 1) (vector-length polygon2)))
		     l1
		     *epsilon1*)
	       (let ((v1 (make-vector (* 3 (vector-length polygons)) 0.0))
		     (v2 (make-vector (* 3 (vector-length polygons)) 0.0)))
		(define (tilt! theta v)
		 (let* ((perp (orthogonal (v- (q l1) (p l1))))
			(sigma
			 (unit
			  (m*v (rotation-matrix-2d theta)
			       (k*v (dot perp (v- (centroid polygon1) p2))
				    perp))))
			(b (m* (vector sigma)
			       (matrix-append
				(k*m -1.0
				     (dlrhodalpha-polygon
				      (rho p2 l1) l1 polygon1))
				;; This is the derivative of {x(p2), y(p2)}
				;; with respect to
				;; {x(c(p2)), y(c(p2)), theta(p2)}.
				(vector
				 (vector
				  1.0
				  0.0
				  (- (y (centroid polygon2)) (y p2)))
				 (vector
				  0.0
				  1.0
				  (- (x p2) (x (centroid polygon2)))))))))
		  (vector-set! v (+ (* 3 i) 0) (matrix-ref b 0 0))
		  (vector-set! v (+ (* 3 i) 1) (matrix-ref b 0 1))
		  (vector-set! v (+ (* 3 i) 2) (matrix-ref b 0 2))
		  (vector-set! v (+ (* 3 j) 0) (matrix-ref b 0 3))
		  (vector-set! v (+ (* 3 j) 1) (matrix-ref b 0 4))
		  (vector-set! v (+ (* 3 j) 2) (matrix-ref b 0 5))))
		(tilt! (degrees->radians (- theta)) v1)
		(tilt! (degrees->radians theta) v2)
		(set! a (cons v1 a))
		(set! a (cons v2 a))))))
	    (vector-length polygon2)))
	  (polygon->line-segments polygon1))
	 (for-each
	  (lambda (l2)
	   (for-each-n
	    (lambda (i1)
	     (let ((p1 (vector-ref polygon1 i1))
		   (l2 (inexactify-line-segment l2)))
	      (when (polygon-endpoint-touches-at-non-corner?
		     p1
		     (vector-ref polygon1
				 (modulo (- i1 1) (vector-length polygon1)))
		     (vector-ref polygon1
				 (modulo (+ i1 1) (vector-length polygon1)))
		     l2
		     *epsilon1*)
	       (let ((v1 (make-vector (* 3 (vector-length polygons)) 0.0))
		     (v2 (make-vector (* 3 (vector-length polygons)) 0.0)))
		(define (tilt! theta v)
		 (let* ((perp (orthogonal (v- (q l2) (p l2))))
			(sigma
			 (unit
			  (m*v (rotation-matrix-2d theta)
			       (k*v (dot perp (v- (centroid polygon2) p1))
				    perp))))
			(b (m* (vector sigma)
			       (matrix-append
				;; This is the derivative of {x(p1), y(p1)}
				;; with respect to
				;; {x(c(p1)), y(c(p1)), theta(p1)}.
				(vector
				 (vector
				  1.0
				  0.0
				  (- (y (centroid polygon1)) (y p1)))
				 (vector
				  0.0
				  1.0
				  (- (x p1) (x (centroid polygon1)))))
				(k*m -1.0
				     (dlrhodalpha-polygon
				      (rho p1 l2) l2 polygon2))))))
		  (vector-set! v (+ (* 3 i) 0) (matrix-ref b 0 0))
		  (vector-set! v (+ (* 3 i) 1) (matrix-ref b 0 1))
		  (vector-set! v (+ (* 3 i) 2) (matrix-ref b 0 2))
		  (vector-set! v (+ (* 3 j) 0) (matrix-ref b 0 3))
		  (vector-set! v (+ (* 3 j) 1) (matrix-ref b 0 4))
		  (vector-set! v (+ (* 3 j) 2) (matrix-ref b 0 5))))
		(tilt! (degrees->radians (- theta)) v1)
		(tilt! (degrees->radians theta) v2)
		(set! a (cons v1 a))
		(set! a (cons v2 a))))))
	    (vector-length polygon1)))
	  (polygon->line-segments polygon2)))))
      j)))
   (matrix-rows same-layer?))
  a))

(define (polygon-make-equations-for-joints-with-i-rigid
	 polygons contacts joint-i?)
 (let ((a '()))
  (for-each (lambda (contact k)
	     (when (vector-ref joint-i? k)
	      (let* ((l1 (contact-l1 contact))
		     (l2 (contact-l2 contact))
		     (i (contact-i contact))
		     (j (contact-j contact))
		     (p1 (vector-ref polygons i))
		     (p2 (vector-ref polygons j)))
	       (unless (line-segments-intersect? l1 l2 *epsilon2*)
		(panic "Some joined line segments don't intersect"))
	       (unless (lines-intersect? l1 l2)
		(panic "Cannot handle joined collinear line segments"))
	       (let* ((v (make-vector (* 3 (vector-length polygons)) 0.0))
		      (l1 (inexactify-line-segment l1))
		      (l2 (inexactify-line-segment l2))
		      (b (drhodalpha-polygon (maybe-swap-endpoints l1 l2)
					     (maybe-swap-endpoints l2 l1)
					     p1
					     p2)))
		(vector-set! v (+ (* 3 i) 0) (vector-ref b 0))
		(vector-set! v (+ (* 3 i) 1) (vector-ref b 1))
		(vector-set! v (+ (* 3 i) 2) (vector-ref b 2))
		(vector-set! v (+ (* 3 j) 0) (vector-ref b 3))
		(vector-set! v (+ (* 3 j) 1) (vector-ref b 4))
		(vector-set! v (+ (* 3 j) 2) (vector-ref b 5))
		(set! a (cons v a))))))
	    contacts
	    (enumerate (length contacts)))
  a))

(define (polygon-make-equations-for-joints-with-j-rigid
	 polygons contacts joint-j?)
 (let ((a '()))
  (for-each (lambda (contact k)
	     (when (vector-ref joint-j? k)
	      (let* ((l1 (contact-l1 contact))
		     (l2 (contact-l2 contact))
		     (i (contact-i contact))
		     (j (contact-j contact))
		     (p1 (vector-ref polygons i))
		     (p2 (vector-ref polygons j)))
	       (unless (line-segments-intersect? l1 l2 *epsilon2*)
		(panic "Some joined line segments don't intersect"))
	       (unless (lines-intersect? l1 l2)
		(panic "Cannot handle joined collinear line segments"))
	       (let* ((v (make-vector (* 3 (vector-length polygons)) 0.0))
		      (l1 (inexactify-line-segment l1))
		      (l2 (inexactify-line-segment l2))
		      (b (drhodalpha-polygon (maybe-swap-endpoints l2 l1)
					     (maybe-swap-endpoints l1 l2)
					     p2
					     p1)))
		(vector-set! v (+ (* 3 j) 0) (vector-ref b 0))
		(vector-set! v (+ (* 3 j) 1) (vector-ref b 1))
		(vector-set! v (+ (* 3 j) 2) (vector-ref b 2))
		(vector-set! v (+ (* 3 i) 0) (vector-ref b 3))
		(vector-set! v (+ (* 3 i) 1) (vector-ref b 4))
		(vector-set! v (+ (* 3 i) 2) (vector-ref b 5))
		(set! a (cons v a))))))
	    contacts
	    (enumerate (length contacts)))
  a))

(define (polygon-make-equations-for-joints-with-w-rigid
	 polygons contacts joint-w?)
 (let ((a '()))
  (for-each (lambda (contact k)
	     (when (vector-ref joint-w? k)
	      (let* ((l1 (contact-l1 contact))
		     (l2 (contact-l2 contact))
		     (i (contact-i contact))
		     (j (contact-j contact)))
	       (unless (line-segments-intersect? l1 l2 *epsilon2*)
		(panic "Some joined line segments don't intersect"))
	       (let ((v (make-vector (* 3 (vector-length polygons)) 0.0)))
		(vector-set! v (+ (* 3 i) 2) 1.0)
		(vector-set! v (+ (* 3 j) 2) -1.0)
		(set! a (cons v a))))))
	    contacts
	    (enumerate (length contacts)))
  a))

(define (polygon-make-equations-for-ground polygons ground?)
 (let ((a '()))
  (for-each-n (lambda (i)
	       (when (vector-ref ground? i)
		(let ((v (make-vector (* 3 (vector-length polygons)) 0.0)))
		 (vector-set! v (+ (* 3 i) 0) 1.0)
		 (set! a (cons v a)))
		(let ((v (make-vector (* 3 (vector-length polygons)) 0.0)))
		 (vector-set! v (+ (* 3 i) 1) 1.0)
		 (set! a (cons v a)))
		(let ((v (make-vector (* 3 (vector-length polygons)) 0.0)))
		 (vector-set! v (+ (* 3 i) 2) 1.0)
		 (set! a (cons v a)))))
	      (vector-length ground?))
  a))

(define (polygon-stable? polygons contacts kinematic-model)
 ;; Only the upper triangle of SAME-LAYER? is used.
 (unless (boolean-upper-triangular?
	  (kinematic-model-same-layer? kinematic-model))
  (panic "The same-layer matrix is not boolean upper triangular"))
 (when #f				;debugging
  (let ((n (length polygons))
	(same-layer? (kinematic-model-same-layer? kinematic-model))
	(joint-i? (kinematic-model-joint-i? kinematic-model))
	(joint-j? (kinematic-model-joint-j? kinematic-model))
	(joint-w? (kinematic-model-joint-w? kinematic-model)))
   ;; This checks the kludge constraint.
   (when (some-n
	  (lambda (i)
	   (some-n (lambda (j)
		    (and (< i j)
			 (matrix-ref same-layer? i j)
			 (some (lambda (contact k)
				(and (or (vector-ref joint-i? k)
					 (vector-ref joint-j? k)
					 (vector-ref joint-w? k))
				     (or (and (= (contact-i contact) i)
					      (= (contact-j contact) j))
					 (and (= (contact-i contact) j)
					      (= (contact-j contact) i)))))
			       contacts
			       (enumerate (length contacts)))))
		   n))
	  n)
    (fuck-up))
   ;; This checks the constraint that every polygon pair have at most one
   ;; joint.
   (when (some (lambda (ci i)
		(some (lambda (cj j)
		       (and (not (= i j))
			    (or (vector-ref joint-i? i)
				(vector-ref joint-j? i)
				(vector-ref joint-w? i))
			    (or (vector-ref joint-i? j)
				(vector-ref joint-j? j)
				(vector-ref joint-w? j))
			    (or (and (= (contact-i ci) (contact-i cj))
				     (= (contact-j ci) (contact-j cj)))
				(and (= (contact-i ci) (contact-j cj))
				     (= (contact-j ci) (contact-i cj))))))
		      contacts
		      (enumerate (length contacts))))
	       contacts
	       (enumerate (length contacts)))
    (fuck-up))))
 (let ((polygons (list->vector polygons)))
  (not (lp-solve-consistent?
	(list->vector (make-inequalities-for-polygons-on-same-layer
		       polygons
		       (kinematic-model-same-layer? kinematic-model)))
	(list->vector
	 (append (polygon-make-equations-for-joints-with-i-rigid
		  polygons contacts (kinematic-model-joint-i? kinematic-model))
		 (polygon-make-equations-for-joints-with-j-rigid
		  polygons contacts (kinematic-model-joint-j? kinematic-model))
		 (polygon-make-equations-for-joints-with-w-rigid
		  polygons contacts (kinematic-model-joint-w? kinematic-model))
		 (polygon-make-equations-for-ground
		  polygons (kinematic-model-ground? kinematic-model))))
	(polygon-make-equation-for-dpe polygons)))))

(define (a-lexicographic-polygon-joint?-daughter joint?)
 (bit-clear joint?
	    (an-integer-between (+ (last-position-vectorq #f joint?) 1)
				(- (vector-length joint?) 1))))

(define (a-lexicographic-polygon-joint?-mother joint?)
 (bit-set joint?
	  (an-integer-between (+ (last-position-vectorq #t joint?) 1)
			      (- (vector-length joint?) 1))))

(define (a-polygon-joint?-daughter joint?)
 (let ((k (an-integer-between 0 (- (vector-length joint?) 1))))
  (unless (vector-ref joint? k) (fail))
  (bit-clear joint? k)))

(define (a-polygon-joint?-mother joint?)
 (let ((k (an-integer-between 0 (- (vector-length joint?) 1))))
  (when (vector-ref joint? k) (fail))
  (bit-set joint? k)))

(define (polygon-joint?-eve contacts) (make-vector (length contacts) #t))

(define (polygon-joint?-star-daughter contacts)
 (make-vector (length contacts) #f))

(define (polygon-joined?-consistent? contacts joined?)
 ;; This enforces the constraint that every polygon pair have at most one
 ;; joint.
 (not (some (lambda (ci i)
	     (some (lambda (cj j)
		    (and (not (= i j))
			 (vector-ref joined? i)
			 (vector-ref joined? j)
			 (or (and (= (contact-i ci) (contact-i cj))
				  (= (contact-j ci) (contact-j cj)))
			     (and (= (contact-i ci) (contact-j cj))
				  (= (contact-j ci) (contact-i cj))))))
		   contacts
		   (enumerate (length contacts))))
	    contacts
	    (enumerate (length contacts)))))

(define (a-lexicographic-polygon-joined?-daughter contacts joined?)
 ;; CONTACTS is only needed to enforce the constraint that every polygon pair
 ;; have at most one joint.
 (a-daughter (last-position-vectorq #f joined?)
	     (lambda (joined?) (polygon-joined?-consistent? contacts joined?))
	     joined?))

(define (a-lexicographic-polygon-joined?-mother contacts joined?)
 ;; CONTACTS is only needed to enforce the constraint that every polygon pair
 ;; have at most one joint.
 (a-mother (last-position-vectorq #t joined?)
	   (lambda (joined?) (polygon-joined?-consistent? contacts joined?))
	   joined?))

(define (a-polygon-joined?-daughter contacts joined?)
 ;; CONTACTS is only needed to enforce the constraint that every polygon pair
 ;; have at most one joint.
 (a-daughter -1
	     (lambda (joined?) (polygon-joined?-consistent? contacts joined?))
	     joined?))

(define (a-polygon-joined?-mother contacts joined?)
 ;; CONTACTS is only needed to enforce the constraint that every polygon pair
 ;; have at most one joint.
 (a-mother -1
	   (lambda (joined?) (polygon-joined?-consistent? contacts joined?))
	   joined?))

(define (a-polygon-joined?-eve contacts)
 (let ((joined? (make-vector (length contacts) #t)))
  (if (polygon-joined?-consistent? contacts joined?)
      joined?
      (a-polygon-joined?-daughter contacts joined?))))

(define (polygon-joined?-star-daughter contacts)
 (make-vector (length contacts) #f))

(define (a-lexicographic-polygon-rigid?-daughter rigid?)
 (bit-clear rigid?
	    (an-integer-between (+ (last-position-vectorq #f rigid?) 1)
				(- (vector-length rigid?) 1))))

(define (a-lexicographic-polygon-rigid?-mother rigid?)
 (bit-set rigid?
	  (an-integer-between (+ (last-position-vectorq #t rigid?) 1)
			      (- (vector-length rigid?) 1))))

(define (a-polygon-rigid?-daughter rigid?)
 (let ((k (an-integer-between 0 (- (vector-length rigid?) 1))))
  (unless (vector-ref rigid? k) (fail))
  (bit-clear rigid? k)))

(define (a-polygon-rigid?-mother rigid?)
 (let ((k (an-integer-between 0 (- (vector-length rigid?) 1))))
  (when (vector-ref rigid? k) (fail))
  (bit-set rigid? k)))

(define (polygon-rigid?-eve joined?)
 (make-vector (countq #t (vector->list joined?)) #t))

(define (polygon-rigid?-star-daughter joined?)
 (make-vector (countq #t (vector->list joined?)) #t))

(define (polygon-same-layer?-consistent?
	 polygons same-layer?
	 ;; needs work: CONTACTS, JOINT-I?, JOINT-J?, and JOINT-W? are only
	 ;;             needed for the kludge.
	 contacts joint-i? joint-j? joint-w?)
 (let ((same-layer? (vector->upper-triangular-matrix same-layer?)))
  (and
   (upper-triangular-transitive? same-layer?)
   (not
    (some-n
     (lambda (j)
      (some-n
       (lambda (i)
	(and
	 (matrix-ref same-layer? i j)
	 (or
	  ;; needs work: The following is a kludge. It disallows same-layer
	  ;;             assertions between attached polygons. We used to only
	  ;;             disallow same-layer assertions between rigidly
	  ;;             attached polygons. Now we also disallow same-layer
	  ;;             assertions between revolute and prismatic joints as
	  ;;             well.
	  (some (lambda (contact k)
		 (and (or (vector-ref joint-i? k)
			  (vector-ref joint-j? k)
			  (vector-ref joint-w? k))
		      (or (and (= (contact-i contact) i)
			       (= (contact-j contact) j))
			  (and (= (contact-i contact) j)
			       (= (contact-j contact) i)))))
		contacts
		(enumerate (length contacts)))
	  (polygons-cross?
	   (list-ref polygons i) (list-ref polygons j) *epsilon1*))))
       j))
     (length polygons))))))

(define (a-lexicographic-polygon-same-layer?-daughter
	 polygons same-layer?
	 ;; needs work: CONTACTS, JOINT-I?, JOINT-J?, and JOINT-W? are only
	 ;;             needed for the kludge.
	 contacts joint-i? joint-j? joint-w?)
 (a-daughter (last-position-vectorq #f same-layer?)
	     (lambda (same-layer?)
	      (polygon-same-layer?-consistent?
	       polygons same-layer? contacts joint-i? joint-j? joint-w?))
	     same-layer?))

(define (a-lexicographic-polygon-same-layer?-mother
	 polygons same-layer?
	 ;; needs work: CONTACTS, JOINT-I?, JOINT-J?, and JOINT-W? are only
	 ;;             needed for the kludge.
	 contacts joint-i? joint-j? joint-w?)
 (a-mother (last-position-vectorq #t same-layer?)
	   (lambda (same-layer?)
	    (polygon-same-layer?-consistent?
	     polygons same-layer? contacts joint-i? joint-j? joint-w?))
	   same-layer?))

(define (a-polygon-same-layer?-daughter
	 polygons same-layer?
	 ;; needs work: CONTACTS, JOINT-I?, JOINT-J?, and JOINT-W? are only
	 ;;             needed for the kludge.
	 contacts joint-i? joint-j? joint-w?)
 (a-daughter -1
	     (lambda (same-layer?)
	      (polygon-same-layer?-consistent?
	       polygons same-layer? contacts joint-i? joint-j? joint-w?))
	     same-layer?))

(define (a-polygon-same-layer?-mother
	 polygons same-layer?
	 ;; needs work: CONTACTS, JOINT-I?, JOINT-J?, and JOINT-W? are only
	 ;;             needed for the kludge.
	 contacts joint-i? joint-j? joint-w?)
 (a-mother -1
	   (lambda (same-layer?)
	    (polygon-same-layer?-consistent?
	     polygons same-layer? contacts joint-i? joint-j? joint-w?))
	   same-layer?))

(define (lucy? polygons)
 (upper-triangular-matrix->vector
  (transitive-closure
   (list->vector
    (map (lambda (polygon1)
	  (list->vector (map (lambda (polygon2)
			      (polygons-touch-at-non-corner?
			       polygon1 polygon2 *epsilon1*))
			     polygons)))
	 polygons)))))

(define (a-polygon-same-layer?-eve
	 ;; needs work: CONTACTS, JOINT-I?, JOINT-J?, and JOINT-W? are only
	 ;;             needed for the kludge.
	 polygons contacts lucy? joint-i? joint-j? joint-w?)
 ;; note: In the case of polygons, we only consider same-layer relations
 ;;       between polygons that touch. This is an optimization since same-layer
 ;;       relations between polygons that don't touch are redundant. We haven't
 ;;       (yet) added this optimization to the line-segment case.
 (let ((same-layer? lucy?))
  (if (polygon-same-layer?-consistent?
       polygons same-layer? contacts joint-i? joint-j? joint-w?)
      same-layer?
      (a-polygon-same-layer?-daughter
       polygons same-layer? contacts joint-i? joint-j? joint-w?))))

(define (polygon-same-layer?-star-daughter polygons)
 (make-vector (quotient (* (length polygons) (- (length polygons) 1)) 2) #f))

(define (old-a-top-down-polygon-ground? polygons contacts lucy?)
 ;; This is the version for use without the filter that removes
 ;; interpretations where a polygon pair has two joints.
 (let loop ((ground? (ground?-eve polygons)))
  (let ((ground?s
	 (remove-if-not
	  (lambda (ground?)
	   (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
			    (joint-j? (polygon-joint?-eve contacts))
			    (joint-w? (polygon-joint?-eve contacts)))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?)))))
	  (all-values (a-lexicographic-ground?-daughter ground?)))))
   (if (null? ground?s)
       (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
			    (joint-j? (polygon-joint?-eve contacts))
			    (joint-w? (polygon-joint?-eve contacts)))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 (a-ground?-daughter ground?)))))
	   (fail)
	   ground?)
       (loop (a-member-of ground?s))))))

(define (old-a-bottom-up-polygon-ground? polygons contacts lucy?)
 ;; This is the version for use without the filter that removes
 ;; interpretations where a polygon pair has two joints.
 (let loop ((ground? (ground?-star-daughter polygons)))
  (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
		       (joint-j? (polygon-joint?-eve contacts))
		       (joint-w? (polygon-joint?-eve contacts)))
		  (polygon-stable?
		   polygons
		   contacts
		   (make-kinematic-model
		    (vector->upper-triangular-matrix
		     (a-polygon-same-layer?-eve
		      polygons contacts lucy? joint-i? joint-j? joint-w?))
		    joint-i?
		    joint-j?
		    joint-w?
		    ground?))))
      (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
			   (joint-j? (polygon-joint?-eve contacts))
			   (joint-w? (polygon-joint?-eve contacts)))
		      (polygon-stable?
		       polygons
		       contacts
		       (make-kinematic-model
			(vector->upper-triangular-matrix
			 (a-polygon-same-layer?-eve
			  polygons contacts lucy? joint-i? joint-j? joint-w?))
			joint-i?
			joint-j?
			joint-w?
			(a-ground?-daughter ground?)))))
	  (fail)
	  ground?)
      (loop (a-lexicographic-ground?-mother ground?)))))

(define (a-top-down-polygon-ground? polygons contacts lucy?)
 (let loop ((ground? (ground?-eve polygons)))
  (let ((ground?s
	 (remove-if-not
	  (lambda (ground?)
	   (possibly? (let* ((joined? (a-polygon-joined?-eve contacts))
			     (joint-i? joined?)
			     (joint-j? joined?)
			     (joint-w? joined?))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?)))))
	  (all-values (a-lexicographic-ground?-daughter ground?)))))
   (if (null? ground?s)
       (if (possibly? (let* ((joined? (a-polygon-joined?-eve contacts))
			     (joint-i? joined?)
			     (joint-j? joined?)
			     (joint-w? joined?))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 (a-ground?-daughter ground?)))))
	   (fail)
	   ground?)
       (loop (a-member-of ground?s))))))

(define (a-bottom-up-polygon-ground? polygons contacts lucy?)
 (let loop ((ground? (ground?-star-daughter polygons)))
  (if (possibly? (let* ((joined? (a-polygon-joined?-eve contacts))
			(joint-i? joined?)
			(joint-j? joined?)
			(joint-w? joined?))
		  (polygon-stable?
		   polygons
		   contacts
		   (make-kinematic-model
		    (vector->upper-triangular-matrix
		     (a-polygon-same-layer?-eve
		      polygons contacts lucy? joint-i? joint-j? joint-w?))
		    joint-i?
		    joint-j?
		    joint-w?
		    ground?))))
      (if (possibly? (let* ((joined? (a-polygon-joined?-eve contacts))
			    (joint-i? joined?)
			    (joint-j? joined?)
			    (joint-w? joined?))
		      (polygon-stable?
		       polygons
		       contacts
		       (make-kinematic-model
			(vector->upper-triangular-matrix
			 (a-polygon-same-layer?-eve
			  polygons contacts lucy? joint-i? joint-j? joint-w?))
			joint-i?
			joint-j?
			joint-w?
			(a-ground?-daughter ground?)))))
	  (fail)
	  ground?)
      (loop (a-lexicographic-ground?-mother ground?)))))

(define (a-top-down-polygon-joint-w? polygons contacts lucy? ground?)
 (let loop ((joint-w? (polygon-joint?-eve contacts)))
  (let ((joint-w?s
	 (remove-if-not
	  (lambda (joint-w?)
	   (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
			    (joint-j? (polygon-joint?-eve contacts)))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?)))))
	  (all-values (a-lexicographic-polygon-joint?-daughter joint-w?)))))
   (if (null? joint-w?s)
       (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
			    (joint-j? (polygon-joint?-eve contacts))
			    (joint-w? (a-polygon-joint?-daughter joint-w?)))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?))))
	   (fail)
	   joint-w?)
       (loop (a-member-of joint-w?s))))))

(define (a-bottom-up-polygon-joint-w? polygons contacts lucy? ground?)
 (let loop ((joint-w? (polygon-joint?-star-daughter contacts)))
  (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
		       (joint-j? (polygon-joint?-eve contacts)))
		  (polygon-stable?
		   polygons
		   contacts
		   (make-kinematic-model
		    (vector->upper-triangular-matrix
		     (a-polygon-same-layer?-eve
		      polygons contacts lucy? joint-i? joint-j? joint-w?))
		    joint-i?
		    joint-j?
		    joint-w?
		    ground?))))
      (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
			   (joint-j? (polygon-joint?-eve contacts))
			   (joint-w? (a-polygon-joint?-daughter joint-w?)))
		      (polygon-stable?
		       polygons
		       contacts
		       (make-kinematic-model
			(vector->upper-triangular-matrix
			 (a-polygon-same-layer?-eve
			  polygons contacts lucy? joint-i? joint-j? joint-w?))
			joint-i?
			joint-j?
			joint-w?
			ground?))))
	  (fail)
	  joint-w?)
      (loop (a-lexicographic-polygon-joint?-mother joint-w?)))))

(define (a-top-down-polygon-joint-j? polygons contacts lucy? joint-w? ground?)
 (let loop ((joint-j? (polygon-joint?-eve contacts)))
  (let ((joint-j?s
	 (remove-if-not
	  (lambda (joint-j?)
	   (possibly? (let ((joint-i? (polygon-joint?-eve contacts)))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?)))))
	  (all-values (a-lexicographic-polygon-joint?-daughter joint-j?)))))
   (if (null? joint-j?s)
       (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
			    (joint-j? (a-polygon-joint?-daughter joint-j?)))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?))))
	   (fail)
	   joint-j?)
       (loop (a-member-of joint-j?s))))))

(define (a-bottom-up-polygon-joint-j? polygons contacts lucy? joint-w? ground?)
 (let loop ((joint-j? (polygon-joint?-star-daughter contacts)))
  (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts)))
		  (polygon-stable?
		   polygons
		   contacts
		   (make-kinematic-model
		    (vector->upper-triangular-matrix
		     (a-polygon-same-layer?-eve
		      polygons contacts lucy? joint-i? joint-j? joint-w?))
		    joint-i?
		    joint-j?
		    joint-w?
		    ground?))))
      (if (possibly? (let ((joint-i? (polygon-joint?-eve contacts))
			   (joint-j? (a-polygon-joint?-daughter joint-j?)))
		      (polygon-stable?
		       polygons
		       contacts
		       (make-kinematic-model
			(vector->upper-triangular-matrix
			 (a-polygon-same-layer?-eve
			  polygons contacts lucy?  joint-i? joint-j? joint-w?))
			joint-i?
			joint-j?
			joint-w?
			ground?))))
	  (fail)
	  joint-j?)
      (loop (a-lexicographic-polygon-joint?-mother joint-j?)))))

(define (a-top-down-polygon-joint-i?
	 polygons contacts lucy? joint-j? joint-w? ground?)
 (let loop ((joint-i? (polygon-joint?-eve contacts)))
  (let ((joint-i?s
	 (remove-if-not
	  (lambda (joint-i?)
	   (possibly? (polygon-stable?
		       polygons
		       contacts
		       (make-kinematic-model
			(vector->upper-triangular-matrix
			 (a-polygon-same-layer?-eve
			  polygons contacts lucy? joint-i? joint-j? joint-w?))
			joint-i?
			joint-j?
			joint-w?
			ground?))))
	  (all-values (a-lexicographic-polygon-joint?-daughter joint-i?)))))
   (if (null? joint-i?s)
       (if (possibly? (let ((joint-i? (a-polygon-joint?-daughter joint-i?)))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?))))
	   (fail)
	   joint-i?)
       (loop (a-member-of joint-i?s))))))

(define (a-bottom-up-polygon-joint-i?
	 polygons contacts lucy? joint-j? joint-w? ground?)
 (let loop ((joint-i? (polygon-joint?-star-daughter contacts)))
  (if (possibly? (polygon-stable?
		  polygons
		  contacts
		  (make-kinematic-model
		   (vector->upper-triangular-matrix
		    (a-polygon-same-layer?-eve
		     polygons contacts lucy? joint-i? joint-j? joint-w?))
		   joint-i?
		   joint-j?
		   joint-w?
		   ground?)))
      (if (possibly? (let ((joint-i? (a-polygon-joint?-daughter joint-i?)))
		      (polygon-stable?
		       polygons
		       contacts
		       (make-kinematic-model
			(vector->upper-triangular-matrix
			 (a-polygon-same-layer?-eve
			  polygons contacts lucy? joint-i? joint-j? joint-w?))
			joint-i?
			joint-j?
			joint-w?
			ground?))))
	  (fail)
	  joint-i?)
      (loop (a-lexicographic-polygon-joint?-mother joint-i?)))))

(define (a-top-down-polygon-same-layer?
	 polygons contacts lucy? joint-i? joint-j? joint-w? ground?)
 (let ((same-layer? (a-polygon-same-layer?-eve
		     polygons contacts lucy? joint-i? joint-j? joint-w?)))
  (unless (polygon-stable? polygons
			   contacts
			   (make-kinematic-model
			    (vector->upper-triangular-matrix same-layer?)
			    joint-i?
			    joint-j?
			    joint-w?
			    ground?))
   (fail))
  (let loop ((same-layer? same-layer?))
   (let ((same-layer?s
	  (remove-if-not
	   (lambda (same-layer?)
	    (polygon-stable? polygons
			     contacts
			     (make-kinematic-model
			      (vector->upper-triangular-matrix same-layer?)
			      joint-i?
			      joint-j?
			      joint-w?
			      ground?)))
	   ;; note: was lexicographic
	   (all-values
	    (a-polygon-same-layer?-daughter
	     polygons same-layer? contacts joint-i? joint-j? joint-w?)))))
    (if (null? same-layer?s)
	(if (possibly?
	     (polygon-stable?
	      polygons
	      contacts
	      (make-kinematic-model
	       (vector->upper-triangular-matrix
		(a-polygon-same-layer?-daughter
		 polygons same-layer? contacts joint-i? joint-j? joint-w?))
	       joint-i?
	       joint-j?
	       joint-w?
	       ground?)))
	    (fail)
	    (vector->upper-triangular-matrix same-layer?))
	(loop (a-member-of same-layer?s)))))))

(define (a-bottom-up-polygon-same-layer?
	 polygons contacts joint-i? joint-j? joint-w? ground?)
 (let loop ((same-layer? (polygon-same-layer?-star-daughter polygons)))
  (if (polygon-stable?
       polygons
       contacts
       (make-kinematic-model (vector->upper-triangular-matrix same-layer?)
			     joint-i?
			     joint-j?
			     joint-w?
			     ground?))
      (if (possibly?
	   (polygon-stable?
	    polygons
	    contacts
	    (make-kinematic-model
	     (vector->upper-triangular-matrix
	      (a-polygon-same-layer?-daughter
	       polygons same-layer? contacts joint-i? joint-j? joint-w?))
	     joint-i?
	     joint-j?
	     joint-w?
	     ground?)))
	  (fail)
	  (vector->upper-triangular-matrix same-layer?))
      (loop (a-lexicographic-polygon-same-layer?-mother
	     polygons same-layer? contacts joint-i? joint-j? joint-w?)))))

(define (top-down-minimal-polygon-kinematic-models polygons)
 (let ((contacts (all-contacts polygons *epsilon2*))
       (lucy? (lucy? polygons)))
  ;; note: This REMOVE-DUPLICATES is necesssary because the lexicographic
  ;;       daughter was changed to daughter in A-TOP-DOWN-POLYGON-SAME-LAYER?.
  (remove-duplicates
   (all-values
    (let* ((ground? (a-top-down-polygon-ground? polygons contacts lucy?))
	   (joint-w?
	    (a-top-down-polygon-joint-w? polygons contacts lucy? ground?))
	   (joint-j? (a-top-down-polygon-joint-j?
		      polygons contacts lucy? joint-w? ground?))
	   (joint-i? (a-top-down-polygon-joint-i?
		      polygons contacts lucy? joint-j? joint-w? ground?))
	   (same-layer?
	    (a-top-down-polygon-same-layer?
	     polygons contacts lucy? joint-i? joint-j? joint-w? ground?)))
     (make-kinematic-model same-layer? joint-i? joint-j? joint-w? ground?))))))

(define (bottom-up-minimal-polygon-kinematic-models polygons)
 (let ((contacts (all-contacts polygons *epsilon2*))
       (lucy? (lucy? polygons)))
  (all-values
   (let* ((ground? (a-bottom-up-polygon-ground? polygons contacts lucy?))
	  (joint-w?
	   (a-bottom-up-polygon-joint-w? polygons contacts lucy? ground?))
	  (joint-j? (a-bottom-up-polygon-joint-j?
		     polygons contacts lucy? joint-w? ground?))
	  (joint-i? (a-bottom-up-polygon-joint-i?
		     polygons contacts lucy? joint-j? joint-w? ground?))
	  (same-layer? (a-bottom-up-polygon-same-layer?
			polygons contacts joint-i? joint-j? joint-w? ground?)))
    (make-kinematic-model same-layer? joint-i? joint-j? joint-w? ground?)))))

(define (minimal-polygon-kinematic-models polygons)
 (let ((top-down (top-down-minimal-polygon-kinematic-models polygons))
       (bottom-up (bottom-up-minimal-polygon-kinematic-models polygons)))
  (unless (set-equal? top-down bottom-up)
   (panic "Top-down and bottom-up differ"))
  bottom-up))

(define (a-top-down-polygon-joined? polygons contacts lucy? ground?)
 (let loop ((joined? (a-polygon-joined?-eve contacts)))
  (let ((joined?s
	 (remove-if-not
	  (lambda (joined?)
	   (possibly? (let ((joint-i? joined?)
			    (joint-j? joined?)
			    (joint-w? joined?))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?)))))
	  (all-values
	   (a-lexicographic-polygon-joined?-daughter contacts joined?)))))
   (if (null? joined?s)
       (if (possibly? (let* ((joined?
			      (a-polygon-joined?-daughter contacts joined?))
			     (joint-i? joined?)
			     (joint-j? joined?)
			     (joint-w? joined?))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?))))
	   (fail)
	   joined?)
       (loop (a-member-of joined?s))))))

(define (a-bottom-up-polygon-joined? polygons contacts lucy? ground?)
 (let loop ((joined? (polygon-joined?-star-daughter contacts)))
  (if (possibly? (let ((joint-i? joined?)
		       (joint-j? joined?)
		       (joint-w? joined?))
		  (polygon-stable?
		   polygons
		   contacts
		   (make-kinematic-model
		    (vector->upper-triangular-matrix
		     (a-polygon-same-layer?-eve
		      polygons contacts lucy? joint-i? joint-j? joint-w?))
		    joint-i?
		    joint-j?
		    joint-w?
		    ground?))))
      (if (possibly? (let* ((joined?
			     (a-polygon-joined?-daughter contacts joined?))
			    (joint-i? joined?)
			    (joint-j? joined?)
			    (joint-w? joined?))
		      (polygon-stable?
		       polygons
		       contacts
		       (make-kinematic-model
			(vector->upper-triangular-matrix
			 (a-polygon-same-layer?-eve
			  polygons contacts lucy? joint-i? joint-j? joint-w?))
			joint-i?
			joint-j?
			joint-w?
			ground?))))
	  (fail)
	  joined?)
      (loop (a-lexicographic-polygon-joined?-mother contacts joined?)))))

(define (rigid?->joint-w? joined? rigid?)
 (let ((joint-w? (make-vector (vector-length joined?) #f)))
  (let loop ((i 0) (j 0))
   (cond
    ((>= i (vector-length joined?)) joint-w?)
    ((not (vector-ref joined? i)) (loop (+ i 1) j))
    ((vector-ref rigid? j) (vector-set! joint-w? i #t) (loop (+ i 1) (+ j 1)))
    (else (loop (+ i 1) (+ j 1)))))))

(define (a-top-down-polygon-rigid? polygons contacts lucy? ground? joined?)
 (let loop ((rigid? (polygon-rigid?-eve joined?)))
  (let ((rigid?s
	 (remove-if-not
	  (lambda (rigid?)
	   (possibly? (let ((joint-i? joined?)
			    (joint-j? joined?)
			    (joint-w? (rigid?->joint-w? joined? rigid?)))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?)))))
	  (all-values (a-lexicographic-polygon-rigid?-daughter rigid?)))))
   (if (null? rigid?s)
       (if (possibly? (let ((joint-i? joined?)
			    (joint-j? joined?)
			    (joint-w?
			     (rigid?->joint-w?
			      joined? (a-polygon-rigid?-daughter rigid?))))
		       (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (vector->upper-triangular-matrix
			  (a-polygon-same-layer?-eve
			   polygons contacts lucy? joint-i? joint-j? joint-w?))
			 joint-i?
			 joint-j?
			 joint-w?
			 ground?))))
	   (fail)
	   (rigid?->joint-w? joined? rigid?))
       (loop (a-member-of rigid?s))))))

(define (a-bottom-up-polygon-rigid? polygons contacts lucy? ground? joined?)
 (let loop ((rigid? (polygon-rigid?-star-daughter joined?)))
  (if (possibly? (let ((joint-i? joined?)
		       (joint-j? joined?)
		       (joint-w? (rigid?->joint-w? joined? rigid?)))
		  (polygon-stable?
		   polygons
		   contacts
		   (make-kinematic-model
		    (vector->upper-triangular-matrix
		     (a-polygon-same-layer?-eve
		      polygons contacts lucy? joint-i? joint-j? joint-w?))
		    joint-i?
		    joint-j?
		    joint-w?
		    ground?))))
      (if (possibly?
	   (let ((joint-i? joined?)
		 (joint-j? joined?)
		 (joint-w? (rigid?->joint-w?
			    joined? (a-polygon-rigid?-daughter rigid?))))
	    (polygon-stable?
	     polygons
	     contacts
	     (make-kinematic-model
	      (vector->upper-triangular-matrix
	       (a-polygon-same-layer?-eve
		polygons contacts lucy? joint-i? joint-j? joint-w?))
	      joint-i?
	      joint-j?
	      joint-w?
	      ground?))))
	  (fail)
	  (rigid?->joint-w? joined? rigid?))
      (loop (a-lexicographic-polygon-rigid?-mother rigid?)))))

(define (top-down-simplified-minimal-polygon-kinematic-models polygons)
 (let ((contacts (all-contacts polygons *epsilon2*))
       (lucy? (lucy? polygons)))
  ;; note: This REMOVE-DUPLICATES is necesssary because the lexicographic
  ;;       daughter was changed to daughter in A-TOP-DOWN-POLYGON-SAME-LAYER?.
  (remove-duplicates
   (all-values
    (let* ((ground? (a-top-down-polygon-ground? polygons contacts lucy?))
	   (joined?
	    (a-top-down-polygon-joined? polygons contacts lucy? ground?))
	   (joint-i? joined?)
	   ;; The call to MAP-VECTOR is so that JOINT-I? and JOINT-J? are not
	   ;; EQ? because this causes problems with the kgp.sc GUI.
	   (joint-j? (map-vector identity joined?))
	   (joint-w? (if *fast?*
			 (map-vector identity joined?)
			 (a-top-down-polygon-rigid?
			  polygons contacts lucy? ground? joined?)))
	   (same-layer?
	    (a-top-down-polygon-same-layer?
	     polygons contacts lucy? joint-i? joint-j? joint-w? ground?)))
     (make-kinematic-model same-layer? joint-i? joint-j? joint-w? ground?))))))

(define (bottom-up-simplified-minimal-polygon-kinematic-models polygons)
 (let ((contacts (all-contacts polygons *epsilon2*))
       (lucy? (lucy? polygons)))
  (all-values
   (let* ((ground? (a-bottom-up-polygon-ground? polygons contacts lucy?))
	  (joined?
	   (a-bottom-up-polygon-joined? polygons contacts lucy? ground?))
	  (joint-i? joined?)	
	  ;; The call to MAP-VECTOR is so that JOINT-I? and JOINT-J? are not
	  ;; EQ? because this causes problems with the kgp.sc GUI.
	  (joint-j? (map-vector identity joined?))
	  (joint-w? (if *fast?*
			(map-vector identity joined?)
			(a-bottom-up-polygon-rigid?
			 polygons contacts lucy? ground? joined?)))
	  (same-layer? (a-bottom-up-polygon-same-layer?
			polygons contacts joint-i? joint-j? joint-w? ground?)))
    (make-kinematic-model same-layer? joint-i? joint-j? joint-w? ground?)))))

(define (check-simplified-minimal-polygon-kinematic-models polygons)
 (let ((top-down
	(top-down-simplified-minimal-polygon-kinematic-models polygons))
       (bottom-up
	(bottom-up-simplified-minimal-polygon-kinematic-models polygons)))
  (unless (set-equal? top-down bottom-up)
   (panic "Top-down and bottom-up differ"))
  bottom-up))

(define (a-cheap-polygon-joined? polygons contacts lucy? joined? ground?)
 (let loop ((joined? joined?))
  (let ((joined?s
	 (remove-if-not
	  (lambda (joined?)
	   (possibly?
	    (let ((joint-i? joined?)
		  (joint-j? joined?)
		  (joint-w? joined?))
	     (polygon-stable?
	      polygons
	      contacts
	      (make-kinematic-model
	       (vector->upper-triangular-matrix
		(a-polygon-same-layer?-eve
		 polygons contacts lucy? joint-i? joint-j? joint-w?))
	       joint-i?
	       joint-j?
	       joint-w?
	       ground?)))))
	  (all-values
	   (a-lexicographic-polygon-joined?-daughter contacts joined?)))))
   (if (null? joined?s)
       (if (possibly?
	    (let* ((joined?
		    (a-polygon-joined?-daughter contacts joined?))
		   (joint-i? joined?)
		   (joint-j? joined?)
		   (joint-w? joined?))
	     (polygon-stable?
	      polygons
	      contacts
	      (make-kinematic-model
	       (vector->upper-triangular-matrix
		(a-polygon-same-layer?-eve
		 polygons contacts lucy? joint-i? joint-j? joint-w?))
	       joint-i?
	       joint-j?
	       joint-w?
	       ground?))))
	   (fail)
	   joined?)
       (loop (a-member-of joined?s))))))

(define (cheap-simplified-minimal-polygon-kinematic-models polygons)
 ;; This generates only a subset of all possible JOINED? relations. But since
 ;; we do not do temporal circumscription over JOINED?, there is no guarantee
 ;; of continuity in the ATTACHED? primitive event type. This also
 ;; intentionally generates only a subset of all possible SAME-LAYER?
 ;; relations. But since we do not do temporal circumscription over
 ;; SAME-LAYER?, there is not guarantee of continuity in the CONTACTS?
 ;; primitive event type.
 (define (contacts? i j contact)
  (or (and (= (contact-i contact) i) (= (contact-j contact) j))
      (and (= (contact-i contact) j) (= (contact-j contact) i))))
 (unless *fast?* (fuck-up))
 (let ((contacts (all-contacts polygons *epsilon2*))
       (lucy? (lucy? polygons)))
  (define (contact? i j)
   (some (lambda (contact) (contacts? i j contact)) contacts))
  (define (spanning-tree root nodes)
   (let loop ((in (list root)) (out (removev root nodes)) (result '()))
    (let ((next (remove-if-not
		 (lambda (o) (some (lambda (i) (contact? i o)) in))
		 out)))
     (if (null? next)
	 result
	 (loop (append in next)
	       (set-differencev out next)
	       (append
		(map (lambda (n)
		      (find-if (lambda (contact)
				(some (lambda (i) (contacts? i n contact)) in))
			       contacts))
		     next)
		result))))))
  (all-values
   (let* ((components (equivalence-classesp (lambda (i j) (contact? i j))
					    (enumerate (length polygons))))
	  (grounded-indices (map a-member-of components))
	  (ground?
	   (map-n-vector (lambda (i) (not (not (memv i grounded-indices))))
			 (length polygons)))
	  (joined-contacts
	   (reduce append
		   (map (lambda (grounded-index component)
			 (spanning-tree grounded-index component))
			grounded-indices
			components)
		   '()))
	  (joined? (list->vector
		    (map (lambda (contact)
			  (not (not (memq contact joined-contacts))))
			 contacts)))
	  (joined?
	   (a-cheap-polygon-joined? polygons contacts lucy? joined? ground?))
	  (joint-i? joined?)
	  ;; The call to MAP-VECTOR is so that JOINT-I? and JOINT-J? are not
	  ;; EQ? because this causes problems with the kgp.sc GUI.
	  (joint-j? (map-vector identity joined?))
	  (joint-w? (map-vector identity joined?))
	  (same-layer?
	   ;; note: We actually generate a single minimal stable SAME-LAYER?
	   ;;       relation. This is purely for aesthetic reasons. To produce
	   ;;       models that look good. Because we don't actually use the
	   ;;       CONTACT primitive event type in any compound event type
	   ;;       definition the SAME-LAYER? relation is never used. We could
	   ;;       just put an EVE here, which is what we used to do, to save
	   ;;       compute time.
	   (one-value
	    (a-top-down-polygon-same-layer?
	     polygons contacts lucy? joint-i? joint-j? joint-w? ground?))))
    (make-kinematic-model same-layer? joint-i? joint-j? joint-w? ground?)))))

(define (simplified-minimal-polygon-kinematic-models polygons)
 (if *cheap?*
     (cheap-simplified-minimal-polygon-kinematic-models polygons)
     (top-down-simplified-minimal-polygon-kinematic-models polygons)))

(define (lexicographically< x y)
 (and (not (null? x))
      (or (< (first x) (first y))
	  (and (= (first x) (first y))
	       (lexicographically< (rest x) (rest y))))))

(define (cardinality-circumscription kinematic-models)
 (define (cardinality kinematic-model)
  (list
   ;; First, minimize the number of grounded objects.
   (countq #t (vector->list (kinematic-model-ground? kinematic-model)))
   ;; Second, minimize the number of joints.
   (countq #t
	   (map (lambda (w j i) (or w j i))
		(vector->list (kinematic-model-joint-w? kinematic-model))
		(vector->list (kinematic-model-joint-j? kinematic-model))
		(vector->list (kinematic-model-joint-i? kinematic-model))))
   ;; Third, minimize the number of rigid joints. The various combinations of
   ;; revolute and prismatic joints are treated equivalently.
   (countq #t
	   (map (lambda (w j i) (and w j i))
		(vector->list (kinematic-model-joint-w? kinematic-model))
		(vector->list (kinematic-model-joint-j? kinematic-model))
		(vector->list (kinematic-model-joint-i? kinematic-model))))
   ;; Fourth, minimize the number of same-layer relations.
   (countq #t
	   (reduce append
		   (map vector->list
			(vector->list
			 (kinematic-model-same-layer? kinematic-model)))
		   '()))))
 (let ((minimal-cardinality
	(minp lexicographically< (map cardinality kinematic-models))))
  (remove-if (lambda (kinematic-model)
	      (lexicographically< minimal-cardinality
				  (cardinality kinematic-model)))
	     kinematic-models)))

(define (temporal-circumscription kinematic-models-movie)
 (define (state-changes kinematic-model-movie)
  (if (or (null? kinematic-model-movie) (null? (rest kinematic-model-movie)))
      0
      (+ (countq
	  #f
	  (vector->list
	   (map-vector
	    eq?
	    (kinematic-model-ground? (first kinematic-model-movie))
	    (kinematic-model-ground? (second kinematic-model-movie)))))
	 (state-changes (rest kinematic-model-movie)))))
 (define (grounded-objects kinematic-model-movie)
  (if (null? kinematic-model-movie)
      0
      (countq
       #t
       (vector->list
	(reduce
	 (lambda (g1 g2) (map-vector (lambda (g1 g2) (or g1 g2)) g1 g2))
	 (map kinematic-model-ground? kinematic-model-movie)
	 (kinematic-model-ground? (first kinematic-model-movie)))))))
 (define (grounded-non-moving-objects kinematic-model-movie)
  (reduce +
	  (map (lambda (kinematic-model)
		(countq #t
			(vector->list
			 (map-vector (lambda (ground? colour)
				      (and ground? (not (eq? colour 'moving))))
				     (kinematic-model-ground? kinematic-model)
				     *colours*))))
	       kinematic-model-movie)
	  0))
 (define (reciprocal-sum-cost kinematic-model-movie)
  (let ((ground? (map kinematic-model-ground? kinematic-model-movie)))
   (reduce
    +
    (map-n
     (lambda (i)
      (sum
       (lambda (i) (/ (+ i 1)))
       (countq #t (map (lambda (ground?) (vector-ref ground? i)) ground?))))
     (vector-length (first ground?)))
    0.0)))
 (define (cost kinematic-model-movie)
  (if *new-temporal-circumscription-cost?*
      (list (reciprocal-sum-cost kinematic-model-movie))
      (list (state-changes kinematic-model-movie))))
 (when (zero? (vector-length kinematic-models-movie)) (panic "Empty movie"))
 (unless (every-vector (lambda (kinematic-models)
			(not (null? kinematic-models)))
		       kinematic-models-movie)
  (panic "Some frame has no models"))
 (unless (every-vector
	  (lambda (kinematic-models)
	   (every (lambda (kinematic-model)
		   (= (vector-length
		       (kinematic-model-ground? kinematic-model))
		      (vector-length
		       (kinematic-model-ground?
			(first (vector-ref kinematic-models-movie 0))))))
		  kinematic-models))
	  kinematic-models-movie)
  (panic "Different frames have different numbers of objects"))
 (let loop ((kinematic-models-movie
	     (rest (vector->list kinematic-models-movie)))
	    (kinematic-model-movies
	     (map list (vector-ref kinematic-models-movie 0))))
  (if (null? kinematic-models-movie)
      (list->vector
       (reverse
	(map list
	     (minp (lambda (kinematic-model-movie1 kinematic-model-movie2)
		    (lexicographically< (cost kinematic-model-movie1)
					(cost kinematic-model-movie2)))
		   kinematic-model-movies))))
      (loop (rest kinematic-models-movie)
	    (map (lambda (kinematic-model)
		  (minp (lambda (kinematic-model-movie1 kinematic-model-movie2)
			 (lexicographically< (cost kinematic-model-movie1)
					     (cost kinematic-model-movie2)))
			(map (lambda (kinematic-model-movie)
			      (cons kinematic-model kinematic-model-movie))
			     kinematic-model-movies)))
		 (first kinematic-models-movie))))))

(define (minimal-membersp <? =? l)
 ;; note: There is a bug in Scheme->C which doesn't allow < and = to shadow
 ;;       the global bindings which is why these are named <? and =?.
 ;; Belongs in QobiScheme.
 (when (null? l) (fuck-up))
 (let loop ((xs (list (first l))) (l (rest l)))
  (if (null? l)
      xs
      (loop (cond ((<? (first l) (first xs)) (list (first l)))
		  ((=? (first l) (first xs)) (cons (first l) xs))
		  (else	xs))
	    (rest l)))))

(define (circumscription kinematic-models-movie)
 (temporal-circumscription
  (if *twice?*
      (map-vector
       (lambda (kinematic-models ground?)
	(cardinality-circumscription
	 (remove-if-not
	  (lambda (kinematic-model)
	   (equal? (kinematic-model-ground? kinematic-model) ground?))
	  kinematic-models)))
       kinematic-models-movie
       (map-vector (lambda (kinematic-models)
		    (kinematic-model-ground? (first kinematic-models)))
		   (temporal-circumscription kinematic-models-movie)))
      (map-vector cardinality-circumscription kinematic-models-movie))))

(define (simplified-minimal-polygon-kinematic-model? kinematic-model polygons)
 (let ((contacts (all-contacts polygons *epsilon2*)))
  (and
   ;; Correct number of polygons.
   (= (length polygons)
      (vector-length (kinematic-model-ground? kinematic-model)))
   ;; Correct number of contacts.
   (= (length contacts)
      (vector-length (kinematic-model-joint-i? kinematic-model)))
   ;; Joined line segments intersect.
   (every
    (lambda (contact k)
     (or (and (not (vector-ref (kinematic-model-joint-i? kinematic-model) k))
	      (not (vector-ref (kinematic-model-joint-j? kinematic-model) k))
	      (not (vector-ref (kinematic-model-joint-w? kinematic-model) k)))
	 (line-segments-intersect?
	  (contact-l1 contact) (contact-l2 contact) *epsilon2*)))
    contacts
    (enumerate (length contacts)))
   ;; Polygons on same layer don't cross.
   (every-n (lambda (j)
	     (every-n (lambda (i)
		       (or (not (matrix-ref
				 (kinematic-model-same-layer? kinematic-model)
				 i
				 j))
			   (not (polygons-cross? (list-ref polygons i)
						 (list-ref polygons j)
						 *epsilon1*))))
		      j))
	    (length polygons))
   ;; needs work: Also need to add in the kludge.
   ;; Stable.
   (polygon-stable? polygons contacts kinematic-model)
   ;; Remove a ground assertion.
   (not (possibly?
	 (polygon-stable?
	  polygons
	  contacts
	  (make-kinematic-model
	   (kinematic-model-same-layer? kinematic-model)
	   (kinematic-model-joint-i? kinematic-model)
	   (kinematic-model-joint-j? kinematic-model)
	   (kinematic-model-joint-w? kinematic-model)
	   (a-ground?-daughter (kinematic-model-ground? kinematic-model))))))
   ;; Remove a same-layer assertion.
   (not (possibly? (polygon-stable?
		    polygons
		    contacts
		    (make-kinematic-model
		     (vector->upper-triangular-matrix
		      (a-polygon-same-layer?-daughter
		       polygons
		       (upper-triangular-matrix->vector
			(kinematic-model-same-layer? kinematic-model))
		       contacts
		       (kinematic-model-joint-i? kinematic-model)
		       (kinematic-model-joint-j? kinematic-model)
		       (kinematic-model-joint-w? kinematic-model)))
		     (kinematic-model-joint-i? kinematic-model)
		     (kinematic-model-joint-j? kinematic-model)
		     (kinematic-model-joint-w? kinematic-model)
		     (kinematic-model-ground? kinematic-model)))))
   ;; Remove a joint.
   (not (possibly?
	 (let ((joined?
		(a-polygon-joined?-daughter
		 contacts (kinematic-model-joint-i? kinematic-model))))
	  (polygon-stable?
	   polygons
	   contacts
	   (make-kinematic-model
	    (kinematic-model-same-layer? kinematic-model)
	    joined?
	    joined?
	    (map-vector (lambda (joined? joint-w?) (and joined? joint-w?))
			joined?
			(kinematic-model-joint-w? kinematic-model))
	    (kinematic-model-ground? kinematic-model))))))
   ;; Demote a joint from rigid to revolute.
   (or *fast?*
       (not (possibly? (polygon-stable?
			polygons
			contacts
			(make-kinematic-model
			 (kinematic-model-same-layer? kinematic-model)
			 (kinematic-model-joint-i? kinematic-model)
			 (kinematic-model-joint-j? kinematic-model)
			 (a-polygon-joint?-daughter
			  (kinematic-model-joint-w? kinematic-model))
			 (kinematic-model-ground? kinematic-model)))))))))

(define (forward-simplified-minimal-polygon-kinematic-models polygons)
 (if *cache?*
     (or (find-if (lambda (kinematic-models)
		   (every (lambda (kinematic-model)
			   (simplified-minimal-polygon-kinematic-model?
			    kinematic-model polygons))
			  kinematic-models))
		  *kinematic-models-cache*)
	 (let ((kinematic-models
		(simplified-minimal-polygon-kinematic-models polygons)))
	  (set! *kinematic-models-cache*
		(cons kinematic-models *kinematic-models-cache*))
	  kinematic-models))
     (simplified-minimal-polygon-kinematic-models polygons)))

(define (colour-of i)
 (if *colours*
     (vector-ref *colours* i)
     (string->symbol (string (integer->char (+ (char->integer #\A) i))))))

(define (format-event e is)
 (format #f "~s@~a"
	 (cons (first e) (map colour-of (rest e)))
	 (string-append "{"
			(reduce (lambda (x y) (string-append x ", " y))
				(map extended-spanning-interval->string is)
				"")
			"}")))

(define (kinematic-model->ppm code? i polygons kinematic-model)
 (let ((pixmap (pnm->pixmap (if code?
				(meteor-code-frame->ppm i)
				(meteor-frame->ppm i)))))
  (define (draw-line-segment l)
   (xdrawline *display* pixmap *thin-gc*
	      (x (p l)) (y (p l)) (x (q l)) (y (q l))))
  (define (draw-polygon p)
   (for-each-n
    (lambda (i)
     (draw-line-segment
      (make-line-segment (vector-ref p i)
			 (vector-ref p (modulo (+ i 1) (vector-length p))))))
    (vector-length p)))
  (define (draw-bar r l)
   (let* ((dx (- (x (p l)) (x (q l))))
	  (dy (- (y (p l)) (y (q l))))
	  (m (sqrt (+ (sqr dx) (sqr dy))))
	  (dx (* 10.0 (/ dx m)))
	  (dy (* 10.0 (/ dy m))))
    (xdrawline *display* pixmap *medium-gc*
	       (- (x r) dx) (- (y r) dy) (+ (x r) dx) (+ (y r) dy))))
  (define (draw-filled-circle p diameter)
   (xfillarc *display* pixmap *thin-gc*
	     (- (x p) (/ diameter 2))
	     (- (y p) (/ diameter 2))
	     diameter diameter 0 (* 360 64)))
  (define (draw-unfilled-circle p diameter)
   (xdrawarc *display* pixmap *thin-gc*
	     (- (x p) (/ diameter 2))
	     (- (y p) (/ diameter 2))
	     diameter diameter 0 (* 360 64)))
  (cond
   (code?
    (xsetforeground *display* *thin-gc* (xcolor-pixel (second *foreground*)))
    (xsetforeground *display* *medium-gc* (xcolor-pixel (second *foreground*)))
    (xsetforeground *display* *roman-gc* (xcolor-pixel (second *foreground*))))
   (else
    (xsetforeground *display* *thin-gc* (xcolor-pixel (second *background*)))
    (xsetforeground *display* *medium-gc* (xcolor-pixel (second *background*)))
    (xsetforeground
     *display* *roman-gc* (xcolor-pixel (second *background*)))))
  (for-each (lambda (polygon) (draw-polygon polygon)) polygons)
  (when kinematic-model
   (let* ((ns (enumerate (length polygons)))
	  (same-layer?(kinematic-model-same-layer? kinematic-model))
	  (joint-i? (kinematic-model-joint-i? kinematic-model))
	  (joint-j? (kinematic-model-joint-j? kinematic-model))
	  (joint-w? (kinematic-model-joint-w? kinematic-model))
	  (ground? (kinematic-model-ground? kinematic-model))
	  (layers
	   (equivalence-classesp
	    (lambda (pi pj)
	     (matrix-ref
	      same-layer? (positionq pi polygons) (positionq pj polygons)))
	    polygons))
	  (contacts (all-contacts polygons *epsilon2*)))
    (for-each (lambda (pi)
	       (let ((layer (number->string (positionp memq pi layers)))
		     (p (layer-id-location pi)))
		(xdrawstring *display* pixmap *roman-gc*
			     (x p) (y p) layer (string-length layer))))
	      polygons)
    (for-each (lambda (contact k)
	       (let ((p (contact-p contact))
		     (l1 (contact-l1 contact))
		     (l2 (contact-l2 contact)))
		(if (vector-ref joint-i? k)
		    (if (vector-ref joint-j? k)
			(if (vector-ref joint-w? k)
			    (draw-filled-circle p 10)
			    (draw-unfilled-circle p 10))
			(if (vector-ref joint-w? k)
			    (begin (draw-filled-circle p 10)
				   (draw-bar p l2))
			    (begin (draw-unfilled-circle p 10)
				   (draw-bar p l2))))
		    (if (vector-ref joint-j? k)
			(if (vector-ref joint-w? k)
			    (begin (draw-filled-circle p 10)
				   (draw-bar p l1))
			    (begin (draw-unfilled-circle p 10)
				   (draw-bar p l1)))
			(if (vector-ref joint-w? k)
			    (begin (draw-filled-circle p 10)
				   (draw-bar p l1)
				   (draw-bar p l2))
			    #f)))))
	      contacts
	      (enumerate (length contacts)))
    (for-each (lambda (pi i)
	       (when (vector-ref ground? i)
		(let ((p (ground-location pi)))
		 (draw-line-segment
		  (make-line-segment p (v+ p '#(0 5))))
		 (draw-line-segment
		  (make-line-segment (v+ p '#(-3 5)) (v+ p '#(3 5))))
		 (draw-line-segment
		  (make-line-segment (v+ p '#(-2 7)) (v+ p '#(2 7))))
		 (draw-line-segment
		  (make-line-segment (v+ p '#(-1 9)) (v+ p '#(1 9)))))))
	      polygons
	      ns)))
  (cond
   (code?
    (xsetforeground *display* *thin-gc* (xcolor-pixel (second *background*)))
    (xsetforeground *display* *medium-gc* (xcolor-pixel (second *background*)))
    (xsetforeground
     *display* *roman-gc* (xcolor-pixel (second *background*))))
   (else
    (xsetforeground *display* *thin-gc* (xcolor-pixel (second *foreground*)))
    (xsetforeground *display* *medium-gc* (xcolor-pixel (second *foreground*)))
    (xsetforeground
     *display* *roman-gc* (xcolor-pixel (second *foreground*)))))
  (let ((ximage (xgetimage *display* pixmap 0 0 320 240 -1 zpixmap)))
   (free-pixmap pixmap)
   (let ((ppm (ximage->ppm ximage 240 320)))
    (xdestroyimage ximage)
    ppm))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
