(MODULE HLIB-LOCATE)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-locate.sch")
(include "c-externals.sc")

;;; TTMTTD
;;;  1. Change location system so that it can accomodate smart point choosing
;;;  2. Make a point tracker and keep track of image-vector/model-vector
;;;     correspondences

;;; needs work: LOCATE:MODEL and LOCATE:PARALLEL? prefixed to avoid name
;;;             conflicts

;;; Structures

(define-structure quad origin three-points inverse)

(define-structure locate:model name points line-segments quads)

;;; Constants

;;; *FOCAL-LENGTH* is the distance of the image plane G from the centre of
;;; projection O.  Corresponds to f in figure 1.
(define *focal-length* 760)

;;; *LLOCATE-EPSILON* is the maximum allowable amount of noise (measured in
;;; pixel distances) between an actual image vector and the
;;; perspective-projected vertex from a possible transformation of the model.
(define *llocate-epsilon* 2)

;;; *MINIMUM-FIT* is the minimum number of vertices from a POSIT3 calculated
;;; transformation of a model which must have corresponding image vectors
;;; before that transformation is considered viable.
(define *minimum-fit* 2)

;;; *MAXIMUM-POSIT-ITERATIONS* is the maximum number of iterations which POSIT3
;;; will perform.
(define *maximum-posit-iterations* 30)

;;; If all of the SOP-vectors from the previous iteration of POSIT are within
;;; *POSIT-CONVERGENCE-THRESHOLD* pixels of the SOP-vectors of the current
;;; iteration, POSIT is considered converged.
(define *posit-convergence-threshold* 1)

(define (set-focal-length! x)
 (unless (positive? x)
  (panic "SET-FOCAL-LENGTH!: ~A is not a valid focal length" x))
 (set! *focal-length* x))

(define (set-epsilon! x)
 (unless (positive? x) (panic "SET-EPSILON!: ~A is not a valid value" x))
 (set! *llocate-epsilon* x))

(define (set-minimum-fit! x)
 (unless (positive? x) (panic "SET-MINIMUM-FIT!: ~A is not a valid value" x))
 (set! *minimum-fit* x))

(define (set-maximum-posit-iterations! x)
 (unless (positive? x)
  (panic "SET-MAXIMUM-POSIT-ITERATIONS!: ~A is not a valid value" x))
 (set! *maximum-posit-iterations* x))

(define (set-posit-convergence-threshold! x)
 (unless (positive? x)
  (panic "SET-POSIT-CONVERGENCE-THRESHOLD!: ~A is not a valid value" x))
 (set! *posit-convergence-threshold* x))

;;; Procedures

;;; Model creation and manipulation

(define (create-model name points line-segments)
 (make-locate:model name
		    points
		    line-segments
		    (all-values (let* ((o (a-member-of points))
				       (after-o (rest (memq o points)))
				       (x (a-member-of after-o))
				       (after-x (rest (memq x after-o)))
				       (y (a-member-of after-x))
				       (after-y (rest (memq y after-x)))
				       (z (a-member-of after-y))
				       (x (v- x o))
				       (y (v- y o))
				       (z (v- z o))
				       (b (invert-matrix (vector x y z))))
				 (unless b (fail))
				 (make-quad o (list x y z) b)))))

(define (create-block name length width height)
 (let ((p0 (vector    0       0      0))
       (p1 (vector    0       0 height))
       (p2 (vector    0   width      0))
       (p3 (vector    0   width height))
       (p4 (vector length     0      0))
       (p5 (vector length     0 height))
       (p6 (vector length width      0))
       (p7 (vector length width height)))
  (create-model name
		(list p0 p1 p2 p3 p4 p5 p6 p7)
		(list (make-line-segment p0 p1)
		      (make-line-segment p0 p2)
		      (make-line-segment p0 p4)
		      (make-line-segment p1 p3)
		      (make-line-segment p1 p5)
		      (make-line-segment p2 p3)
		      (make-line-segment p2 p6)
		      (make-line-segment p3 p7)
		      (make-line-segment p4 p5)
		      (make-line-segment p4 p6)
		      (make-line-segment p5 p7)
		      (make-line-segment p6 p7)))))

;;; POSIT calculations

;;; SOP3 returns a list of three 2D VECTORs.  THREE-IMAGE-POINTS is a list of
;;; three 2D VECTORs, THREE-MODEL-POINTS is a list of three 3D VECTORs.  K is a 3D
;;; VECTOR and SCALE is a REAL.  See eq (9) in the paper for details.
(define (sop3 three-image-points three-model-points k scale)
 (let ((ratio (/ scale *focal-length*)))
  (map (lambda (model-point image-point)
	(k*v (+ 1 (* ratio (dot k model-point))) image-point))
       three-model-points three-image-points)))

;;; CONVERGED? determines whether POSIT has converged on a value by checking if
;;; each 2D VECTOR in the NEW-SOP and OLD-SOP lists are within the posit
;;; convergence threshold distance (in pixels) away from each other.
(define (converged? new-sop old-sop)
 (every (lambda (x y) (< (distance x y) *posit-convergence-threshold*))
	new-sop old-sop))

;;; POSIT3 calculates object pose using four image vectors and the
;;; corresponding model vectors using the algorithm described in `Model-Based
;;; Object Pose in 25 Lines of Code' by DeMenthon and Davis.  IMAGE-ORIGIN is
;;; the image vector corresponding the model reference point and is a
;;; 2D VECTOR.  THREE-IMAGE-POINTS is a list of three 2D VECTORs from the
;;; true-perspective-projection of the model.  THREE-MODEL-POINTS is a list of
;;; three 3D VECTORs which, in the model coordinate system, represent the three
;;; vertices corresponding to the vectors in TPP.  INVERSE is the object
;;; matrix described near the top of page 338 of the paper.
(define (posit3 image-origin
		three-image-points
		model-origin
		three-model-points
		inverse)
 (define (loop sop iteration-number)
  (let* ((i0x (x image-origin))
	 (i0y (y image-origin))
	 (x (list->vector (map (lambda (p) (- (x p) i0x)) sop)))
	 (y (list->vector (map (lambda (p) (- (y p) i0y)) sop)))
	 (unscaled-i (m*v inverse x))	; equation 6(a)
	 (unscaled-j (m*v inverse y))	; equation 6(a)
	 ;; programming by analogy to their code
	 (scale (/ (+ (magnitude unscaled-i) (magnitude unscaled-j)) 2))
	 (i (unit unscaled-i))	
	 (j (unit unscaled-j))
	 (k (cross i j))
	 (new-sop (sop3 three-image-points three-model-points k scale)))
   (cond
    ((converged? new-sop sop)
     (and (< (abs (dot i j)) 0.2)
	  (let ((transform
		 (compose-transforms
		  (make-transform (map-vector - model-origin)
				  (vector (vector 1 0 0)
					  (vector 0 1 0)
					  (vector 0 0 1)))
		  (make-transform (vector (/ i0x scale)
					  (/ i0y scale)
					  (/ *focal-length* scale))
				  (vector i j k)))))
	   ;; note: I'm not sure why the following check is necessary?
	   (and (every (lambda (image-point model-point)
			(< (distance
			    (project (apply-transform transform model-point)
				     *focal-length*)
			    image-point)
			   *llocate-epsilon*))
		       three-image-points
		       three-model-points)
		transform))))
    ((>= iteration-number *maximum-posit-iterations*) #f)
    (else (loop new-sop (+ 1 iteration-number))))))
 (loop three-image-points 1))

;;; Instance finder

(define (n-distinct-members-of n l)
 (if (zero? n)
     '()
     (let ((x (a-member-of l)))
      (cons x (n-distinct-members-of (- n 1) (rest (memq x l)))))))

(define (conjugate v) (vector (- (y v)) (x v)))

(define (orthogonal? u v)
 ;; hardwird
 (< (abs (dot (unit u) (unit v))) 0.15))

(define (locate:parallel? l1 l2)
 (orthogonal? (unit (v- (q l1) (p l1))) (conjugate (unit (v- (q l2) (p l2))))))

(define (distance-from-2d-point-to-line r l)
 ;; worst case:
 ;; 4 +
 ;; 1 unary -
 ;; 5 binary -
 ;; 9 *
 (let* ((p (p l))
	(q (q l))
	(rx (x r))
	(ry (y r))
	(px (x p))
	(py (y p))
	(qx (x q))
	(qy (y q))
	(c1 (- qx px))
	(c2 (- qy py))
	(c3 (* c1 c1))
	(c4 (* c2 c2))
	(c5 (- rx px)))
  (sqrt
   (cond ((zero? c1) (* c5 c5))
	 ((zero? c2) (* (- ry py) (- ry py)))
	 (else
	  (let* ((dy (- ry (/ (+ (* py c3) (* c5 c1 c2) (* ry c4)) (+ c3 c4))))
		 (dx (- (/ (* dy c2) c1))))
	   (+ (* dx dx) (* dy dy))))))))

(define (distance-between-lines l1 l2)
 (min (distance-from-2d-point-to-line (p l1) l2)
      (distance-from-2d-point-to-line (q l1) l2)
      (distance-from-2d-point-to-line (p l2) l1)
      (distance-from-2d-point-to-line (q l2) l1)))

(define (tangential-position r l) (dot (v- r (p l)) (tangent l)))

(define (overlap l1 l2)
 (abs (- (min (line-segment-length l2)
	      (max 0 (tangential-position (p l1) l2)))
	 (min (line-segment-length l2)
	      (max 0 (tangential-position (q l1) l2))))))

(define (transform-and-project-model transform model)
 (map (lambda (line-segment)
       (make-line-segment
	(project (apply-transform transform (p line-segment)) *focal-length*)
	(project (apply-transform transform (q line-segment)) *focal-length*)))
      (locate:model-line-segments model)))

(define (find-model image-points image-line-segments model)
 (one-value
  (let* ((quad (a-member-of (locate:model-quads model)))
	 (four-image-points (n-distinct-members-of 4 image-points))
	 (image-origin (a-member-of four-image-points))
	 (three-image-points
	  (a-permutation-of (removeq image-origin four-image-points)))
	 (transform (posit3 image-origin
			    three-image-points
			    (quad-origin quad)
			    (quad-three-points quad)
			    (quad-inverse quad))))
   (unless (and
	    transform
	    (>= (count-if
		 (lambda (model-line-segment)
		  (some
		   (lambda (image-line-segment)
		    (and
		     (parallel? model-line-segment image-line-segment)
		     (<= (distance-between-lines
			  model-line-segment image-line-segment)
			 ;; hardwired
			 2)
		     (> (overlap image-line-segment model-line-segment)
			(* 0.75		;hardwired
			   (min (line-segment-length image-line-segment)
				(line-segment-length model-line-segment))))))
		   image-line-segments))
		 (transform-and-project-model transform model))
		;; hardwired
		5))
    (fail))
   transform)
  #f))

;;; Tam V'Nishlam Shevah L'El Borei Olam
