(MODULE HLIB-MERGE-LINES)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-merge-lines.sch")
(include "c-externals.sc")

;;; Variables

(define *angle-threshold* (degrees->radians 5))
(define *scale* 20)
(define *relative-tangential-distance-threshold* 0.5)
(define *relative-normal-distance-threshold* 0.10)

;;; Procedures

;;; vector from p to q
(define (line-segment->vector l) (v- (q l) (p l)))

;;; angle traversing from v1 to v2 (-pi .. pi) (positive if ccw)
(define (vectors-angle v1 v2)
 (asin (/ (cross-2d v1 v2) (* (magnitude v1) (magnitude v2)))))

;;; angle traversing from l1 to l2 (-pi/2 .. pi/2) (positive if ccw)
(define (line-segments-angle l1 l2)
 (let* ((v1 (line-segment->vector l1))
	(v2 (line-segment->vector l2))
	;; flip vector if directions different
	(v2 (if (< (dot v2 v1) 0) (vector (- (x v2)) (- (y v2))) v2)))
  (vectors-angle v1 v2)))

(define (line-segment-centroid l)
 (let ((p (p l))
       (q (q l)))
  (vector (/ (+ (x p) (x q)) 2) (/ (+ (y p) (y q)) 2))))

;;; Merge lines to single line or return #f if lines not merged
(define (old-merge-two-line-segments l1 l2)
 (let* ((c1 (line-segment-centroid l1))
	(c2 (line-segment-centroid l2))
	(m1 (line-segment-length l1))
	(m2 (line-segment-length l2))
	(axis (make-line-segment (vector 0 0) (vector 1 0)))
	(a1 (line-segments-angle axis l1))
	(a2 (line-segments-angle axis l2))
	;; make new line from (weighted) average centroid and angle
	(c (vector (/ (+ (* m1 (x c1)) (* m2 (x c2))) (+ m1 m2))
		   (/ (+ (* m1 (y c1)) (* m2 (y c2))) (+ m1 m2))))
	(a (/ (+ (* m1 a1) (* m2 a2)) (+ m1 m2)))
	(v (vector (cos a) (sin a)))
	(n (vector (- (y v)) (x v))))
  ;; check conditions for merging, return #f if fails
  (cond ((> (abs (- a1 a)) (* *angle-threshold* (/ *scale* m1))) #f)
	((> (abs (- a2 a)) (* *angle-threshold* (/ *scale* m2))) #f)
	((> (/ (- (distance c1 c2) (/ (+ m1 m2) 2)) (min m1 m2))
	    *relative-tangential-distance-threshold*)
	 #f)
	((> (dot (v- c1 c) n) (* *scale* *relative-normal-distance-threshold*))
	 #f)
	((> (dot (v- c2 c) n) (* *scale* *relative-normal-distance-threshold*))
	 #f)
	;; perform merge and return line
	(else (let* ((p1 (p l1))
		     (q1 (q l1))
		     (p2 (p l2))
		     (q2 (q l2))
		     ;; set line to be longest of any parts
		     (p1c (dot (v- p1 c) v))
		     (q1c (dot (v- q1 c) v))
		     (p2c (dot (v- p2 c) v))
		     (q2c (dot (v- q2 c) v))
		     (d1 (min p1c q1c p2c q2c))
		     (d2 (max p1c q1c p2c q2c)))
	       (make-line-segment
		(vector (inexact->exact (round (+ (x c) (* (x v) d1))))
			(inexact->exact (round (+ (y c) (* (y v) d1)))))
		(vector (inexact->exact (round (+ (x c) (* (x v) d2))))
			(inexact->exact (round (+ (y c) (* (y v) d2)))))))))))

(define (acute x)
 (cond ((> x half-pi) (- x pi))
       ((<= x (- half-pi)) (+ x pi))
       (else x)))

(define (merge-two-line-segments l1 l2)
 (let* ((l1px (x (p l1)))
	(l1py (y (p l1)))
	(l1qx (x (q l1)))
	(l1qy (y (q l1)))
	(l2px (x (p l2)))
	(l2py (y (p l2)))
	(l2qx (x (q l2)))
	(l2qy (y (q l2)))
	(c1x (/ (+ l1px l1qx) 2))
	(c1y (/ (+ l1py l1qy) 2))
	(c2x (/ (+ l2px l2qx) 2))
	(c2y (/ (+ l2py l2qy) 2))
	(m1 (sqrt (+ (sqr (- l1px l1qx)) (sqr (- l1py l1qy)))))
	(m2 (sqrt (+ (sqr (- l2px l2qx)) (sqr (- l2py l2qy)))))
	(a1 (acute (atan (- l1py l1qy) (- l1px l1qx))))
	(a2 (acute (atan (- l2py l2qy) (- l2px l2qx))))
	;; make new line from (weighted) average centroid and angle
	(cx (/ (+ (* m1 c1x) (* m2 c2x)) (+ m1 m2)))
	(cy (/ (+ (* m1 c1y) (* m2 c2y)) (+ m1 m2)))
	(a (/ (+ (* m1 a1) (* m2 a2)) (+ m1 m2)))
	(vx (cos a))
	(vy (sin a)))
  ;; check conditions for merging, return #f if fails
  (cond ((> (abs (- a1 a)) (* *angle-threshold* (/ *scale* m1))) #f)
	((> (abs (- a2 a)) (* *angle-threshold* (/ *scale* m2))) #f)
	((> (/ (- (sqrt (+ (sqr (- c1x c2x)) (sqr (- c1y c2y))))
		  (/ (+ m1 m2) 2))
	       (min m1 m2))
	    *relative-tangential-distance-threshold*)
	 #f)
	((> (+ (* (- c1x cx) (- vy)) (* (- c1y cy) vx))
	    (* *scale* *relative-normal-distance-threshold*))
	 #f)
	((> (+ (* (- c2x cx) (- vy)) (* (- c2y cy) vx))
	    (* *scale* *relative-normal-distance-threshold*))
	 #f)
	;; perform merge and return line
	;; set line to be longest of any parts
	(else (let* ((p1c (+ (* (- l1px cx) vx) (* (- l1py cy) vy)))
		     (q1c (+ (* (- l1qx cx) vx) (* (- l1qy cy) vy)))
		     (p2c (+ (* (- l2px cx) vx) (* (- l2py cy) vy)))
		     (q2c (+ (* (- l2qx cx) vx) (* (- l2qy cy) vy)))
		     (d1 (min p1c q1c p2c q2c))
		     (d2 (max p1c q1c p2c q2c)))
	       (make-line-segment
		(vector (inexact->exact (round (+ cx (* vx d1))))
			(inexact->exact (round (+ cy (* vy d1)))))
		(vector (inexact->exact (round (+ cx (* vx d2))))
			(inexact->exact (round (+ cy (* vy d2)))))))))))

(define (group-line-segments line-segments)
 (let ((again? #f))
  (for-each
   (lambda (l1)
    (let ((tail (memq l1 line-segments)))
     (when tail
      (call-with-current-continuation
       (lambda (return)
	(for-each
	 (lambda (l2)
	  (let ((l (merge-two-line-segments l1 l2)))
	   (when l
	    (set! again? #t)
	    (set! line-segments (remq! l2 (remq! l1 line-segments)))
	    (set-cdr! (last-pair line-segments) (list l))
	    (return #t))))
	 (rest tail)))))))
   line-segments)
  (if again? (group-line-segments line-segments) line-segments)))

;;; filter line segments for length
(define (filter-line-segments line-segments)
 (remove-if-not (lambda (l) (>= (line-segment-length l) *scale*))
		line-segments))

(define (merge-line-segments line-segments)
 (filter-line-segments
  (group-line-segments (map (lambda (x) x) line-segments))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
