(MODULE HLIB-RATIO-REGIONS)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-ratio-regions.sch")
(include "c-externals.sc")

;;; Procedures

;;; The vertices are s, t, and (y,x).
;;; C-RIGHT[y,x] is the capacity from (y,x) to (y,x+1) which is the same as the
;;; capacity from (y,x+1) to (y,x).
;;; C-DOWN[y,x] is the capacity from (y,x) to (y+1,x) which is the same as the
;;; capacity from (y+1,x) to (y,x).
;;; The capacity from s to (y,0), (0,x), (y,Y-1), (0,X-1) is implicitly
;;; infinite.
;;; The capacity from (x,y) to t is V*W[y,x].
;;; F-RIGHT[y,x] is the preflow from (y,x) to (y,x+1) which is the negation of
;;; the preflow from (y,x+1) to (y,x).
;;; F-DOWN[y,x] is the preflow from (y,x) to (y+1,x) which is the negation of
;;; the preflow from (y+1,x) to (y,x).
;;; We do not record the preflow from s to (y,X-1), (y,0), (Y-1,x), and (0,x)
;;; and from (y,X-1), (y,0), (Y-1,x), and (0,x) to s.
;;; F-OUT[y,x] is the preflow from (y,x) to t.
;;; We do not record the preflow from t to (y,x).
;;; {C,F}-RIGHT[0:Y-1,0:X-2].
;;; {C,F}-DOWN[0:Y-2,0:X-1].
;;; F-OUT[0:Y-1,0:X-1]
;;; For now, we will keep all capacities (and thus all preflows) as integers.
;;; (CF-RIGHT y x) is the residual capacity from (y,x) to (y,x+1).
;;; (CF-LEFT y x) is the residual capacity from (y,x) to (y,x-1).
;;; (CF-DOWN y x) is the residual capacity from (y,x) to (y+1,x).
;;; (CF-UP y x) is the residual capacity from (y,x) to (y-1,x).
;;; We do not compute the residual capacities from s to (y,X-1), (y,0),
;;; (Y-1,x), and (0,x) because they are all infinite.
;;; We do not compute the residual capacities from (y,X-1), (y,0), (Y-1,x),
;;; and (0,x) to s because they will never be used.
;;; (CF-OUT y x) is the residual capacity from (y,x) to t.
;;; We do not compute the residual capacity from t to (y,x) because it will
;;; be used.
;;; (EF-RIGHT? y x) is true if there is an edge from (y,x) to (y,x+1) in the
;;; residual network.
;;; (EF-LEFT? y x) is true if there is an edge from (y,x) to (y,x-1) in the
;;; residual network.
;;; (EF-DOWN? y x) is true if there is an edge from (y,x) to (y+1,x) in the
;;; residual network.
;;; (EF-UP? y x) is true if there is an edge from (y,x) to (y-1,x) in the
;;; residual network.
;;; (EF-OUT? y x) is true if there is an edge from (y,x) to t in the
;;; residual network.
;;; There are always edges in the residual network from s to (y,X-1), (y,0),
;;; (Y-1,x), and (0,x).
;;; We don't care whether there are edges in the residual network from
;;; (y,X-1), (y,0), (Y-1,x), and (0,x) to s because they will never be used.
;;; We don't care whether there are edges in the residual network from t to
;;; (y,x) because they will never be used.

(define (positive-min x y) (if (negative? x) y (min x y)))

(define (positive- x y) (if (negative? x) x (- x y)))

(define (positive+ x y) (if (negative? x) x (+ x y)))

(define (ratio-region mask c-right c-down w lg-max-v)
 (let* ((height (matrix-rows w))
	(width (matrix-columns w))
	(f-right (make-matrix height (- width 1) 0))
	(f-down (make-matrix (- height 1) width 0))
	(f-out (make-matrix height width 0))
	(h (make-matrix height width 0))
	(e (make-matrix height width 0))
	(marked? (make-matrix height width #f))
	(m0 (* height width))
	(m1 (+ (* height width) 1))
	(q (make-vector m1 '())))
  (define (cf-right y x)
   (- (matrix-ref c-right y x) (matrix-ref f-right y x)))
  (define (cf-left y x)
   (+ (matrix-ref c-right y (- x 1)) (matrix-ref f-right y (- x 1))))
  (define (cf-down y x)
   (- (matrix-ref c-down y x) (matrix-ref f-down y x)))
  (define (cf-up y x)
   (+ (matrix-ref c-down (- y 1) x) (matrix-ref f-down (- y 1) x)))
  (define (ef-right? y x) (positive? (cf-right y x)))
  (define (ef-left? y x) (positive? (cf-left y x)))
  (define (ef-down? y x) (positive? (cf-down y x)))
  (define (ef-up? y x) (positive? (cf-up y x)))
  (define (preflow-push! v)
   (let ((pushes 0)
	 (lifts 0)
	 (relabels 0)
	 (total-flow 0))
    (define (enqueue! y x)
     (unless (matrix-ref marked? y x)
      (vector-set! q
		   (matrix-ref h y x)
		   (cons (vector x y) (vector-ref q (matrix-ref h y x))))
      (matrix-set! marked? y x #t)))
    (define (cf-out y x) (- (* v (matrix-ref w y x)) (matrix-ref f-out y x)))
    (define (ef-out? y x) (positive? (cf-out y x)))
    (define (can-push-out? y x)
     (and (not (zero? (matrix-ref e y x)))
	  (ef-out? y x)
	  (= (matrix-ref h y x) 1)))
    (define (can-push-right? y x)
     (and (< x (- width 1))
	  (not (zero? (matrix-ref e y x)))
	  (matrix-ref mask y (+ x 1))
	  (ef-right? y x)
	  (= (matrix-ref h y x) (+ (matrix-ref h y (+ x 1)) 1))))
    (define (can-push-left? y x)
     (and (> x 0)
	  (not (zero? (matrix-ref e y x)))
	  (matrix-ref mask y (- x 1))
	  (ef-left? y x)
	  (= (matrix-ref h y x) (+ (matrix-ref h y (- x 1)) 1))))
    (define (can-push-down? y x)
     (and (< y (- height 1))
	  (not (zero? (matrix-ref e y x)))
	  (matrix-ref mask (+ y 1) x)
	  (ef-down? y x)
	  (= (matrix-ref h y x) (+ (matrix-ref h (+ y 1) x) 1))))
    (define (can-push-up? y x)
     (and (> y 0)
	  (not (zero? (matrix-ref e y x)))
	  (matrix-ref mask (- y 1) x)
	  (ef-up? y x)
	  (= (matrix-ref h y x) (+ (matrix-ref h (- y 1) x) 1))))
    (define (can-lift? y x)
     (and (not (zero? (matrix-ref e y x)))
	  (< (matrix-ref h y x) m1)
	  (or (not (ef-out? y x)) (= (matrix-ref h y x) 0))
	  (or (= x (- width 1))
	      (not (matrix-ref mask y (+ x 1)))
	      (not (ef-right? y x))
	      (<= (matrix-ref h y x) (matrix-ref h y (+ x 1))))
	  (or (= x 0)
	      (not (matrix-ref mask y (- x 1)))
	      (not (ef-left? y x))
	      (<= (matrix-ref h y x) (matrix-ref h y (- x 1))))
	  (or (= y (- height 1))
	      (not (matrix-ref mask (+ y 1) x))
	      (not (ef-down? y x))
	      (<= (matrix-ref h y x) (matrix-ref h (+ y 1) x)))
	  (or (= y 0)
	      (not (matrix-ref mask (- y 1) x))
	      (not (ef-up? y x))
	      (<= (matrix-ref h y x) (matrix-ref h (- y 1) x)))))
    (define (push-out! y x)
     ;;(format #t "Push out ~s ~s~%" y x)
     (let ((df-u-v (positive-min (matrix-ref e y x) (cf-out y x))))
      (matrix-set! f-out y x (+ (matrix-ref f-out y x) df-u-v))
      (matrix-set! e y x (positive- (matrix-ref e y x) df-u-v))
      (set! total-flow (+ total-flow df-u-v))))
    (define (push-right! y x)
     ;;(format #t "Push right ~s ~s~%" y x)
     (let ((df-u-v (positive-min (matrix-ref e y x) (cf-right y x))))
      (matrix-set! f-right y x (+ (matrix-ref f-right y x) df-u-v))
      (matrix-set! e y x (positive- (matrix-ref e y x) df-u-v))
      (matrix-set! e y (+ x 1) (positive+ (matrix-ref e y (+ x 1)) df-u-v))
      (enqueue! y (+ x 1))))
    (define (push-left! y x)
     ;;(format #t "Push left ~s ~s~%" y x)
     (let ((df-u-v (positive-min (matrix-ref e y x) (cf-left y x))))
      (matrix-set! f-right y (- x 1) (- (matrix-ref f-right y (- x 1)) df-u-v))
      (matrix-set! e y x (positive- (matrix-ref e y x) df-u-v))
      (matrix-set! e y (- x 1) (positive+ (matrix-ref e y (- x 1)) df-u-v))
      (enqueue! y (- x 1))))
    (define (push-down! y x)
     ;;(format #t "Push down ~s ~s~%" y x)
     (let ((df-u-v (positive-min (matrix-ref e y x) (cf-down y x))))
      (matrix-set! f-down y x (+ (matrix-ref f-down y x) df-u-v))
      (matrix-set! e y x (positive- (matrix-ref e y x) df-u-v))
      (matrix-set! e (+ y 1) x (positive+ (matrix-ref e (+ y 1) x) df-u-v))
      (enqueue! (+ y 1) x)))
    (define (push-up! y x)
     ;;(format #t "Push up ~s ~s~%" y x)
     (let ((df-u-v (positive-min (matrix-ref e y x) (cf-up y x))))
      (matrix-set! f-down (- y 1) x (- (matrix-ref f-down (- y 1) x) df-u-v))
      (matrix-set! e y x (positive- (matrix-ref e y x) df-u-v))
      (matrix-set! e (- y 1) x (positive+ (matrix-ref e (- y 1) x) df-u-v))
      (enqueue! (- y 1) x)))
    (define (lift! y x)
     ;;(format #t "Lift ~s ~s~%" y x)
     (let ((t (min (if (ef-out? y x) 0 m0)
		   (if (or (= x (- width 1))
			   (not (matrix-ref mask y (+ x 1)))
			   (not (ef-right? y x)))
		       m0
		       (matrix-ref h y (+ x 1)))
		   (if (or (= x 0)
			   (not (matrix-ref mask y (- x 1)))
			   (not (ef-left? y x)))
		       m0
		       (matrix-ref h y (- x 1)))
		   (if (or (= y (- height 1))
			   (not (matrix-ref mask (+ y 1) x))
			   (not (ef-down? y x)))
		       m0
		       (matrix-ref h (+ y 1) x))
		   (if (or (= y 0)
			   (not (matrix-ref mask (- y 1) x))
			   (not (ef-up? y x)))
		       m0
		       (matrix-ref h (- y 1) x)))))
      (when (>= t m1) (fuck-up))
      (matrix-set! h y x (+ t 1))))
    (define (relabel!)
     ;;(format #t "Relabel~%")
     (let ((q '())
	   (tail #f))
      (define (enqueue! y x value)
       (when (> value m1) (fuck-up))
       (when (< value (matrix-ref h y x))
	(matrix-set! h y x value)
	(unless (matrix-ref marked? y x)
	 (matrix-set! marked? y x #t)
	 (cond ((eq? tail #f)
		(set! tail (cons (vector x y) '()))
		(set! q tail))
	       (else (set-cdr! tail (cons (vector x y) '()))
		     (set! tail (cdr tail)))))))
      (define (dequeue!)
       (let ((p (first q)))
	(matrix-set! marked? (y p) (x p) #f)
	(set! q (rest q))
	(when (null? q) (set! tail #f))
	p))
      (do ((y 0 (+ y 1))) ((>= y height))
       (do ((x 0 (+ x 1))) ((>= x width))
	(matrix-set! h y x m1)
	(matrix-set! marked? y x #f)))
      (do ((y 0 (+ y 1))) ((>= y height))
       (do ((x 0 (+ x 1))) ((>= x width))
	(when (and (matrix-ref mask y x)
		   (ef-out? y x)
		   (> (matrix-ref h y x) 1))
	 (enqueue! y x 1))))
      (let loop ()
       (unless (null? q)
	(let* ((p (dequeue!))
	       (x (x p))
	       (y (y p))
	       (value (+ (matrix-ref h y x) 1)))
	 (when (and (> x 0)
		    (matrix-ref mask y (- x 1))
		    (ef-right? y (- x 1)))
	  (enqueue! y (- x 1) value))
	 (when (and (< x (- width 1))
		    (matrix-ref mask y (+ x 1))
		    (ef-left? y (+ x 1)))
	  (enqueue! y (+ x 1) value))
	 (when (and (> y 0)
		    (matrix-ref mask (- y 1) x)
		    (ef-down? (- y 1) x))
	  (enqueue! (- y 1) x value))
	 (when (and (< y (- height 1))
		    (matrix-ref mask (+ y 1) x)
		    (ef-up? (+ y 1) x))
	  (enqueue! (+ y 1) x value)))
	(loop)))))
    (do ((y 0 (+ y 1))) ((>= y height))
     (do ((x 0 (+ x 1))) ((>= x width))
      (matrix-set! e y x 0)
      (matrix-set! f-out y x 0)))
    (do ((y 0 (+ y 1))) ((>= y height))
     (do ((x 0 (+ x 1))) ((>= x (- width 1)))
      (matrix-set! f-right y x 0)))
    (do ((y 0 (+ y 1))) ((>= y (- height 1)))
     (do ((x 0 (+ x 1))) ((>= x width))
      (matrix-set! f-down y x 0)))
    (do ((y 0 (+ y 1))) ((>= y height))
     (do ((x 0 (+ x 1))) ((>= x width))
      (when (or (= x (- width 1))
		(= x 0)
		(= y (- height 1))
		(= y 0)
		(and (matrix-ref mask y x)
		     (or (not (matrix-ref mask y (+ x 1)))
			 (not (matrix-ref mask y (- x 1)))
			 (not (matrix-ref mask (+ y 1) x))
			 (not (matrix-ref mask (- y 1) x)))))
       (matrix-set! e y x -1))))
    (let loop ((i 0) (p? #f))
     (cond
      ((and (zero? (modulo i 6)) (not p?))
       (relabel!)
       (set! relabels (+ relabels 1))
       (cond
	((every-n (lambda (y)
		   (every-n (lambda (x)
			     (or (not (matrix-ref mask y x))
				 (zero? (matrix-ref e y x))
				 (= (matrix-ref h y x) m1)))
			    width))
		  height)
	 ;; Every vertex with excess capacity is not reachable from the sink in
	 ;; the inverse residual network. So terminate early because we have
	 ;; already found a min cut. In this case, the preflows and excess
	 ;; capacities will not be correct. But the cut is indicated by the
	 ;; heights. Vertices reachable from the source have height
	 ;; (+ (* HEIGHT WIDTH) 1) while vertices reachable from the sink have
	 ;; smaller height. Early termination is necessary with relabeling to
	 ;; prevent an infinite loop. The loop arises because vertices that are
	 ;; not reachable from the sink in the inverse residual network have
	 ;; their height reset to (+ (* HEIGHT WIDTH) 1) by the relabeling
	 ;; process. If there are such vertices with excess capacity, this is
	 ;; not high enough for the excess capacity to be pushed back to the
	 ;; perimeter. So after relabeling, vertices get lifted to try to push
	 ;; excess capacity back to the perimeter but then a relabeling happens
	 ;; to soon and foils this lifting. Terminating when all vertices with
	 ;; excess capacity are not reachable from the sink in the inverse
	 ;; residual network eliminates this problem.
	 (format #t "~s push~a, ~s lift~a, ~s relabel~a, ~s wave~a, flow=~s, e.t.~%"
		 pushes
		 (if (= pushes 1) "" "es")
		 lifts
		 (if (= lifts 1) "" "s")
		 relabels
		 (if (= relabels 1) "" "s")
		 i
		 (if (= i 1) "" "s")
		 total-flow))
	(else
	 ;; We need to rebuild the priority queue after relabeling since the
	 ;; heights might have changed and the priority queue is indexed by
	 ;; height. This also assumes that a relabel is done before any pushes
	 ;; or lifts.
	 (do ((k 0 (+ k 1))) ((>= k (vector-length q))) (vector-set! q k '()))
	 (do ((y 0 (+ y 1))) ((>= y height))
	  (do ((x 0 (+ x 1))) ((>= x width))
	   (matrix-set! marked? y x #f)))
	 (do ((y 0 (+ y 1))) ((>= y height))
	  (do ((x 0 (+ x 1))) ((>= x width))
	   (when (and (matrix-ref mask y x)
		      (not (zero? (matrix-ref e y x)))
		      (not (= (matrix-ref h y x) m1)))
	    (enqueue! y x))))
	 (loop i #t))))
      ((some-vector
	(lambda (ps)
	 (some (lambda (p) (not (zero? (matrix-ref e (y p) (x p)))))  ps))
	q)
       (let loop ((k (- (vector-length q) 1)))
	(unless (negative? k)
	 (let ((ps (vector-ref q k)))
	  (vector-set! q k '())
	  (for-each (lambda (p) (matrix-set! marked? (y p) (x p) #f)) ps)
	  (for-each (lambda (p)
		     (let ((x (x p))
			   (y (y p)))
		      (when (can-push-out? y x)
		       (set! pushes (+ pushes 1))
		       (push-out! y x))
		      (when (can-push-right? y x)
		       (set! pushes (+ pushes 1))
		       (push-right! y x))
		      (when (can-push-left? y x)
		       (set! pushes (+ pushes 1))
		       (push-left! y x))
		      (when (can-push-down? y x)
		       (set! pushes (+ pushes 1))
		       (push-down! y x))
		      (when (can-push-up? y x)
		       (set! pushes (+ pushes 1))
		       (push-up! y x))
		      (when (and (not (zero? (matrix-ref e y x)))
				 (< (matrix-ref h y x) m1))
		       ;;(unless (can-lift? y x) (fuck-up))
		       (set! lifts (+ lifts 1))
		       (lift! y x))
		      (unless (or (zero? (matrix-ref e y x))
				  (= (matrix-ref h y x) m1))
		       (enqueue! y x))))
		    ps))
	 (loop (- k 1))))
       (loop (+ i 1) #f))
      ;; All overflowing vertices that cannot be discharged should be
      ;; unreachable from the sink in the inverse residual network. This can
      ;; happen because we don't push back to the source.
      (else (when (some-n (lambda (y)
			   (some-n (lambda (x)
				    (and (matrix-ref mask y x)
					 (not (= (matrix-ref h y x) m1))
					 (not (zero? (matrix-ref e y x)))))
				   width))
			  height)
	     (fuck-up))
	    ;; This is so MIN-CUT and MIN-CUT-INCLUDES-EVERY-EDGE-TO-T? work.
	    (relabel!)
	    (set! relabels (+ relabels 1))
	    (format #t
		    "~s push~a, ~s lift~a, ~s relabel~a, ~s wave~a, flow=~s~%"
		    pushes
		    (if (= pushes 1) "" "es")
		    lifts
		    (if (= lifts 1) "" "s")
		    relabels
		    (if (= relabels 1) "" "s")
		    i
		    (if (= i 1) "" "s")
		    total-flow))))))
  (define (min-cut-includes-every-edge-to-t?)
   ;; This requires that a relabel was done immediately before returning from
   ;; PREFLOW-PUSH!.
   (every-n
    (lambda (y)
     (every-n
      (lambda (x) (or (not (matrix-ref mask y x)) (= (matrix-ref h y x) m1)))
      width))
    height))
  (define (min-cut)
   ;; This requires that a relabel was done immediately before returning from
   ;; PREFLOW-PUSH!.
   (map-n-vector
    (lambda (y)
     (map-n-vector
      (lambda (x) (and (matrix-ref mask y x) (not (= (matrix-ref h y x) m1))))
      width))
    height))
  (let loop ((lg-v lg-max-v) (v-max 0))
   (cond
    ((negative? lg-v)
     (format #t "V-MAX=~s~%" v-max)
     (preflow-push! (+ v-max 1))
     (when (min-cut-includes-every-edge-to-t?) (fuck-up))
     (min-cut))
    (else
     (let ((v (+ v-max
		 (let loop ((i lg-v) (c 1))
		  (if (zero? i) c (loop (- i 1) (+ c c)))))))
      (format #t "LG-V=~s, V-MAX=~s, V=~s~%" lg-v v-max v)
      (preflow-push! v)
      (loop (- lg-v 1) (if (min-cut-includes-every-edge-to-t?) v v-max))))))))

(define (c-ratio-region mask c-right c-down w lg-max-v)
 (let* ((height (matrix-rows w))
	(width (matrix-columns w))
	(c-mask (boolean-matrix->c-int-matrix mask))
	(c-c-right (matrix->c-int-matrix c-right))
	(c-c-down (matrix->c-int-matrix c-down))
	(c-w (matrix->c-int-matrix w))
	(c-marked? (ratio-region-internal
		    height width c-mask c-c-right c-c-down c-w lg-max-v))
	(marked? (c-int-matrix->boolean-matrix c-marked? height width)))
  (free-c-int-matrix! c-mask height)
  (free-c-int-matrix! c-c-right height)
  (free-c-int-matrix! c-c-down (- height 1))
  (free-c-int-matrix! c-w height)
  (free-c-int-matrix! c-marked? height)
  marked?))

;;; File   : sort.scm
;;; Author : Richard A. O'Keefe (based on Prolog code by D.H.D.Warren)
;;; Updated: 11 June 1991
;;; Defines: sorted?, merge, merge!, sort, sort!

;;; --------------------------------------------------------------------
;;; Many Scheme systems provide some kind of sorting functions.  They do
;;; not, however, always provide the _same_ sorting functions, and those
;;; that I have had the opportunity to test provided inefficient ones (a
;;; common blunder is to use quicksort which does not perform well).
;;; Because sort and sort! are not in the standard, there is very little
;;; agreement about what these functions look like.  For example, Dybvig
;;; says that Chez Scheme provides
;;;     (merge predicate list1 list2)
;;;     (merge! predicate list1 list2)
;;;     (sort predicate list)
;;;     (sort! predicate list),
;;; while the MIT Scheme 7.1 manual, following Common Lisp, offers
;;;     (sort list predicate),
;;; TI PC Scheme offers
;;;     (sort! list/vector predicate?)
;;; and Elk offers
;;;     (sort list/vector predicate?)
;;;     (sort! list/vector predicate?)
;;; Here is a comprehensive catalogue of the variations I have found.
;;; (1) Both sort and sort! may be provided.
;;; (2) sort may be provided without sort!
;;; (3) sort! may be provided without sort
;;; (4) Neither may be provided
;;; ---
;;; (5) The sequence argument may be either a list or a vector.
;;; (6) The sequence argument may only be a list.
;;; (7) The sequence argument may only be a vector.
;;; ---
;;; (8) The comparison function may be expected to behave like <
;;; (9) or it may be expected to behave like <=
;;; ---
;;; (10) The interface may be (sort predicate? sequence)
;;; (11) or (sort sequence predicate?)
;;; (12) or (sort sequence &optional (predicate? <))
;;; ---
;;; (13) The sort may be stable
;;; (14) or it may be unstable.
;;; ---
;;; All of this variation really does not help anybody.  A nice simple
;;; merge sort is both stable and fast (quite a lot faster than `quick'
;;; sort).
;;; I am providing this source code with no restrictions at all on its
;;; use (but please retain D.H.D.Warren's credit for the original idea).
;;; You may have to rename some of these functions in order to use them
;;; in a system which already provides incompatible or inferior sorts.
;;; For each of the functions, only the top-level define needs to be
;;; edited to do that.
;;; I could have given these functions names which would not clash with
;;; any Scheme that I know of, but I would like to encourage implementors
;;; to converge on a single interface, and this may serve as a hint.
;;; The argument order for all functions has been chosen to be as close
;;; to Common Lisp as made sense, in order to avoid NIH-itis.

;;; Each of the five functions has a required *last* parameter which is
;;; a comparison function.  A comparison function f is a function of 2
;;; arguments which acts like <.  For example,
;;;     (not (f x x))
;;;     (and (f x y) (f y z)) => (f x z)
;;; The standard functions <, >, char<?, char>?, char-ci<?, char-ci>?,
;;; string<?, string>?, string-ci<?, and string-ci>? are suitable for
;;; use as comparison functions.  Think of (less? x y) as saying when
;;; x must *not* precede y.

;;; (sorted? sequence less?)
;;;     returns #t when the sequence argument is in non-decreasing order
;;;     according to less? (that is, there is no adjacent pair ... x y ...
;;;     for which (less? y x))
;;;     returns #f when the sequence contains at least one out-of-order pair.
;;;     It is an error if the sequence is neither a list nor a vector.

;;; (ok:merge list1 list2 less?)
;;;     This merges two lists, producing a completely new list as result.
;;;     I gave serious consideration to producing a Common-Lisp-compatible
;;;     version.  However, Common Lisp's `sort' is our `sort!' (well, in
;;;     fact Common Lisp's `stable-sort' is our `sort!', merge sort is
;;;     *fast* as well as stable!) so adapting CL code to Scheme takes a
;;;     bit of work anyway.  I did, however, appeal to CL to determine
;;;     the *order* of the arguments.

;;; (ok:merge! list1 list2 less?)
;;;     merges two lists, re-using the pairs of list1 and list2 to build
;;;     the result.  If the code is compiled, and less? constructs no new
;;;     pairs, no pairs at all will be allocated.  The first pair of the
;;;     result will be either the first pair of list1 or the first pair
;;;     of list2, but you can't predict which.

;;;     The code of merge and merge! could have been quite a bit simpler,
;;;     but they have been coded to reduce the amount of work done per
;;;     iteration.  (For example, we only have one null? test per iteration.)

;;; (ok:sort sequence less?)
;;;     accepts either a list or a vector, and returns a new sequence which
;;;     is sorted.  The new sequence is the same type as the input.  Always
;;;     (sorted? (ok:sort sequence less?) less?).
;;;     The original sequence is not altered in any way.  The new sequence
;;;     shares its _elements_ with the old one; no elements are copied.

;;; (ok:sort! sequence less?)
;;;     returns its sorted result in the original boxes.  If the original
;;;     sequence is a list, no new storage is allocated at all.  If the
;;;     original sequence is a vector, the sorted elements are put back
;;;     in the same vector.

;;; Note that these functions do NOT accept a CL-style ":key" argument.
;;; A simple device for obtaining the same expressiveness is to define
;;; (define (keyed less? key) (lambda (x y) (less? (key x) (key y))))
;;; and then, when you would have written
;;;     (ok:sort a-sequence #'my-less :key #'my-key)
;;; in Common Lisp, just write
;;;     (ok:sort! a-sequence (keyed my-less? my-key))
;;; in Scheme.
;;; --------------------------------------------------------------------

;;; (sorted? sequence less?)
;;; is true when sequence is a list (x0 x1 ... xm) or a vector #(x0 ... xm)
;;; such that for all 1 <= i <= m,
;;;     (not (less? (list-ref list i) (list-ref list (- i 1)))).

(define (sorted? seq less?)
 (cond ((null? seq) #t)
       ((vector? seq)
	(let ((n (vector-length seq)))
	 (if (<= n 1)
	     #t
	     (do ((i 1 (+ i 1)))
	       ((or (= i n)
		    (less? (vector-ref seq (- i 1))
			   (vector-ref seq i)))
		(= i n))))))
       (else (let loop ((last (car seq)) (next (cdr seq)))
	      (or (null? next)
		  (and (not (less? (car next) last))
		       (loop (car next) (cdr next))))))))

;;; (ok:merge a b less?)
;;; takes two lists a and b such that (sorted? a less?) and (sorted? b less?)
;;; and returns a new list in which the elements of a and b have been stably
;;; interleaved so that (sorted? (ok:merge a b less?) less?).
;;; Note:  this does _not_ accept vectors.  See below.

(define (ok:merge a b less?)
 (cond ((null? a) b)
       ((null? b) a)
       (else (let loop ((x (car a)) (a (cdr a)) (y (car b)) (b (cdr b)))
	      ;; The loop handles the merging of non-empty lists.  It has
	      ;; been written this way to save testing and car/cdring.
	      (if (less? y x)
		  (if (null? b)
		      (cons y (cons x a))
		      (cons y (loop x a (car b) (cdr b))))
		  ;; x <= y
		  (if (null? a)
		      (cons x (cons y b))
		      (cons x (loop (car a) (cdr a) y b))))))))

;;; (ok:merge! a b less?)
;;; takes two sorted lists a and b and smashes their cdr fields to form a
;;; single sorted list including the elements of both.
;;; Note:  this does _not_ accept vectors.

(define (ok:merge! a b less?)
 (define (loop r a b)
  (if (less? (car b) (car a))
      (begin (set-cdr! r b)
	     (if (null? (cdr b))
		 (set-cdr! b a)
		 (loop b a (cdr b))))
      ;; (car a) <= (car b)
      (begin (set-cdr! r a)
	     (if (null? (cdr a))
		 (set-cdr! a b)
		 (loop a (cdr a) b)))))
 (cond ((null? a) b)
       ((null? b) a)
       ((less? (car b) (car a))
	(if (null? (cdr b))
	    (set-cdr! b a)
	    (loop b a (cdr b)))
	b)
       (else				; (car a) <= (car b)
	(if (null? (cdr a))
	    (set-cdr! a b)
	    (loop a (cdr a) b))
	a)))

;;; (ok:sort! sequence less?)
;;; sorts the list or vector sequence destructively.  It uses a version
;;; of merge-sort invented, to the best of my knowledge, by David H. D.
;;; Warren, and first used in the DEC-10 Prolog system.  R. A. O'Keefe
;;; adapted it to work destructively in Scheme.

(define (ok:sort! seq less?)
 (define (step n)
  (cond ((> n 2)
	 (let* ((j (quotient n 2))
		(a (step j))
		(k (- n j))
		(b (step k)))
	  (ok:merge! a b less?)))
	((= n 2)
	 (let ((x (car seq))
	       (y (cadr seq))
	       (p seq))
	  (set! seq (cddr seq))
	  (if (less? y x)
	      (begin (set-car! p y)
		     (set-car! (cdr p) x)))
	  (set-cdr! (cdr p) '())
	  p))
	((= n 1)
	 (let ((p seq))
	  (set! seq (cdr seq))
	  (set-cdr! p '())
	  p))
	(else
	 '())))
 (if (vector? seq)
     (let ((n (vector-length seq))
	   (vector seq))		; save original vector
      (set! seq (vector->list seq))	; convert to list
      (do ((p (step n) (cdr p))		; sort list destructively
	   (i 0 (+ i 1)))		; and store elements back
	((null? p) vector)              ; in original vector
       (vector-set! vector i (car p))))
     ;; otherwise, assume it is a list
     (step (length seq))))

;;; (ok:sort sequence less?)
;;; sorts a vector or list non-destructively.  It does this by sorting a
;;; copy of the sequence.  My understanding is that the Standard says
;;; that the result of append is always "newly allocated" except for
;;; sharing structure with "the last argument", so (append x '()) ought
;;; to be a standard way of copying a list x.

(define (ok:sort seq less?)
 (if (vector? seq)
     (list->vector (ok:sort! (vector->list seq) less?))
     (ok:sort! (append seq '()) less?)))

(define (cuts->perimeter cuts ppm)
 (let* ((red (ppm-red ppm))
	(green (ppm-green ppm))
	(blue (ppm-blue ppm))
	(new-red (map-vector (lambda (row) (map-vector identity row)) red))
	(new-green (map-vector (lambda (row) (map-vector identity row)) green))
	(new-blue (map-vector (lambda (row) (map-vector identity row)) blue))
	(height (pnm-height ppm))
	(width (pnm-width ppm)))
  (for-each (lambda (cut)
	     (do ((y 1 (+ y 1))) ((>= y (- height 1)))
	      (do ((x 1 (+ x 1))) ((>= x (- width 1)))
	       (when (and (matrix-ref cut y x)
			  (or (not (matrix-ref cut y (- x 1)))
			      (not (matrix-ref cut y (+ x 1)))
			      (not (matrix-ref cut (- y 1) x))
			      (not (matrix-ref cut (+ y 1) x))))
		(matrix-set! new-red y x 255)
		(matrix-set! new-green y x 0)
		(matrix-set! new-blue y x 0)))))
	    cuts)
  (make-ppm (ppm-raw? ppm) 255 new-red new-green new-blue)))

(define (lg x) (/ (log x) (log 2.0)))

(define (binary-search-positionv value vector)
 (let loop ((n (- (inexact->exact (ceiling (lg (vector-length vector)))) 1))
	    (i 0))
  (if (negative? n)
      i
      (let ((i-prime (+ i (expt 2 n))))
       (loop (- n 1)
	     (if (and (< i-prime (vector-length vector))
		      (<= (vector-ref vector i-prime) value))
		 i-prime
		 i))))))

(define (edge-weight r1 r2 g1 g2 b1 b2 weights method p)
 (case method
  ((a)
   (inexact->exact
    (floor (expt (- (* (sqrt 3.0) 255.0)
		    (sqrt (+ (sqr (- r1 r2)) (sqr (- g1 g2)) (sqr (- b1 b2)))))
		 p))))
  ((b)
   (inexact->exact
    (floor (expt (- (* 3.0 (sqr 255.0))
		    (+ (sqr (- r1 r2)) (sqr (- g1 g2)) (sqr (- b1 b2))))
		 p))))
  ((c)
   (inexact->exact
    (floor (expt (+ (sqr (- r1 r2)) (sqr (- g1 g2)) (sqr (- b1 b2))) p))))
  ((d)
   (inexact->exact (floor (expt (+ 1.0
				   (* 100.0
				      (- 1.0 (/ (abs (- r1 r2)) 255.0))
				      (- 1.0 (/ (abs (- g1 g2)) 255.0))
				      (- 1.0 (/ (abs (- b1 b2)) 255.0))))
				p))))
  ((e)
   (- (vector-length weights)
      (binary-search-positionv
       (+ (sqr (- r1 r2)) (sqr (- g1 g2)) (sqr (- b1 b2))) weights)))
  (else (fuck-up))))

(define (sorted-edge-weights ppm)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (list->vector
   (ok:merge
    (ok:sort
     (reduce
      append
      (map-n
       (lambda (y)
	(map-n
	 (lambda (x)
	  (+ (sqr (- (matrix-ref red y x) (matrix-ref red y (+ x 1))))
	     (sqr (- (matrix-ref green y x) (matrix-ref green y (+ x 1))))
	     (sqr (- (matrix-ref blue y x) (matrix-ref blue y (+ x 1))))))
	 (- (pnm-width ppm) 1)))
       (pnm-height ppm))
      '())
     <)
    (ok:sort
     (reduce
      append
      (map-n
       (lambda (y)
	(map-n
	 (lambda (x)
	  (+ (sqr (- (matrix-ref red y x) (matrix-ref red (+ y 1) x)))
	     (sqr (- (matrix-ref green y x) (matrix-ref green (+ y 1) x)))
	     (sqr (- (matrix-ref blue y x) (matrix-ref blue (+ y 1) x)))))
	 (pnm-width ppm)))
       (- (pnm-height ppm) 1))
      '())
     <)
    <))))

(define (c-right ppm method p)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm))
       (weights (sorted-edge-weights ppm)))
  (map-n-vector (lambda (y)
		 (map-n-vector (lambda (x)
				(edge-weight (matrix-ref red y x)
					     (matrix-ref red y (+ x 1))
					     (matrix-ref green y x)
					     (matrix-ref green y (+ x 1))
					     (matrix-ref blue y x)
					     (matrix-ref blue y (+ x 1))
					     weights
					     method
					     p))
			       (- (pnm-width ppm) 1)))
		(pnm-height ppm))))

(define (c-down ppm method p)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm))
       (weights (sorted-edge-weights ppm)))
  (map-n-vector (lambda (y)
		 (map-n-vector (lambda (x)
				(edge-weight (matrix-ref red y x)
					     (matrix-ref red (+ y 1) x)
					     (matrix-ref green y x)
					     (matrix-ref green (+ y 1) x)
					     (matrix-ref blue y x)
					     (matrix-ref blue (+ y 1) x)
					     weights
					     method
					     p))
			       (pnm-width ppm)))
		(- (pnm-height ppm) 1))))

(define (c-right-fake ppm)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (map-n-vector
   (lambda (y)
    (map-n-vector
     (lambda (x)
      (if (and (= (matrix-ref red y x) (matrix-ref red y (+ x 1)))
	       (= (matrix-ref green y x) (matrix-ref green y (+ x 1)))
	       (= (matrix-ref blue y x) (matrix-ref blue y (+ x 1))))
	  128
	  1))
     (- (pnm-width ppm) 1)))
   (pnm-height ppm))))

(define (c-down-fake ppm)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (map-n-vector
   (lambda (y)
    (map-n-vector
     (lambda (x)
      (if (and (= (matrix-ref red y x) (matrix-ref red (+ y 1) x))
	       (= (matrix-ref green y x) (matrix-ref green (+ y 1) x))
	       (= (matrix-ref blue y x) (matrix-ref blue (+ y 1) x)))
	  128
	  1))
     (pnm-width ppm)))
   (- (pnm-height ppm) 1))))

(define (n-by-n-ppm n)
 (let ((red (make-matrix n n 255))
       (green (make-matrix n n 0))
       (blue (make-matrix n n 0)))
  (do ((y (quotient n 4) (+ y 1))) ((>= y (* 3 (quotient n 4))))
   (do ((x (quotient n 4) (+ x 1))) ((>= x (* 3 (quotient n 4))))
    (matrix-set! red y x 0)
    (matrix-set! green y x 255)))
  (make-ppm #t 255 red green blue)))

(define (connected-components cut)
 (setup-proximity-clusterer! 1)
 (map (lambda (points)
       (pbm-bitmap
	(points->pbm-of-size points (matrix-rows cut) (matrix-columns cut))))
      (c-pbm-proximity-clusterer (make-pbm #t cut) 1)))

(define (ratio-region-components c-right c-down w lg-max-v)
 (connected-components
  (ratio-region (make-matrix (matrix-rows w) (matrix-columns w) #t)
		c-right
		c-down
		w
		lg-max-v)))

(define (c-ratio-region-components c-right c-down w lg-max-v)
 (connected-components
  (c-ratio-region (make-matrix (matrix-rows w) (matrix-columns w) #t)
		  c-right
		  c-down
		  w
		  lg-max-v)))

(define (pixel-count cut)
 (let ((count 0))
  (do ((y 0 (+ y 1))) ((>= y (matrix-rows cut)))
   (do ((x 0 (+ x 1))) ((>= x (matrix-columns cut)))
    (when (matrix-ref cut y x) (set! count (+ count 1)))))
  count))

(define (recursive-ratio-region c-right c-down w lg-max-v q-low q-high)
 (let loop ((mask (make-matrix (matrix-rows w) (matrix-columns w) #t)))
  (reduce
   append
   (map
    (lambda (mask)
     (if (< (pixel-count mask) (* q-low (matrix-rows w) (matrix-columns w)))
	 '()
	 (let ((cut (ratio-region mask c-right c-down w lg-max-v)))
	  ;; debugging
	  (format #t "Mask: ~s, Cut: ~s(~s%)~%"
		  (pixel-count mask)
		  (pixel-count cut)
		  (inexact->exact
		   (round (/ (* 100 (pixel-count cut)) (pixel-count mask)))))
	  (if (< (* q-low (matrix-rows w) (matrix-columns w))
		 (pixel-count cut)
		 (* q-high (pixel-count mask)))
	      (append
	       (loop cut)
	       (loop
		(map-vector
		 (lambda (row1 row2)
		  (map-vector
		   (lambda (element1 element2) (and element1 (not element2)))
		   row1 row2))
		 mask cut)))
	      (list mask)))))
    (connected-components mask))
   '())))

(define (c-recursive-ratio-region c-right c-down w lg-max-v q-low q-high)
 (let loop ((mask (make-matrix (matrix-rows w) (matrix-columns w) #t)))
  (reduce
   append
   (map
    (lambda (mask)
     (if (< (pixel-count mask) (* q-low (matrix-rows w) (matrix-columns w)))
	 '()
	 (let ((cut (c-ratio-region mask c-right c-down w lg-max-v)))
	  ;; debugging
	  (format #t "Mask: ~s, Cut: ~s(~s%)~%"
		  (pixel-count mask)
		  (pixel-count cut)
		  (inexact->exact
		   (round (/ (* 100 (pixel-count cut)) (pixel-count mask)))))
	  (if (< (* q-low (matrix-rows w) (matrix-columns w))
		 (pixel-count cut)
		 (* q-high (pixel-count mask)))
	      (append
	       (loop cut)
	       (loop
		(map-vector
		 (lambda (row1 row2)
		  (map-vector
		   (lambda (element1 element2) (and element1 (not element2)))
		   row1 row2))
		 mask cut)))
	      (list mask)))))
    (connected-components mask))
   '())))

(define (write-dimacs pathname c-right c-down w v)
 (call-with-output-file (replace-extension pathname "dimacs")
  (lambda (output-port)
   (let* ((height (matrix-rows w))
	  (width (matrix-columns w))
	  (s (* height width))
	  (t (+ (* height width) 1))
	  (inf 100000000))
    (define (yx y x) (+ (* y width) x))
    (format output-port "p max ~s ~s~%"
	    (+ (* height width) 2)
	    (- (* 5 height width) 4))
    (format output-port "n ~s s~%" s)
    (format output-port "n ~s t~%" t)
    (do ((y 0 (+ y 1))) ((>= y height))
     (do ((x 0 (+ x 1))) ((>= x width))
      (format output-port "a ~s ~s ~s~%" (yx y x) t (* v (matrix-ref w y x)))
      (unless (= x (- width 1))
       (format output-port "a ~s ~s ~s~%"
	       (yx y x) (yx y (+ x 1)) (matrix-ref c-right y x)))
      (unless (= x 0)
       (format output-port "a ~s ~s ~s~%"
	       (yx y x) (yx y (- x 1)) (matrix-ref c-right y (- x 1))))
      (unless (= y (- height 1))
       (format output-port "a ~s ~s ~s~%"
	       (yx y x) (yx (+ y 1) x) (matrix-ref c-down y x)))
      (unless (= y 0)
       (format output-port "a ~s ~s ~s~%"
	       (yx y x) (yx (- y 1) x) (matrix-ref c-down (- y 1) x)))))
    (do ((x 0 (+ x 1))) ((>= x width))
     (format output-port "a ~s ~s ~s~%" s (yx (- height 1) x) inf)
     (format output-port "a ~s ~s ~s~%" s (yx 0 x) inf))
    (do ((y 1 (+ y 1))) ((>= y (- height 1)))
     (format output-port "a ~s ~s ~s~%" s (yx y (- width 1)) inf)
     (format output-port "a ~s ~s ~s~%" s (yx y 0) inf))))))

(define (ratio-region-cost c-right c-down w cut)
 (let ((height (matrix-rows w))
       (width (matrix-columns w)))
  (/ (+ (sum (lambda (y)
	      (sum (lambda (x)
		    (if (or (and (matrix-ref cut y x)
				 (not (matrix-ref cut y (+ x 1))))
			    (and (not (matrix-ref cut y x))
				 (matrix-ref cut y (+ x 1))))
			(matrix-ref c-right y x)
			0))
		   (- width 1)))
	     height)
	(sum (lambda (y)
	      (sum (lambda (x)
		    (if (or (and (matrix-ref cut y x)
				 (not (matrix-ref cut (+ y 1) x)))
			    (and (not (matrix-ref cut y x))
				 (matrix-ref cut (+ y 1) x)))
			(matrix-ref c-down y x)
			0))
		   width))
	     (- height 1)))
     (sum (lambda (y)
	   (sum (lambda (x) (if (matrix-ref cut y x) (matrix-ref w y x) 0))
		width))
	  height))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
