/* LaHaShem HaAretz U'Mloah */

/* Program to train and evaluate hidden Markov models */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "hlib-c.h"
#include "hmm.h"

static RMat g_alpha = NULL;
static RMat g_delta = NULL;
static RMat g_deltav = NULL;	/* This could be int. */
static RMat g_beta = NULL;
static RMat3d g_psi = NULL;
static Real g_HMM_errorSumTol = 0.01;

/* -- utility functions -- */

Hmm *allocHMM(int uu);
void constantHMM(Hmm *m, int flags);
void copyHMM(Hmm *m1, Hmm *m2);
void displayHMM(FILE *f, Hmm *m);
void freeHMM(Hmm *m);
void normaliseHMM(Hmm *m);
void randomiseHMM(Hmm *m, int flags);
void zeroHMM(Hmm *m);

/* -- processing functions -- */

void HMM_calcAlphas(Hmm *m, RVec *logL);
void HMM_calcDeltas(Hmm *m, RVec *logL);
void HMM_initGlobals(int uu, int tt);
Real HMM_logL(Hmm *m, RVec *logL);
int *HMM_best_state_sequence(Hmm *m, RVec *logL);
Real HMM_updateModel(Hmm *m, Hmm *new, RVec *logL, RVec *gamma);

/* -- utility functions -- */

/* allocHMM - Allocates space for an HMM and returns a pointer to it. */
Hmm *allocHMM(int uu)
{ Hmm *m = (Hmm *)safe_malloc(sizeof(Hmm));
  assert(uu>0);
  m->logA = allocRMat(uu, uu);
  m->logB = allocRVec(uu);
  m->uu = uu;
  return m;}

/* constantHMM -
 *  Initialises an HMM with constant transition log probabilities.
 */
void constantHMM(Hmm *m, int flags)
{ int u, v, uu = m->uu;
  zeroHMM(m);
  for (u = 0; u<uu; u++)
  { VECTOR(m->logB)[u] = 0;
    for (v = (flags==0)?0:u; v<uu; v++) MATRIX(m->logA)[u][v] = 0;}
  normaliseHMM(m);}

void copyHMM(Hmm *m1, Hmm *m2)
{ int u, uu = m1->uu;
  assert(m1->uu==m2->uu);
  copyRVec(VECTOR(m1->logB), VECTOR(m2->logB), uu);
  for (u = 0; u<uu; u++)
  { copyRVec(MATRIX(m1->logA)[u], MATRIX(m2->logA)[u], uu);}}

/* displayHMM - Writes the HMM `m' in a human-readable format to output
 *  stream `f'.  Only displays a maximum of DISPLAY_HMM_MAX_U states
 *  on a single line.
 */
#define DISPLAY_HMM_MAX_U 10
void displayHMM(FILE *f, Hmm *m)
{ int i, u, uu = m->uu;
  /* display B vector */
  fprintf(f, "%8s", "");
  for (u = 0; u<DISPLAY_HMM_MAX_U&&u<uu; u++) fprintf(f, " %4d", u);
  fprintf(f, "\n");
  fprintf(f, "HMM:  B:");
  for (u = 0; u<DISPLAY_HMM_MAX_U&&u<uu; u++)
  { fprintf(f, " %4.2f", my_exp(VECTOR(m->logB)[u]));}
  fprintf(f, "\n");
  for (;u<uu; u++)
  { fprintf(f, "\n");
    fprintf(f, "%8s", "");
    for (i = 0; i<DISPLAY_HMM_MAX_U&&i+u<uu; u++, i++) fprintf(f, " %4d", i+u);
    fprintf(f, "\n");
    for (i = 0; i<DISPLAY_HMM_MAX_U&&u<uu; u++, i++)
    { fprintf(f, " %4.2f", my_exp(VECTOR(m->logB)[u]));}
    fprintf(f, "\n");}
  /* display A matrix */
  fprintf(f, "%8s", "A:");
  for (u = 0; u<DISPLAY_HMM_MAX_U&&u<uu; u++) fprintf(f, " %4d", u);
  fprintf(f, "\n");
  for (u = 0; u<DISPLAY_HMM_MAX_U&&u<uu; u++)
  { fprintf(f, "%8d", u);
    for (i = 0; i<DISPLAY_HMM_MAX_U&&i<uu; i++)
    { fprintf(f, " %4.2f", my_exp(MATRIX(m->logA)[u][i]));}
    fprintf(f, "\n");}}

/* freeHMM - Frees a previously allocated HMM. */
void freeHMM(Hmm *m)
{ freeRMat(m->logA);
  freeRVec(m->logB);
  free(m);}

/* normaliseHMM -
 *  Normalises an HMM so that the transition log probabilities from each
 *  state sum to zero.
 */
void normaliseHMM(Hmm *m)
{ int u, v, uu = m->uu;
  Real sum;
  sum = sumOfLogs(VECTOR(m->logB), uu);
  assert(sum>NEGATIVE_INFINITY);
  for (u = 0; u<uu; u++) VECTOR(m->logB)[u] -= sum;
  for (u = 0; u<uu; u++)
  { sum = sumOfLogs(MATRIX(m->logA)[u], uu);
    assert(sum>NEGATIVE_INFINITY);
    for (v = 0; v<uu; v++) MATRIX(m->logA)[u][v] -= sum;}}

/* randomiseHMM -
 *  Initialises an HMM with normalised random transition log probabilities.
 */
void randomiseHMM(Hmm *m, int flags)
{ int u, v, uu = m->uu;
  zeroHMM(m);
  for (u = 0; u<uu; u++)
  { VECTOR(m->logB)[u] = my_log(RANDOM);
    for (v = (flags==0)?0:u; v<uu; v++)
    { MATRIX(m->logA)[u][v] = my_log(RANDOM);}}
  normaliseHMM(m);}

/* zeroHMM - Sets all of the transition probabilities to 0.0. */
void zeroHMM(Hmm *m)
{ int u, v, uu = m->uu;
  for (u = 0; u<uu; u++)
  { VECTOR(m->logB)[u] = NEGATIVE_INFINITY;
    for (v = 0; v<uu; v++)
    { MATRIX(m->logA)[u][v] = NEGATIVE_INFINITY;}}}

/* -- processing functions -- */

/* HMM_calcAlphas -
 *  Given an HMM and a vector of log likelihoods for states assigned to
 *  data vectors in the sequence, calculates the alpha values from the
 *  forward-backward algorithm.
 */
void HMM_calcAlphas(Hmm *m, RVec *logL)
{ int t, u, v, uu = m->uu, tt = VLENGTH(logL[0]);
  Real **al, **a = MATRIX(m->logA), *b = VECTOR(m->logB);
  HMM_initGlobals(uu, tt);
  al = MATRIX(g_alpha);
  /* alpha[u][1] = b[u]*logL[u][1]
   * alpha[u][t] = sum(v = 1 -> uu, a[v][u]*alpha[v][t-1])*logL[u][t]
   */
  for (u = 0; u<uu; u++) al[u][0] = VECTOR(logL[u])[0]+b[u];
  for (t = 1; t<tt; t++)
  { for (u = 0; u<uu; u++)
    { al[u][t] = NEGATIVE_INFINITY;
      for (v = 0; v<uu; v++)
      { al[u][t] = add_exp(al[u][t], a[v][u]+al[v][t-1]);}
      al[u][t] += VECTOR(logL[u])[t];}}}

/* HMM_calcDeltas -
 *  Given an HMM and a vector of log likelihoods for states assigned to
 *  data vectors in the sequence, calculates the delta values from the
 *  forward-backward algorithm.
 */
void HMM_calcDeltas(Hmm *m, RVec *logL)
{ int t, u, v, uu = m->uu, tt = VLENGTH(logL[0]);
  /* dev could be int. */
  Real **de, **dev, **a = MATRIX(m->logA), *b = VECTOR(m->logB), d;
  HMM_initGlobals(uu, tt);
  de = MATRIX(g_delta);
  dev = MATRIX(g_deltav);	/* This could be int. */
  /* delta[u][1] = b[u]*logL[u][1]
   * delta[u][t] = max(v = 1 -> uu, a[v][u]*delta[v][t-1])*logL[u][t]
   */
  for (u = 0; u<uu; u++) de[u][0] = VECTOR(logL[u])[0]+b[u];
  for (t = 1; t<tt; t++)
  { for (u = 0; u<uu; u++)
    { de[u][t] = NEGATIVE_INFINITY;
      dev[u][t] = 0;		/* This could be int. */
      for (v = 0; v<uu; v++)
      { d = a[v][u]+de[v][t-1];
	if (d>de[u][t])
        { de[u][t] = d;
	  dev[u][t] = v;}}	/* This could be int. */
      de[u][t] += VECTOR(logL[u])[t];}}}

/* HMM_initGlobals -
 *  Given a number of states and a sequence length, allocates or reallocates
 *  memory for internal variables.
 */
void HMM_initGlobals(int uu, int tt)
{ if (g_alpha==NULL||COLUMNS(g_alpha)<tt||ROWS(g_alpha)<uu)
  { if (g_alpha!=NULL)
    { freeRMat(g_alpha);
      freeRMat(g_delta);
      freeRMat(g_deltav);	/* This could be int. */
      freeRMat(g_beta);
      freeRMat3d(g_psi);}
    g_alpha = allocRMat(uu, tt);
    g_delta = allocRMat(uu, tt);
    g_deltav = allocRMat(uu, tt); /* This could be int. */
    g_beta = allocRMat(uu, tt);
    g_psi = allocRMat3d(uu, uu, tt);}}

/* HMM_logL -
 *  Given an HMM and a vector of log likelihoods for states assigned to
 *  data vectors in the sequence, returns the log likelihood of the data
 *  given the HMM.
 */
Real HMM_logL(Hmm *m, RVec *logL)
{ int u, tt = VLENGTH(logL[0]);
  Real l = NEGATIVE_INFINITY;
  /* P(sequence|m) = sum(u = 1->uu, alpha[u][tt]) */
  HMM_calcAlphas(m, logL);
  for (u = 0; u<m->uu; u++) l = add_exp(l, MATRIX(g_alpha)[u][tt-1]);
  return l;}

/* HMM_best_state_sequence -
 *  Given an HMM and a vector of log likelihoods for states assigned to
 *  data vectors in the sequence, returns the best state sequence of the data
 *  given the HMM.
 */
int *HMM_best_state_sequence(Hmm *m, RVec *logL)
{ int u, t, tt = VLENGTH(logL[0]), *rv;
  Real b, d;
  rv = (int *)safe_malloc(tt*sizeof(int));
  HMM_calcDeltas(m, logL);
  b = NEGATIVE_INFINITY;
  rv[tt-1] = 0;
  for (u = 0; u<m->uu; u++)
  { d = MATRIX(g_delta)[u][tt-1];
    if (d>b)
    { b = d;
      rv[tt-1] = u;}}
  /* This could be int. */
  for (t = tt-2; t>=0; t--) rv[t] = MATRIX(g_deltav)[rv[t+1]][t+1];
  return rv;}

/* HMM_updateModel -
 *  Given an HMM and a vector of log likelihoods for states in the sequences,
 *  calculates the responsibilities of each state in the HMM for each symbol
 *  in the sequences, and maximises the model parameters based on the
 *  assigned log likelihoods.
 */
Real HMM_updateModel(Hmm *m, Hmm *new, RVec *logL, RVec *gamma)
{ int t, u, v, uu = m->uu, tt = VLENGTH(logL[0]);
  Real **a = MATRIX(m->logA), *b = VECTOR(m->logB), **al, **be, ***ps;
  Real sum, like;
  assert(VLENGTH(logL[0])==VLENGTH(gamma[0]));
  HMM_initGlobals(uu, tt);
  al = MATRIX(g_alpha);
  be = MATRIX(g_beta);
  ps = MATRIX(g_psi);
  /* calculate alpha's */
  HMM_calcAlphas(m, logL);
  /* calculate beta's -
   * beta[u][tt] = 1
   * beta[u][t] = sum(v = 1 -> uu, a[u][v]*beta[v][t+1]*logL[v][t+1])
   */
  for (u = 0; u<uu; u++) be[u][tt-1] = 0.0;
  for (t = tt-2; t>=0; t--)
  { for (u = 0; u<uu; u++)
    { be[u][t] = NEGATIVE_INFINITY;
      for (v = 0; v<uu; v++)
      { be[u][t] =
	add_exp(be[u][t], a[u][v]+be[v][t+1]+VECTOR(logL[v])[t+1]);}}}
  /* calculate logL of sequence -
   * P(sequence|m) = sum(u = 1->uu, alpha[u][tt])
   */
  like = NEGATIVE_INFINITY;
  for (u = 0; u<uu; u++) like = add_exp(like, al[u][tt-1]);
  /* calculate responsibilities
   *               alpha[u][t]*beta[u][t]
   * gamma[u][t] = ----------------------
   *                    P(data|model)
   */
  for (t = 0; t<tt; t++)
  { for (u = 0; u<uu; u++) VECTOR(gamma[u])[t] = al[u][t]+be[u][t]-like;}
  /* calculate time-indexed transition probabilities
   *                alpha[u][t]*a[u][v]*logL[v][t+1]*beta[v][t+1]
   * psi[u][v][t] = ---------------------------------------------
   *                               P(data|model)
   */
  for (u = 0; u<uu; u++)
  { for (v = 0; v<uu; v++)
    { for (t = 0; t<tt-1; t++)
      { ps[u][v][t] = al[u][t]+a[u][v]+VECTOR(logL[v])[t+1]+be[v][t+1]-like;}}}
  /* update new model */
  a = MATRIX(new->logA);
  b = VECTOR(new->logB);
  /* calculate B
   * b[u] = gamma[u][1]
   * - added scaling by sum of gammas to catch any numerical accuracy problems
   */
  sum = NEGATIVE_INFINITY;
  for (u = 0; u<uu; u++)
  { b[u] = add_exp(b[u], VECTOR(gamma[u])[0]);
    sum = add_exp(sum, VECTOR(gamma[u])[0]);}
  if (fabs(sum)>g_HMM_errorSumTol)
  { reportWarn("B sum disparity exceeds tolerance: %f\n", my_exp(fabs(sum)));}
  /* calculate A matrix
   *                    sum(t = 1->tt-1, psi[u][v][t])
   * a[u][v] = ----------------------------------------------
   *           sum(t = 1->tt-1, sum(w = 1->uu, psi[u][w][t]))
   */
  for (u = 0; u<uu; u++)
  { for (v = 0; v<uu; v++)
    { a[u][v] = add_exp(a[u][v], sumOfLogs(ps[u][v], tt-1));}}
  for (t = 0; t<tt; t++)
  { for (u = 0; u<uu; u++) VECTOR(gamma[u])[t] = my_exp(VECTOR(gamma[u])[t]);}
  return like;}

/* Tam V'Nishlam Shevah L'El Borei Olam */
